/* C/bricks/API/API0/anaLib0_kernel.c
 * 
 * implementation of API calls for the so plugin functioning mode
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch> 
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */
#include "anaLib0.h"
//#include "anaError.h"
/*
anaHandle_t anaL0_attach(char *brickName, void *pluginHandle, pthread_t mainThread, pthread_t dataThread){
	
	struct commSpecs_s auxComm;
	memset(&auxComm,0, sizeof(struct commSpecs_s));
	
	auxComm.msgQueue = messageQueue;
	auxComm.queueLock = &queueLock;
	auxComm.condLock = &waitLock;
	auxComm.pluginHandle = pluginHandle;
	auxComm.mainThread = mainThread;
	auxComm.dataThread = dataThread;	
	
	return handleAttach(brickName, &auxComm);
}

*/


/* this function is the one that receives all notifications for the brick and then calls the proper event handler */
void notificationWrapper(struct anaEventNotif_s *notif, int len, anaLabel_t input,  void *aux){
	/* aux contains the pointer to the event handler */
	eventHandler_t *Phandler = (eventHandler_t *) aux;
	eventHandler_t handler = *Phandler;

	handler(notif);
	
	if(BRICK_ATTACHED_EVENT&notif->eventID || BRICK_DETACHED_EVENT&notif->eventID || IDP_UNPUBLISHED_EVENT&notif->eventID)
		free(notif->name);
	
	free(aux);
}


int anaL0_eventSubscribe(struct anaEventSub_s* sub, eventHandler_t handler, int threadFlag){
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->registerEventHandler(MY_HANDLE, MY_HANDLE, sub, (anaCallback_t) notificationWrapper, &handler,  sizeof(void*), threadFlag);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
}


int anaL0_eventUnsubscribe(struct anaEventSub_s* sub){
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->unregisterEventHandler(MY_HANDLE, sub );
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
	
}


int anaL0_eventNotify( struct anaEventNotif_s *notif){
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->notifyEvent(MY_HANDLE, notif );
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
	
}

int anaL0_getIDPinfo(anaLabel_t label,  anaHandle_t viewer, struct IDPinfo_s *result){
	int ret = -1;
	
	if(result == NULL)
		return -1;	
	
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->getIDPinfo(MY_HANDLE, viewer, label, result);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
	
}

int anaL0_setIDPinfo(struct IDPinfo_s *info){
	int ret = -1;
	anaHandle_t callee;
	
	if(info == NULL)
		return -1;	
	
	callee = (info->binding).brick.handle;
	
	if(callee == 0) {
		info->IDPtype = BRICK;
		(info->binding).brick.handle = MY_HANDLE;
	}
	
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->setIDPinfo(MY_HANDLE, info);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
	
}


int anaL0_redirect(const anaLabel_t originLabel, const anaHandle_t originViewer, const anaLabel_t destLabel, const anaHandle_t destViewer){
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->redirect(myHandle, originViewer, originLabel,  destViewer, destLabel);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
}


int anaL0_forkData(const anaLabel_t originLabel, const anaHandle_t originViewer, const anaLabel_t destLabel,const anaHandle_t destViewer){
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->forkData(myHandle , originViewer, originLabel, destViewer, destLabel);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
}

/*???*/
int anaL0_send(const anaLabel_t label, void *msg, int len) {
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->callback(myHandle, label, myHandle, msg, len);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
}

anaLabel_t anaL0_registerCallback( anaCallback_t fct, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux, 
						const anaHandle_t viewer, int perm, int thread) {
	int ret = -1;
	anaLabel_t label;
	analock_read_lock(&attachLock);
	if(attached) {
		ret =  myminmex->regist(myHandle, fct, (void *)aux, copyAux,  freeAux, viewer, perm, thread, &label);
	} else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return label;
}


int anaL0_changeLabelAux(const anaLabel_t label, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux){
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = 	myminmex->changeLabelAux( myHandle, label, aux,  copyAux,  freeAux);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
}


/* sends a request to the minmex to unregister a callBack IDP. 
 * The entry in the shadowIDT corresponding to this IDP is also removed 
 */
int anaL0_unregisterCallback(const anaLabel_t label, anaHandle_t viewer) {
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->unregist(myHandle, viewer, label);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
}


/* this function sends a message to the MINMEX asking to change an IDP viewer*/
int anaL0_changeLabelView(const anaLabel_t label, const anaHandle_t oldViewer,  const anaHandle_t newViewer){
	int ret = -1;
	analock_read_lock(&attachLock);
	if(attached)
		ret = myminmex->changeView(myHandle,  oldViewer, label, newViewer);
	else
		ret = ANA_NOMINMEX_ERROR;
	analock_read_unlock(&attachLock);
	return ret;
}	


/* this function stops ANA activity correctly*/
void anaL0_stopANA(int ret){
	brick_exit();
	anatimer_exit();
	anaThreadExit();
	analock_write_lock(&attachLock);
	if(attached)
		myminmex->detach(myHandle, 255);
		attached = 0;
	analock_write_unlock(&attachLock);

}

void wrapInThread(struct dataMessageQueue_s *msg){

	if(msg == NULL)
		return;
		
	/* we launch the callBack function */
	msg->callBackFct(msg->data, msg->dataLen, msg->inputLabel, msg->aux);
	
	/* and we free the message  IMPORTANT : WE DO NOT TOUCH THE AUX*/
	free(msg->data);
	free(msg);
}

void heartBeat(void){
	anaPrint(ANA_DEBUG, "+++heart beat timer started for handle %i\n", MY_HANDLE);
	analock_read_lock(&attachLock);
	if(attached)
		myminmex->heartBeat(MY_HANDLE);
	analock_read_unlock(&attachLock);
}

#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
void recvData(struct work_struct *work) {
#else
void recvData(void *work) {
#endif
	struct dataMessageQueue_s *msg;
	if(recvDataQueue == NULL)
		anaPrint(ANA_ERR, "recvData no message received\n");	
	while(recvDataQueue != NULL) {
		/* we unqueue the head message */
		analock_lock(&queueLock);
		msg = recvDataQueue;
		recvDataQueue = msg->next;
		analock_unlock(&queueLock);

		if (msg->thread) {
			_anaThread(myHandle, (void *) wrapInThread, msg, NULL);
		}
		else {
			/* we execute the callback function*/
			msg->callBackFct(msg->data, msg->dataLen, msg->inputLabel, msg->aux);
			/* and we free the data  IMPORTANT : WE DO NOT TOUCH THE AUX*/
			free(msg->data);
			free(msg);
		}
	}
}



















