/* C/bricks/API/AL1/anaLib1_multi.c
 * 
 * implementation of the al1 functions
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "anaLib1.h"
xrpSpecs_t defaultXrpSpecs = {DEFAULT_XRP_MSGSIZE, DEFAULT_XRP_CMDSIZE, DEFAULT_XRP_CLASSSIZE, DEFAULT_XRP_VERSION};


/*
int anaL1_initDefault(void){

	defaultXrpSpecs.maxMsgSize = DEFAULT_XRP_MSGSIZE;	 		
	defaultXrpSpecs.commandSize = DEFAULT_XRP_CMDSIZE; 		
	defaultXrpSpecs.classSize = DEFAULT_XRP_CLASSSIZE;
	defaultXrpSpecs.xrpSpecVersion = DEFAULT_XRP_VERSION; 
	return 0;	
}
*/


int anaL1_encData(xrpMsg_t msg, void *data , int lenData, struct context_s *senderContext, struct service_s *senderService, const anaLabel_t responseLabel){ 
	int totalLength=0, arglen;
	
	if (!msg || !data || lenData<=0 ){ /*senderID and responseLabel can be NULL*/
		anaPrint(ANA_ERR,  "al1_encData : NULL Arguments \n");
		return -1;
	}
	
	
	totalLength = fillXRPCommand(msg, XRP_CMD_DATA);

	if (totalLength<0) {
                anaPrint(ANA_ERR, "anaL1_encData: Unable to fill command in message\n");
		return -1;
        }  
	
	arglen = addXRPArg(msg, XRP_CLASS_MESSAGE, data, lenData);
	
        if (arglen<0) {
        	anaPrint(ANA_ERR, "anaL1_encData: Unable to add data to message\n");
		return -1;
        }
	
	totalLength += arglen;
	
	if (senderContext!=NULL){
		
		arglen = addXRPArg(msg, XRP_CLASS_SRC_CXT, senderContext->value, senderContext->valueLen);
		
		if (arglen<0) {
        		anaPrint(ANA_ERR, "anaL1_encData: Unable to add sender context to message\n");
			return -1;
        	}		
		
		totalLength += arglen;
	}
	
	
	if (senderService!=NULL){
	
		arglen = addXRPArg(msg, XRP_CLASS_SRC_SRV, senderService->value, senderService->valueLen);
	
		if (arglen<0) {
        		anaPrint(ANA_ERR, "anaL1_encData: Unable to add sender context to message\n");
        	        return -1;
        	}		
		
		totalLength += arglen;
	}
	
	if (responseLabel>0){
	
		arglen = addXRPArg(msg, XRP_CLASS_LABEL, (void *)(&responseLabel), sizeof(anaLabel_t));
	
		if (arglen<0) {
        		anaPrint(ANA_ERR, "anaL1_encData: Unable to add response Label to message\n");
        	        return -1;
        	}		
		
		totalLength += arglen;
	}
	
	return totalLength;
}



/*registers a callBack IDP to the minmex. adds the returned label to the XRP command in order to inform the detinator
where to send the reply, and finally sends the message to the minmex */
/* int thread: whether the function processing the reply should be executed as dedicated thread.
 * should be set to 1 if a) the function executes "long" or b) the function wants to receive further messages
 */

int anaL1_requestReply(const anaLabel_t label, void *msg, int len, anaCallback_t fct, void *aux ,copyAuxFunction cpy, freeAuxFunction fr, int thread){

	int arglen; 
	int ret;
	
	/* First we register a callBackIDP for handling the reply */
	anaLabel_t callBackIDP;
	if ( !msg || len<=0 || !fct ){
		anaPrint(ANA_ERR,  "requestReply : NULL Arguments \n");
		return -1;
	}
		
		
	/* WARNING : TOBE CHANGED: for now we assume the labels ceated with request reply 
 	 * are volatile (the 0 argument in registerCallBack)*/
	
	callBackIDP = anaL0_registerCallback(fct, aux, cpy, fr, 0, 0, thread); 
	if (callBackIDP == 0){
		anaPrint(ANA_ERR,  "requestReply : Unable to register Callback IDP to ANA minmex \n");
		return -1;
	}
	
	
	/* we add the callBack label to the message in order to indicate to the receiver where to send the response */
	arglen = addXRPArg((xrpMsg_t) msg, XRP_CLASS_LABEL, &callBackIDP, sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_ERR, "anaL1_requestReply : Unable to add response label to message\n");
		return -1;
	}
	
	/* we update the total length of the message */
	len += arglen;
	
	/* Then we send the message */
	ret = anaL0_send(label, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR,  "requestReply : Unable to send message to  ANA minmex \n");
		return -1;
	}
	
	return len;
	
}

xrpMsg_t anaL1_allocateMessage(void){
	
	xrpMsg_t msg=NULL;
	
	if (allocateXRPMsg(&msg)<0)
		return NULL;
	return msg;
}


int anaL1_encResolve(xrpMsg_t msg, struct context_s *context, struct service_s *target, char chanType, struct service_s *querierService) {

        int totalLength=0, arglen;

        if (msg == NULL || target==NULL || target->valueLen>512 ) /* if context is NULL we will replace it with "."*/ 
                return -1;
        
		
	/*first we clean the message buffer*/
        memset(msg, 0, defaultXrpSpecs.maxMsgSize);

	totalLength = fillXRPCommand(msg, XRP_CMD_RESOLVE);

        if (totalLength<0) {
                anaPrint(ANA_ERR, "anaL1_encResolve: Unable to fill command in message\n");
                return -1;
        }

        arglen = addXRPArg(msg, XRP_CLASS_DST_SRV, target->value,  target->valueLen);

        if (arglen<0) {
        	anaPrint(ANA_ERR, "anaL1_encResolve : Unable to add query as an argument\n");
                return -1;
        }
	
	totalLength += arglen;
	
	if (context != NULL && context->valueLen <= 512 )
		arglen = addXRPArg(msg, XRP_CLASS_DST_CXT, context->value, context->valueLen);
	else
		arglen = addXRPArg(msg, XRP_CLASS_DST_CXT, ".", 2);
	
	if (arglen<0) {
        	anaPrint(ANA_ERR, "anaL1_encResolve : Unable to add context as an argument\n");
               	return -1;
        }
	
	totalLength += arglen;
	
        
	/*add the chan Type parameter*/
	arglen = addXRPArg(msg, XRP_CLASS_CHANTYPE, &chanType, 1);
	
	if (arglen < 0) {
        	anaPrint(ANA_ERR, "anaL1_encResolve : Unable to add aggregate parameter as an argument\n");
                return -1;
        }
	totalLength += arglen;
	
	
	if (querierService!=NULL && querierService->valueLen <=512 ){
	
		arglen = addXRPArg(msg, XRP_CLASS_SRC_SRV, querierService->value, querierService->valueLen);
	
		if (arglen<0){
			anaPrint(ANA_ERR, "anaL1_encResolve unable to add description\n");
		}
		
		totalLength += arglen;
	}
	
        return totalLength;
}

int anaL1_encLookup(xrpMsg_t msg, struct context_s *context, struct service_s *target, struct service_s *querierService) {

        int totalLength=0, arglen;

        if (msg == NULL || target==NULL || target->valueLen>512) 
                return -1;
		
		
        /*first we clean the message buffer*/
        memset(msg, 0, defaultXrpSpecs.maxMsgSize);

	totalLength = fillXRPCommand(msg, XRP_CMD_LOOKUP);

        if (totalLength<0) {
                anaPrint(ANA_ERR, "anaL1_encLookup: Unable to fill command in message\n");
                return -1;
        }

        arglen = addXRPArg(msg, XRP_CLASS_DST_SRV, target->value,  target->valueLen);

        if (arglen<0) {
        	anaPrint(ANA_ERR, "anaL1_encLookup : Unable to add query as an argument\n");
                return -1;
        }
	
	totalLength += arglen;
	
	if (context!=NULL && context->valueLen <= 512)
		arglen = addXRPArg(msg, XRP_CLASS_DST_CXT, context->value, context->valueLen);
	else
		arglen = addXRPArg(msg, XRP_CLASS_DST_CXT, ".", 2);
		
	if (arglen<0) {
        	anaPrint(ANA_ERR, "anaL1_encLookup : Unable to add context as an argument\n");
               	return -1;
        }
		
	totalLength += arglen;
	        
	
	if (querierService!=NULL && querierService->valueLen <= 512){
	
		arglen = addXRPArg(msg, XRP_CLASS_SRC_SRV, querierService->value, querierService->valueLen);
	
		if (arglen<0){
			anaPrint(ANA_ERR, "anaL1_encLookup unable to add querier description\n");
		}
		
		totalLength += arglen;
	}

        return totalLength;
}



/*encodes a Publish request  :
	- keyTab is a char* Table conaining all keywords associated with entry to puclish
	- int nbKeys is the number of elements of the keyTab.
	- char *name is the name of the entry to publish
	- IDP is the label associated with the entry
	- int labelSize
*/
int anaL1_encCompartmentPublish(xrpMsg_t msg, const anaLabel_t idp, struct context_s *ctxt, struct service_s *srv, int permanent) {
	
	int totalSize=0,  arglen;
	
	if (!idp || !msg || !srv || !ctxt || srv->valueLen > 512 || ctxt->valueLen > 512)
		return -1;
	
	/*first we clean the message buffer*/
	memset(msg, 0, defaultXrpSpecs.maxMsgSize);
	
	totalSize = fillXRPCommand(msg, XRP_CMD_PUBLISH);
	
	if (totalSize<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentPublish : Unable to fill command in message\n");
		return -1;
	}	
	

	arglen = addXRPArg(msg, XRP_CLASS_LABEL, (void *)(&idp), sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentPublish : Unable to add label to message\n");
		return -1;
	}
	totalSize += arglen;
	
	arglen = addXRPArg(msg, XRP_CLASS_SRC_SRV, srv->value, srv->valueLen);
	if (arglen<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentPublish : Unable to add description to message\n");
		return -1;
	}
	totalSize += arglen;
	
	arglen = addXRPArg(msg, XRP_CLASS_SRC_CXT, ctxt->value, ctxt->valueLen);
	if (arglen<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentPublish : Unable to add context to message\n");
		return -1;
	}
	totalSize += arglen;
		
	if (permanent){ /* caller wants to publish a permanent IDP */
		char c[1];
		c[0]='y';
		
		arglen = addXRPArg(msg, XRP_CLASS_PERMANENT, c, 1);
		if (arglen<0) {
			anaPrint(ANA_ERR, "anaL1_encCompartmentPublish : Unable to add Permanent flag to message\n");
			return -1;
		}
		totalSize += arglen;
	}
	
	return totalSize;	 
}

int anaL1_encCompartmentUnpublish(xrpMsg_t msg, const anaLabel_t idp, struct context_s *ctxt, struct service_s *srv) {
	int totalLength=0, arglen;
	
	if (!msg || !srv || !ctxt || srv->valueLen > 512 || ctxt->valueLen > 512)
		return -1;
	
	
	/* first we clean the message buffer */
	memset(msg, 0, defaultXrpSpecs.maxMsgSize);
	
	totalLength = fillXRPCommand(msg, XRP_CMD_UNPUBLISH);
	
	if (totalLength<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentUnpublish : Unable to fill command in message\n");
		return -1;
	}	


	arglen = addXRPArg(msg, XRP_CLASS_LABEL, (void *)(&idp), sizeof(anaLabel_t));
	if (arglen<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentUnpublish : Unable to add the IDP value as an argument\n");
		return -1;
	}
	totalLength += arglen;
	
	arglen = addXRPArg(msg, XRP_CLASS_SRC_CXT, ctxt->value, ctxt->valueLen);
	if (arglen<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentUnpublish : Unable to add context as an argument\n");
		return -1;
	}
	totalLength += arglen;

	arglen = addXRPArg(msg, XRP_CLASS_SRC_SRV, srv->value, srv->valueLen);
	if (arglen<0) {
		anaPrint(ANA_ERR, "anaL1_encCompartmentUnpublish : Unable to add service as an argument\n");
		return -1;
	}
	
	return (totalLength + arglen);
}



/*decodes a Lookup reply by, */
int anaL1_decResponse(xrpMsg_t replyMsg, xrpClass_t className, int occurence, void** ptPtval){
	
	int lengthContent;
	void *ptContent;
	
	if (!replyMsg || !className || !ptPtval || occurence<0 )
		return -1;
	
	ptContent = (void *)getXRPArg(replyMsg, className, &lengthContent, occurence);
	if (ptContent==NULL)
		return -1;
	
	*ptPtval = ptContent;
	return lengthContent;
}

int anaL1_isError(xrpMsg_t replyMsg){  /* returns 1 if ERROR command, 0 otherwise */
	
	if (replyMsg == NULL || equalXRPCommands(replyMsg, XRP_CMD_ERROR ))
		return 1;
	return 0;
}














