/* C/bricks/regexp/using_regexp.c
 *   
 * user of yellow pages 
 * depends on: regexp
 *  
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
#include <linux/module.h>
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Christophe Jelger");
//static char *cd = "kern://using_regexp_data"; //"udp://127.0.0.1:5559"; /* brick data url*/
//static char *cc = "kern://using_regexp_ctrl"; //"udp://127.0.0.1:8889"; /* brick control url*/
//static char *nc = "kern://anaMinmex"; //"udp://127.0.0.1:6666"; /* minmex control url*/
//static char * debug  = "ANA_DEBUG";
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

#define MODULE_NAME using_regexp
char mymodename[]="using_regexp";

#include "brick_template.h"

/*-------------------------------------------------------------------------------------*/

#include "anaLib2.h"

#define IPv6_REGEXP "^([[:xdigit:]]{1,4}:){7}[[:xdigit:]]{1,4}$"
#define ETH_REGEXP "^([[:xdigit:]]{2}:){5}[[:xdigit:]]{2}$"

anaLabel_t cfinder_cpe;

void brick_exit()
{
	anaPrint(ANA_DEBUG, "Brick EXIT\n");
}

void dummy(struct anaL2_message *msg)
{
	/* dummy function for callback */

	return;
}

anaLabel_t lookup(char *str)
{
	struct context_s ctx;
	struct service_s srv;
	anaLabel_t tmp;
	
	if (!str)
		return 0;
		
	ctx.value = ".";
	ctx.valueLen = 2;
	srv.value = str;
	srv.valueLen = strlen(srv.value)+1;

	tmp = anaL2_resolve(cfinder_cpe, &ctx, &srv, 'u', NULL, NULL);
	
	anaPrint(ANA_DEBUG, "%s --> %x\n", str, tmp);
	
	return tmp;
}

int brick_start()
{
	struct context_s ctx;
	struct service_s srv;
	anaLabel_t pub1, pub2, lkp;
	
	ctx.value = ".";
	ctx.valueLen = 2;
	srv.value = "cfinder";
	srv.valueLen = strlen(srv.value)+1;

	anaL2_initDefault();
	
	cfinder_cpe = anaL2_resolve(NODE_LABEL, &ctx, &srv, 'u', NULL, NULL);
	
	if (!cfinder_cpe)
	{
		anaPrint(ANA_ERR, "cannot find cfinder brick in KVR\n");
		anaL0_stopANA(-1);
		/* exit here */
	}
	
	srv.value = IPv6_REGEXP;
	srv.valueLen = strlen(srv.value)+1;
	pub1 = anaL2_publish(cfinder_cpe, &ctx, &srv, dummy, NOTHREAD, NULL);
	if (!pub1)
		anaPrint(ANA_ERR, "cannot publish regexp\n");
				
	sleep(5);
	
	anaL2_unpublish(cfinder_cpe, pub1, &ctx, &srv, NULL);
	sleep(5);
	
	pub1 = anaL2_publish(cfinder_cpe, &ctx, &srv, dummy, NOTHREAD, NULL);
	
	srv.value = ETH_REGEXP;
	srv.valueLen = strlen(srv.value)+1;
	pub2 = anaL2_publish(cfinder_cpe, &ctx, &srv, dummy, NOTHREAD, NULL);
	if (!pub2)
		anaPrint(ANA_ERR, "cannot publish regexp\n");
	
	lkp = lookup("whatever");
	
	lkp = lookup("00:11:22:33:44:55");
	
	if (lkp)
	{
		struct context_s eth_ctx;
		struct service_s dummy;
	
		eth_ctx.value = "00:11:22:33:44:55";
		eth_ctx.valueLen = strlen(eth_ctx.value)+1;
		dummy.value = "dummy";
		dummy.valueLen = strlen(dummy.value)+1;
	
		lkp = anaL2_resolve(lkp, &eth_ctx, &dummy, 'u', NULL, NULL);
	}
	
	lkp = lookup("00:11:22:33:44");
	
	lkp = lookup("0123:4567:789a:bcde:f123:4567:89ab:cdef");
	
	lkp = lookup("192.168.42.1");
	
	srv.value = ETH_REGEXP;
	srv.valueLen = strlen(srv.value)+1;
	anaL2_unpublish(cfinder_cpe, pub2, &ctx, &srv, NULL);
		
	srv.value = IPv6_REGEXP;
	srv.valueLen = strlen(srv.value)+1;
	anaL2_unpublish(cfinder_cpe, pub1, &ctx, &srv, NULL);
	
	return 1;
}

