/* C/bricks/chat/chat.c
 * 
 * ANA chat application
 * command line arguments: -a USERNAME -a PROTOCOL
 *         where USERNAME = user name
 *               PROTOCOL = protocol over which chat runs (e.g. "IP").
 * send messages with: @DestUserName Hello world!
 * 
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   	
 * Copyright (c) 2008, Ariane Keller, Swiss federal institute of technology (ETH Zurich) 
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
/* !!! Change the two variables below to the name of your brick !!! */
#include "brick_template.h"

char mymodename[]="chat";

/*-------------------------------------------------------------------------------------*/

#include <ncurses.h>
#include "client-tui.h"

#define MAX_USERNAME_LEN 20
char USERNAME[MAX_USERNAME_LEN];

struct user {
	char * username;
	anaLabel_t userIDP;
	struct user * next;
};
struct user * head = NULL;

anaLabel_t recvDataIDPUser = 0;
anaLabel_t ethLabel = 0;
anaLabel_t peerLabel = 0;
const int MAX_BUFFER = 100;

const char * EXIT = "EXIT*%"; /* magic string to distinguish between user messages 
			   * and the exit message which leads to a 
			   * notification that the user has disappeared
			   */

/* receive chat messages (from network) */
static void recvData(struct anaL2_message *msg ) {
	if (msg!=NULL) {
		/* a user disappears */
		if (!memcmp(msg->data, EXIT, strlen(EXIT))) {
			struct user * tmp = head;
			struct user * old = tmp;
			/* display notification */
			int len = strlen("user disappeared: ");
			int buflen = len + msg->senderService->valueLen;
			char * buff = malloc(buflen + 1);
			memcpy(buff, "user disappeared: ", len);
			memcpy(buff + len, msg->senderService->value, msg->senderService->valueLen);
			buff[buflen] = '\n';
			post_message(buff);
			free(buff);

			/* loop through all registered users */
			while (tmp != NULL) {
				if (!(memcmp(tmp->username, msg->senderService->value, strlen(msg->senderService->value) + 1))) {
					if (tmp == head)
						head = tmp->next;
					else {
						old->next = tmp->next;
						free(tmp->username);
					}
				}
				if (tmp != head);
					old = old->next;
				tmp = tmp->next;

			}	
		}
		/* its a chat message */
		else {
			post_bold_message(msg->data);
		}
	}
	/* this should not happen...*/
	else
		post_message("no message?");
}

/*function which waits for user input*/
static void *sendfct(void * unused){
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	int ch;
	/* in front of each chat message we place "USERNAME: "*/
	char buffer[MAX_BUFFER + MAX_USERNAME_LEN];
	char sendbuffer[MAX_BUFFER + MAX_USERNAME_LEN]; /* we need not to send the destination user name */
	memset(sendbuffer, 0, MAX_BUFFER + MAX_USERNAME_LEN);
	memset(buffer, 0, MAX_BUFFER + MAX_USERNAME_LEN);
	memcpy(buffer, USERNAME, strlen(USERNAME));
	memcpy(buffer + strlen(USERNAME), ": ", 2);
	int index= strlen(USERNAME) + 2;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&timeout, 0, sizeof(struct timespec));
	memset(&description, 0, sizeof(struct service_s));
	timeout.tv_sec = 1; 
	anaValidate("buffer: %s\n", buffer);
	/* wait for user input */
	while ((ch = getch()) != KEY_F(9)) {
        switch (ch) {
        case ERR:{
            // No input available right now. sleep for 1ms
	    usleep(1000);
            break;
	}

	/* delete last character */
	case KEY_BACKSPACE: {
		if(index > 0){
			index--;
			buffer[index] = ' ';
			del_char();
		}
		break;
	}

	/* send the message */
        case '\n':{
		/* message format in buffer: "USERNAME: @destUserName chat message" */
		if(!(memcmp(buffer + strlen(USERNAME) + 2, "@", 1))){
			struct user * tmp = head;
			char * destUser = buffer + strlen(USERNAME) + 3;
			int destUserLen = 0;
			while (memcmp(destUser + destUserLen, " ", 1) && destUserLen < MAX_BUFFER + MAX_USERNAME_LEN)
				destUserLen++;
			memcpy(sendbuffer, buffer, strlen(USERNAME) + 1);
			memcpy(sendbuffer + strlen(USERNAME) + 1, destUser + destUserLen, strlen(destUser + destUserLen));
			/* we loop through all registered peer useres */
			while (tmp != NULL) {
				if(!memcmp(destUser, tmp->username, strlen(tmp->username))){
					/* we have the user in the list */
					anaL0_send(tmp->userIDP, sendbuffer, strlen(sendbuffer) + 1);
					post_message(buffer + strlen(USERNAME) + 2 );
					break;
				}
				tmp = tmp->next;
			}

			/* we did not find the user, do a resolve*/
			if (tmp == NULL) { 
				int destUserLen = 0;
				while (memcmp(destUser + destUserLen, " ", 1) && destUserLen < MAX_BUFFER + MAX_USERNAME_LEN)
					destUserLen++;
				char * destUserName = malloc(destUserLen + 1);
				memcpy(destUserName, destUser, destUserLen);
				destUserName[destUserLen] = '\0';
				context.value = "*";
				context.valueLen = 2;
				service.value = destUserName;
				service.valueLen = destUserLen + 1;
				description.value = USERNAME;
				description.valueLen = strlen(USERNAME) + 1;
				anaLabel_t peerLabel = anaL2_resolve(ethLabel, &context, &service, 'u', &description, &timeout);
				if (peerLabel == 0) {
					char * tmp = malloc(30 + destUserLen);
					sprintf(tmp, "User %s cannot be reached!", destUserName);
					post_message(tmp);
					free(tmp);		
					free(destUserName);	
				}
				/* idp received */
				else {
					struct user * tmp = malloc(sizeof(struct user));
					tmp->username = destUserName;
					tmp->userIDP = peerLabel;
					tmp->next = NULL;
					/* add to the list of known user names */
					if (head == NULL)
						head = tmp;				
					else {
						tmp->next = head;
						head = tmp;
					}
					anaL0_send(peerLabel, sendbuffer, strlen(sendbuffer) + 1);
					post_message(buffer + strlen(USERNAME) + 2);
				}
			}
		}
		else
			post_message("invalid message format! Did you specify a user name?");
	
		/* reset buffer */
		index = strlen(USERNAME) + 2;
		memset(buffer + strlen(USERNAME) + 2, 0, MAX_BUFFER);
		memset(sendbuffer, 0, MAX_BUFFER + MAX_USERNAME_LEN);
	        reset_input_win();
            	refresh_tui();
		break;
	} 
	
	/* add the character to the buffer */
	default:{
		if (index<MAX_BUFFER) {
                	buffer[index++] = ch;
            		echo_character(ch);
		}
		else
			post_message("max buffer length exceeded!");
            break;
        }
	} /*switch*/
	} /*while*/
	
	return NULL;
}


void brick_exit() {
	anaPrint(ANA_DEBUG, "Brick EXIT\n");
	struct context_s context;
	struct service_s service;
	struct timespec timeout;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	close_tui();
	
	/* unpbulish */
	
	context.value = "*";
	context.valueLen = 2;
	service.value = USERNAME;
	service.valueLen = strlen((char*)USERNAME)+1;

	anaL2_unpublish(ethLabel, recvDataIDPUser, &context, &service, &timeout);

	anaL0_unregisterCallback(recvDataIDPUser, 0);

	/* send a notifiaction that one is about to leave */
	struct user * tmp = head;
	while (tmp != NULL) {
		anaL0_send(tmp->userIDP, (char *)EXIT, strlen(EXIT)+1);
		tmp = tmp->next;
	}

	/* free all peers */
	if (head != NULL) {
		tmp = head->next;
		while (tmp != NULL) {
			head->next = tmp->next;
			free(tmp->username);
			free(tmp);
			tmp = head->next;
		}
		free(head->username);
		free(head);
	}

	anatimer_exit();
}

int brick_start() {
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&timeout, 0, sizeof(struct timespec));
	memset(&description, 0, sizeof(struct service_s));
	timeout.tv_sec = 1; 

	if(anaDebugLevel != ANA_VALIDATE)
		anaDebugLevel = ANA_EMERG;
	
	if (validate_init("agnostic_chat") < 0)
               return -1;

	anaL2_initDefault();
	init_tui();

	char * user = getAuxArg(0);
	if(user != NULL)
		strncpy(USERNAME, user, MAX_USERNAME_LEN);
	else
		strncpy(USERNAME, "ANA_CHAT_USER", MAX_USERNAME_LEN);

	USERNAME[MAX_USERNAME_LEN - 1] = '\0';
	char * protocol = getAuxArg(1);
	if (protocol == NULL)
		protocol = "eth01";
	char welcome[100];
	sprintf(welcome, "Hi %s!", USERNAME);
	post_bold_message(welcome);
	sprintf(welcome, "You are using ANA-Chat over %s.", protocol);
	post_bold_message(welcome);
	post_bold_message(" ");

	/* look for the network label, currently eth01 */
	context.value = ".";
	context.valueLen = 2;
	service.value = protocol;
	service.valueLen = strlen(protocol) + 1;

	ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	if(ethLabel == 0){
		post_message("Discover of network brick failed\n");
		close_tui();	
		return 1;
	}

	/* publish our user name in the network compartment */
	context.value = "*";
	context.valueLen = 2;
	service.value = USERNAME;
	service.valueLen = strlen(USERNAME) + 1;

	recvDataIDPUser = anaL2_publish(ethLabel, &context, &service, recvData, THREAD, &timeout);
	if( recvDataIDPUser == 0){
		post_message("Publish inside network compartment failed\n");
		close_tui();
		return 1;
	}
	anaValidate("chat initialization ok\n");

	/* start the thread which is waiting for the user input */
	anaThread(sendfct, NULL, NULL);


	return 1;
}


