 /* C/bricks/ip/ip_chat.c
 *  
 * chat application for ip compartment
 * depends on: ip_enc
 * runs in user space
 * KVR entries: none
 * 
 * use:
 *
 * @{ip address} {message}; in order to send to a specific receiver
 *
 * {message}; in order to send to all receivers
 *
 * @list; in order to display a list of receivers
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */
 
/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[]="chat";

#include "brick_template.h"

/*-------------------------------------------------------------------------------------*/

#include <ncurses.h>
#include "client-tui.h"

const char *EXIT = "EXIT*%"; /* magic string */

struct user {
	char *address;
	anaLabel_t userIDP;
	struct user *next;
};

struct user *head = NULL;

anaLabel_t recvDataIDP = 0;
anaLabel_t ipLabel = 0;

const int MAX_BUFFER = 100;

/* receive data function */
static void recvData(struct anaL2_message *data) {
	int length;
	char *message, *buffer, *saddr;
	anaLabel_t targetIDP;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

       	if (data->data==NULL || data->dataLen<0)
       		return;
	
	length = data->dataLen;
	saddr = (char *) data->senderContext;
	message = (char *) data->data;
	
	if (!memcmp(message, EXIT, strlen(EXIT))) { /* it's an EXIT message */
		struct user *tmp = head;
		struct user *old = tmp;
		while (tmp!=NULL) {
			if (!memcmp(tmp->address, saddr, strlen(saddr) + 1)) {
				if (tmp==head)
					head = tmp->next;
				else {
					old->next = tmp->next;
					free(tmp->address);
				}
			}
			if (tmp!=head);
				old = old->next;
			tmp = tmp->next;
		}
		
		buffer = (char *) malloc(50);
		memcpy(buffer, "--> ", 4);
		memcpy(buffer + 4, saddr, strlen(saddr));
		memcpy(buffer + 4 + strlen(saddr), " has left the chat!", 20);
		post_message(buffer);
		free(buffer);
	}
	else {
		if (memcmp(saddr, "localhost", strlen(saddr) + 1)) {
			/* we loop through all registered addresses */
			struct user *tmp = head;
			while (tmp!=NULL) {
				if (!memcmp(saddr, tmp->address, strlen(tmp->address) + 1)) {
					/* we have the user in the list */
					break;
				}
				tmp = tmp->next;
			}
			if (tmp==NULL) {
				/* not in list, resolve and add */
				context.value = saddr;
				context.valueLen = strlen(saddr) + 1;
				service.value = "chat";
				service.valueLen = strlen("chat")+1;

				targetIDP = anaL2_resolve(ipLabel, &context, &service, 'u', &service, &timeout);
				if (targetIDP != 0) {
					struct user *tmp = malloc(sizeof(struct user));
					tmp->address = malloc(strlen(saddr) + 1);
					memcpy(tmp->address, saddr, strlen(saddr) + 1);
					tmp->userIDP = targetIDP;
					tmp->next = NULL;
					if (head==NULL)
						head = tmp;				
					else {
						tmp->next = head;
						head = tmp;
					}
				}
			}
		}
		
		/* show message */
		buffer = (char *) malloc(length + 20);
		memcpy(buffer, "(", 1);
		memcpy(buffer + 1, saddr, strlen(saddr));
		memcpy(buffer + 1 + strlen(saddr), ")\t: ", 4);
		memcpy(buffer + 1 + strlen(saddr) + 4, message, strlen(message) + 1);
		post_message(buffer);
		free(buffer);
	}		
}

/* function which waits for user input */
static void * sendfct(void *unused) {
	xrpMsg_t msg;
	int index = 0, i = 0, addressLen, j =0, null = 0;
	anaLabel_t targetIDP;
	char *buffer, *buffer2, *address;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

	buffer = (char *) malloc(MAX_BUFFER + 20);
	memset(buffer, 0, MAX_BUFFER + 20);
	
	int ch;
	
	while ((ch = getch())!=KEY_F(9)) {
		switch (ch) {
			
			case ERR: {
				// No input available right now. sleep for 1ms
				usleep(1000);
			break;
			}
		
			case KEY_BACKSPACE: {
				if(index > 0){
					index--;
					buffer[index] = ' ';
				}
				del_char();
				break;
				
			}
	
			case '\n':	{
				if (index==0)
					break;
				if (!memcmp(buffer, "@", 1)) {
					while (memcmp(buffer + i, " ", 1) && memcmp(buffer + i, &null, 1)) {
						i++;
					}
					addressLen = i - 1;
					i = 0;
					
					if (!memcmp(buffer + 1, "list", 4)) {
						buffer2 = (char *) malloc(30);
						memcpy(buffer2, "--> list of known users:", 25);
						post_message(buffer2);
						free(buffer2);
						struct user *tmp = head;
						while (tmp!=NULL) {
							buffer2 = (char *) malloc(20);
							memcpy(buffer2, "@", 1);
							memcpy(buffer2 + 1, tmp->address, strlen(tmp->address) + 1);
							post_message(buffer2);
							free(buffer2);
							tmp = tmp->next;
						}
						buffer2 = (char *) malloc(30);
						memcpy(buffer2, "--> end of list!", 17);
						post_message(buffer2);
						free(buffer2);
						/* reset buffer */
						index = 0;
						memset(buffer, 0, MAX_BUFFER + 20);
						reset_input_win();
						refresh_tui();
						break;
					}
					
					if ((addressLen + 1)==strlen(buffer) || (addressLen + 2)==strlen(buffer)) {
						buffer2 = (char *) malloc(30);
						memcpy(buffer2, "--> no message given!", 22);
						post_message(buffer2);
						free(buffer2);
						/* reset buffer */
						index = 0;
						memset(buffer, 0, MAX_BUFFER + 20);
						reset_input_win();
						refresh_tui();
						break;
					}
					
					address = (char *) malloc(20);
					memset(address, 0, 20);
					memcpy(address, buffer + 1, addressLen);
					
					/* we loop through all registered addresses */
					struct user *tmp = head;
					while (tmp!=NULL) {
						if (!memcmp(buffer + 1, tmp->address, strlen(tmp->address))) {
							/* we have the user in the list */
							anaL0_send(tmp->userIDP, buffer + addressLen + 2, strlen(buffer) + 1 - addressLen - 2);
							post_message(buffer);
							break;
						}
						tmp = tmp->next;
					}
					if (tmp==NULL) {
						/* not in list, resolve */
						context.value = address;
						context.valueLen = strlen(address) + 1;
						service.value = "chat";
						service.valueLen = strlen("chat")+1;

						targetIDP = anaL2_resolve(ipLabel, &context, &service, 'u', &service, &timeout);
						free(msg);
						if (targetIDP == 0) {
							buffer2 = (char *) malloc(100);
							memcpy(buffer2, "--> ", 4);
							memcpy(buffer2 + 4, buffer + 1, addressLen);
							memcpy(buffer2 + 4 + addressLen, " is not reachable! (not online or no default gw specified)", strlen(" is not reachable! (not online or no default gw specified)") + 1);
							post_message(buffer2);
							free(buffer2);
						}
						else {
							/* add to the list */
							struct user *tmp = malloc(sizeof(struct user));
							tmp->address = malloc(strlen(address) + 1);
							memcpy(tmp->address, address, strlen(address) + 1);
							tmp->userIDP = targetIDP;
							tmp->next = NULL;
							if (head==NULL)
								head = tmp;
							else {
								tmp->next = head;
								head = tmp;
							}
							/* we send our message */
							anaL0_send(targetIDP, buffer + addressLen + 2, strlen(buffer) + 1 - addressLen - 2);
							post_message(buffer);
						}
					}
				}
				else {
					/* we loop through all registered usernames */
					struct user *tmp = head;
					j = 0;
					while (tmp!=NULL) {
						anaL0_send(tmp->userIDP, buffer, strlen(buffer) + 1);
						j++;
						tmp = tmp->next;
					}
					if (j==0) {
						buffer2 = (char *) malloc(30);
						memcpy(buffer2, "--> connected to nobody!", 25);
						post_message(buffer2);
						free(buffer2);
					}
					else
						post_message(buffer);
				}
				
				/* reset buffer */
				index = 0;
				memset(buffer, 0, MAX_BUFFER + 20);
				reset_input_win();
				refresh_tui();
				break;
			}
	
			default: {
				/* add the character to the buffer */
				if (index<MAX_BUFFER)
					buffer[index++] = ch;
				echo_character(ch);
				break;
			}
		}
	}
	return NULL;
}

void brick_exit() {
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	struct timespec timeout;
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

	/* send EXIT to all in list */
	struct user *tmp = head;
	while (tmp!=NULL) {
		anaL0_send(tmp->userIDP, (char *) EXIT, strlen(EXIT) + 1);
		tmp = tmp->next;
	}
	
	anaPrint(ANA_NOTICE, "CHAT EXIT\n");
	close_tui();
	context.value = "*";
	context.valueLen = 2;
	service.value = "chat";
	service.valueLen = strlen("chat")+1;

	anaL2_unpublish(ipLabel, recvDataIDP, &context, &service, &timeout);

 	anaL0_unregisterCallback(recvDataIDP, 0);	

	anatimer_exit();
}

int brick_start() {
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

	anaDebugLevel = ANA_ERR;
	anaL2_initDefault();
 	
	/* look for the ip label */
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip";
	service.valueLen = strlen("ip")+1;

	ipLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	if (ipLabel == 0) {
		anaPrint(ANA_ERR, "discover of ip brick failed!\n");	
		return 1;
	}
	context.value = "*";
	context.valueLen = 2;
	service.value = "chat";
	service.valueLen = strlen("chat")+1;

	recvDataIDP = anaL2_publish(ipLabel, &context, &service, recvData, THREAD, &timeout);
	if (recvDataIDP == 0) {
		anaPrint(ANA_ERR, "publish failed!\n");
		return 1;
	}

	init_tui();
	
	/* start the thread which is waiting for the user input */
	anaThread(sendfct, NULL, NULL);
	
	return 1;
}
