#ifndef __ETH_BRICK_
#define __ETH_BRICK_

#ifdef USER
#include <net/ethernet.h>
#endif
#ifdef KERNEL
/*userspace and kernel space have different names for the ethernet header structure...*/
#include <linux/if_ether.h>
struct ether_header{
	unsigned char   ether_dhost[ETH_ALEN];       /* destination eth addr */
        unsigned char   ether_shost[ETH_ALEN];     /* source ether addr    */
        uint16_t          ether_type;                /* packet type ID field */
} __attribute__((packed));


#endif

#include "quickRepository.h"
#include "anaCommon.h"

/* some macros for using ETH MAC addresses */
#define ETH_HEX(addr) \
	((unsigned char *)addr)[0], \
	((unsigned char *)addr)[1], \
	((unsigned char *)addr)[2], \
	((unsigned char *)addr)[3], \
	((unsigned char *)addr)[4], \
	((unsigned char *)addr)[5]

#define ETH_FMT "%02x:%02x:%02x:%02x:%02x:%02x"

/*ether_type fields for the ethernet header */
#define ETH_CMD 0x0889
#define ETH_DLV 0x0890
#define ETH_RES 0x0891 /*this is for resolve request */
#define ETH_RSP 0x0892 /*this is for resolve reply */

/*default size of resolve request identifier*/
#define RES_ID_SIZE 4

#define RESOLVE_TIMEOUT 5000

/* maximum transmission unit. ethBrick packet is payload for vlink*/
#define ETH_MTU 1500 - sizeof(struct ether_header)

/* regular expression for compartment finder (added by Christophe) */
#define ETH_REGEXP "^([[:xdigit:]]{2}:){5}[[:xdigit:]]{2}$"

unsigned char  broadcastAddress[ETH_ALEN] ;

struct fullEthHeader{
	struct ether_header *ethhdr;	/*ethernet header*/
	uint16_t  opCode;
	void *destDescrip;		/*description of the destination ethernet user*/
	int destDescLen;	
	void *srcDescrip;		/*description of the source ethernet user*/
	int srcDescLen;
};

struct resolveRequestList{
	char resId[RES_ID_SIZE];
	anaLabel_t replyTo;
	void *context;
	int contextLen;
	int timerId;
	void *target;
	int targetLen;
	char chanType;
	void *description;
	int descLen;
	struct resolveRequestList *next;
};


/*defualt values for the repository*/
#define QREP_KEYTAB_DEFAULTMASK 10
#define QREP_ENTRYTAB_DEFAULTMASK 10
#endif
