 /* C/bricks/fbr/fbr_clie.c
 *  
 * sample client application
 * depends on: field-based publish / subscribe compartment
 * runs in user space
 * KVR entries: fbr_clie
 * 
 * use:
 *
 * fbr_serv -a "{service-type}"
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "anaLib2.h"
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "fbr_clie";


/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "fbr.h"


anaLabel_t recvIDP;
anaLabel_t forwLabel;

/* called upon leaving the fbr_clie */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

        anaPrint(ANA_NOTICE, "FBR_CLIE HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex!= NULL\n");
	msg = anaL1_allocateMessage();
	
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_clie";
	service.valueLen = strlen("fbr_clie")+1;

	len = anaL1_encCompartmentUnpublish(msg, recvIDP, &context, &service);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request in KVR!\n");
	}
	free(msg);
	
	anaL0_unregisterCallback(recvIDP, 0);
}

/* receive data function */
void recvData_cli(void *data, int len, anaLabel_t input, void *aux) {
	int messageLen, idLen, totalLength, arglen;
	char *message, *id;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	anaPrint(ANA_NOTICE, "\n\n--> message received\n\n");
	
	message =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGE, &messageLen, 0);
	id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGEID, &idLen, 0);
	
	anaPrint(ANA_NOTICE, "message-id: %s\n", id);
	anaPrint(ANA_NOTICE, "message: %s\n", message);
	
	/* send some data */
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_DATA);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "recvData: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_MESSAGEID, id, strlen(id) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "recvData: unable to add message-id to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_MESSAGE, "Hi! Now, I know you!", strlen("Hi! Now, I know you!") + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "recvData: unable to add message to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	if (forwLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_forw";
		service.valueLen = strlen("fbr_forw")+1;

		forwLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (forwLabel == 0) {
			anaPrint(ANA_EMERG, "forwarding brick not found!\n");
			free(msg);
			return;
		}
	}
	anaL0_send(forwLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "message sent to forwarding brick\n");
	free(msg);
}

void handleLookupReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name;
	int labelSize, totalLength, arglen;
	char *target;
	anaLabel_t *temp;
	xrpMsg_t msg;

	anaPrint(ANA_NOTICE, "received lookup reply for forwarding brick in the KVR\n");
	
	if (anaL1_isError(data)) { /* if ERROR command */
		anaPrint(ANA_ERR, "is error\n");
		return;
	}
	
	if (anaL1_decResponse(data, XRP_CLASS_NAME, 0, (void**) &name)>0) {
		labelSize =  anaL1_decResponse(data, XRP_CLASS_LABEL, 0, (void**) &temp);
		if (name!=NULL && temp != 0) {
			forwLabel = *temp;
			target ="gateway";
			if (getAuxArg(0)!=NULL)
				target = getAuxArg(0);
			
			/* send subscribe message */
			
			msg = anaL1_allocateMessage();
			totalLength = fillXRPCommand(msg, XRP_CMD_SUBSCRIBE);
			if (totalLength<0) {
				anaPrint(ANA_EMERG, "handleLookupReply: unable to fill command in message!\n");
				free(msg);
				return;
			}  
			arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, target, strlen(target) + 1);
			if (arglen < 0) {
				anaPrint(ANA_EMERG, "handleLookupReply: unable to add service-type to message!\n");
				free(msg);
				return;
			}
			totalLength += arglen;
			arglen = addXRPArg(msg, XRP_CLASS_REQUESTER, "fbr_clie", strlen("fbr_clie") + 1);
			if (arglen < 0) {
				anaPrint(ANA_EMERG, "handleLookupReply: unable to add requester to message!\n");
				free(msg);
				return;
			}
			totalLength += arglen;
			anaL0_send(forwLabel, msg, totalLength);
			
			free(msg);
			anaPrint(ANA_NOTICE, "subscribe message sent: %s\n", target);
		}
	}
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	struct service_s description;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));

	/* register data receive callback function */
	recvIDP = anaL0_registerCallback((anaCallback_t) & recvData_cli, NULL, NULL, NULL , 0, IDP_PERM, THREAD);
	if (recvIDP < 0) {
		return 1;
	}
	anaPrint(ANA_NOTICE,  "registered recvIDP\n");
	
	/* publish in the KVR */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_clie";
	service.valueLen = strlen("fbr_clie")+1;

	lenMsg = anaL1_encCompartmentPublish(msg, recvIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_ERR, "problem encoding publish request for recvIDP in KVR!\n");
		free(msg);
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_ERR, "problem sending publish request for recvIDP in KVR!\n");
		free(msg);
		return -1;
	}
	free(msg);
	
	/* lookup the forwarding brick by name */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_forw";
	service.valueLen = strlen("fbr_forw")+1;

	lenMsg = anaL1_encResolve(msg, &context, &service, 'u', &description);
	if (lenMsg<0) {
		free(msg);
		return -1;
	}
	ret = anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handleLookupReply, NULL, NULL, NULL, 1); 
	if (ret<0) {
		free(msg);
		return -1;
	}
	anaPrint(ANA_NOTICE, "lookup query sent\n");
	free(msg);
	return 1;
}

