/* C/bricks/fbr/fbr_forw.c
 *  
 * forwards client messages in field-based publish / subscribe compartment
 * depends on: dgrmEthBrick
 * runs in user space
 * KVR entries: fbr_forw
 * 
 * IMPORTANT: use minmex with -b option to extend default message size
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "anaLib2.h"
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "fbr_forw";


/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "fbr.h"

anaLabel_t recvNodeIDP;
anaLabel_t recvEthIDP;
anaLabel_t ethLabel;
anaLabel_t ftabLabel;

struct mids *mid;
struct mids *mid_start;
struct mids *mid_test;

void recvNodeData(char *data, int len, anaLabel_t input, void *aux);
void recvEthData(char *data, int len, anaLabel_t input, void *aux);

int garbage = GARBAGE_LIFES;

/* called upon leaving fbr_forw */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

        anaPrint(ANA_NOTICE, "FBR_FORW HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex! = NULL\n");
	
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_forw";
	service.valueLen = strlen("fbr_forw")+1;

	len = anaL1_encCompartmentUnpublish(msg, recvNodeIDP, &context, &service);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request in KVR!\n");
	}
	free(msg);
	msg = anaL1_allocateMessage();
	len = anaL1_encCompartmentUnpublish(msg, recvEthIDP, &context, &service);
	ret = anaL0_send(ethLabel, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request in ethernet brick!\n");
	}
	free(msg);
	
	anatimer_exit();
	
	anaL0_unregisterCallback(recvNodeIDP, 0);
	anaL0_unregisterCallback(recvEthIDP, 0);
	
}

void dataMessage(void *data, int len) {
	int idLen, found = 0;
	char *id;
	anaLabel_t label;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGEID, &idLen, 0);
	anaPrint(ANA_NOTICE, "message-id: %s\n", id);
	
	while (mid_test->type!=NULL) {
		if (strcmp(mid_test->id, id)==0) {
			found = 1;
			if (mid_test->sown==0) {
				anaL0_send(mid_test->slabel, data, len);
				anaPrint(ANA_NOTICE, "message successfully forwarded to next node\n");
			} else {
				context.value = ".";
				context.valueLen = 2;
				service.value = mid_test->type;
				service.valueLen = strlen(mid_test->type) + 1;
				label = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
				if (label == 0) {
					anaPrint(ANA_EMERG, "service-type not found in local node!\n");
					return;
				}
				anaL0_send(label, data, len);
				anaPrint(ANA_NOTICE, "message successfully delivered local to %s\n", mid_test->type);
			}
		}
		mid_test = mid_test->next;
	}
	mid_test = mid_start;
	
	if (!found)
		anaPrint(ANA_EMERG, "unable to forward message!\n");
}

void localResponse(void *data, int len) {
	int idLen, found = 0, dummy = 0;
	char *id;
	anaLabel_t label;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	
	timeout.tv_sec = 1; 

	id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGEID, &idLen, 0);
	anaPrint(ANA_NOTICE, "message-id: %s\n", id);
	
	while (mid_test->type!=NULL) {
		if (strcmp(mid_test->id, id)==0) {
			found = 1;
			dummy = 1;
			memcpy(&mid_test->sown, &dummy, sizeof(int));
			if (mid_test->cown==0) {
				anaL0_send(mid_test->clabel, data, len);
				anaPrint(ANA_NOTICE, "message successfully forwarded to next node\n");
			} else {
				context.value = ".";
				context.valueLen = 2;
				service.value = mid_test->requester;
				service.valueLen = strlen(mid_test->requester) + 1;

				label = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
				if (label == 0) {
					anaPrint(ANA_EMERG, "service-type not found in local node!\n");
					return;
				}
				anaL0_send(label, data, len);
				anaPrint(ANA_NOTICE, "message successfully delivered local to %s\n", mid_test->type);
			}
		}
		mid_test = mid_test->next;
	}
	mid_test = mid_start;
	
	if (!found)
		anaPrint(ANA_EMERG, "unable to forward message!\n");
}

void ethResponse(void *data, int len, char *sender, int senderlen) {
	int idLen, found = 0, j;
	char *id, *eth, *jch;
	anaLabel_t label = 0, tmpLabel = 0, IDP = 0;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGEID, &idLen, 0);
	anaPrint(ANA_NOTICE, "message-id: %s\n", id);
	
	for (j=1; j<MAX_INTERFACES; j++) {
		eth = (char *) malloc(20);
		jch = (char *) malloc(10);
		strcat(eth, "eth0");
		sprintf(jch, "%i", j);
		strcat(eth, jch);
		context.value = ".";
		context.valueLen = 2;
		service.value = eth;
		service.valueLen = strlen(eth) + 1;
		tmpLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
		if (tmpLabel != 0) {
			context.value = sender;
			context.valueLen =  senderlen;
			service.value = "fbr_forw";
			service.valueLen = strlen("fbr_forw")+1;
			IDP = anaL2_resolve(tmpLabel, &context, &service, 'u', &description, &timeout);
			if (IDP != 0) {
				break;
			}
		}
		free(eth);
		free(jch);
	}
	
	if (IDP == 0) {
		anaPrint(ANA_EMERG, "ethResponse: sender not found!\n");
		return;
	}
	
	while (mid_test->type!=NULL) {
		if (strcmp(mid_test->id, id)==0) {
			found = 1;
			mid_test->slabel = IDP;
			if (mid_test->cown==0) {
				anaL0_send(mid_test->clabel, data, len);
				anaPrint(ANA_NOTICE, "message successfully forwarded to next node\n");
			} else {
				context.value = ".";
				context.valueLen = 2;
				service.value = mid_test->requester;
				service.valueLen = strlen(mid_test->requester) + 1;

				label = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
				if (label == 0) {
					anaPrint(ANA_EMERG, "service-type not found in local node!\n");
					return;
				}
				anaL0_send(label, data, len);
				anaPrint(ANA_NOTICE, "message successfully delivered local to %s\n", mid_test->type);
			}
		}
		mid_test = mid_test->next;
	}
	mid_test = mid_start;
	
	if (!found)
		anaPrint(ANA_EMERG, "unable to forward message!\n");
}

void localSubscribe(void *data, int len) {
	int typeLen, ttl, totalLength, arglen, IDPLen, reqLen;
	char *type, *id, *ttl_m, *requester;
	xrpMsg_t msg, msg2;
	anaLabel_t IDP;
	void *reply, *test;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	type =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	requester =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_REQUESTER, &reqLen, 0);
	
	id = (char *) malloc(9);
	sprintf(id, "%d", randomsnr(10000000, 99999999));
	ttl = SUBSCRIBE_TTL;
	
	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	anaPrint(ANA_NOTICE, "message-id: %s\n", id);
	
	msg2 = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg2, XRP_CMD_DISCOVER);
	if (totalLength<0) {
		anaPrint(ANA_ERR, "localSubscribe: problem encoding discover message!\n");
		free(msg2);
		return;
	}
	arglen = addXRPArg(msg2, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen<0) {
		anaPrint(ANA_ERR, "localSubscribe: unable to add service-type to discover message!\n");
		free(msg2);
		return;
	}
	totalLength += arglen;
	
	if (ftabLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_ftab";
		service.valueLen = strlen("fbr_ftab")+1;

		ftabLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (ftabLabel == 0) {
			anaPrint(ANA_EMERG, "forwarding table brick not found!\n");
			free(msg2);
			return;
		}
	}
	
	reply = anaL2_requestReply(ftabLabel, msg2, totalLength, &totalLength, &timeout);
	free(msg2);
	
	if (reply == 0) {
		anaPrint(ANA_ERR, "request in ftab failed\n");
		return;
	}

	test = (void *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_MESSAGE, &IDPLen, 0);
	if (strcmp((char *) test, "KVR")!=0) {
		IDP = *(anaLabel_t *)test;
		anaPrint(ANA_DEBUG, "set IDP to %i\n", IDP);
	} else {
		context.value = ".";
		context.valueLen = 2;
		service.value = type;
		service.valueLen = strlen(type)+1;
		IDP = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (IDP == 0) {
			anaPrint(ANA_EMERG, "service-type not found in local node!\n");
			return;
		}
	}
	
	mid->next = (struct mids *) malloc(sizeof(struct mids));
	mid->type = (char*) malloc(strlen(type) + 1);
	memcpy(mid->type, type, strlen(type) + 1);
	mid->id = (char*) malloc(strlen(id) + 1);
	memcpy(mid->id, id, strlen(id) + 1);
	mid->cown = 1;
	mid->sown = 0;
	mid->clabel = 0;
	mid->slabel = 0;
	mid->requester = (char*) malloc(strlen(requester) + 1);
	memcpy(mid->requester, requester, strlen(requester) + 1);
	mid = mid->next;
	mid->type = NULL;
	mid->id = NULL;
	mid->cown = 0;
	mid->sown = 0;
	mid->clabel = 0;
	mid->slabel = 0;
	mid->requester = NULL;
	mid->next = NULL;
	
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_SUBSCRIBE);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "localSubscribe: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "localSubscribe: unable to add service-type to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	arglen = addXRPArg(msg, XRP_CLASS_MESSAGEID, id, strlen(id) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "localSubscribe: unable to add message-id to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	ttl_m = (char *) malloc(5);
	sprintf(ttl_m, "%i", ttl);
	arglen = addXRPArg(msg, XRP_CLASS_TTL, ttl_m, strlen(ttl_m) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "localSubscribe: unable to add ttl to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	anaL0_send(IDP, msg, totalLength);
	anaPrint(ANA_NOTICE, "subscribe message successfully forwarded\n");
	free(msg);
}

void ethSubscribe(void *data, int len, char *sender, int senderLen) {
	int typeLen, idLen, ttlLen, ttl, newttl, totalLength, arglen, IDPLen, j;
	char *type, *id, *ttl_m, *newttl_m, *eth, *jch;
	void *reply, *test;
	anaLabel_t label = 0, IDP = 0, tmpLabel = 0;
	xrpMsg_t msg2;
	xrpSpecs_t *xrpSpec = &defaultXrpSpecs;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 
	
	type =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGEID, &idLen, 0);
	ttl_m =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_TTL, &ttlLen, 0);
	sscanf(ttl_m, "%i", &ttl);
	
	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	anaPrint(ANA_NOTICE, "message-id: %s\n", id);
	
	/* decrease TTL */
	if (ttl<=1) {
		anaPrint(ANA_EMERG, "TTL is over!\n");
		return; /* drop */
	}
	else {
		newttl = ttl - 1;
		newttl_m = (char *) malloc(5);
		sprintf(newttl_m, "%i", newttl);
		memcpy(data + xrpSpec->commandSize + sizeof(uint16_t) + ((xrpSpec->classSize + sizeof(uint16_t)) * 3) + strlen(type) + strlen(id) + 2, newttl_m, strlen(newttl_m) + 1);
		anaPrint(ANA_NOTICE, "new TTL: %i\n", newttl);
	}
	
	for (j=1; j<MAX_INTERFACES; j++) {
		eth = (char *) malloc(20);
		jch = (char *) malloc(10);
		strcat(eth, "eth0");
		sprintf(jch, "%i", j);
		strcat(eth, jch);
		context.value = ".";
		context.valueLen = 2;
		service.value = eth;
		service.valueLen =  strlen(eth) + 1;
		tmpLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (tmpLabel != 0) {
			context.value = sender;
			context.valueLen = senderLen;
			service.value = "fbr_forw";
			service.valueLen = strlen("fbr_forw")+1;

			label = anaL2_resolve(tmpLabel, &context, &service, 'u', &description, &timeout);
			if (label != 0) {
				break;
			}
		}
		free(eth);
		free(jch);
	}
	
	if (label == 0) {
		anaPrint(ANA_EMERG, "eth Subscribe: sender not found!\n");
		return;
	}
	
	mid->next = (struct mids *) malloc(sizeof(struct mids));
	mid->type = (char*) malloc(strlen(type) + 1);
	memcpy(mid->type, type, strlen(type) + 1);
	mid->id = (char*) malloc(strlen(id) + 1);
	memcpy(mid->id, id, strlen(id) + 1);
	mid->cown = 0;
	mid->sown = 0;
	mid->clabel = label;
	mid->slabel = 0;
	mid->requester = NULL;
	mid = mid->next;
	mid->type = NULL;
	mid->id = NULL;
	mid->cown = 0;
	mid->sown = 0;
	mid->clabel = 0;
	mid->slabel = 0;
	mid->requester = NULL;
	mid->next = NULL;
	
	msg2 = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg2, XRP_CMD_DISCOVER);
	if (totalLength<0) {
		anaPrint(ANA_ERR, "localSubscribe: problem encoding discover message!\n");
		free(msg2);
		return;
	}
	arglen = addXRPArg(msg2, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
	if (arglen<0) {
		anaPrint(ANA_ERR, "localSubscribe: unable to add service-type to discover message!\n");
		free(msg2);
		return;
	}
	totalLength += arglen;
	
	if (ftabLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "fbr_ftab";
		service.valueLen = strlen("fbr_ftab")+1;

		ftabLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (ftabLabel == 0) {
			anaPrint(ANA_EMERG, "forwarding table brick not found!\n");
			free(msg2);
			return;
		}
	}
	
	reply = anaL2_requestReply(ftabLabel, msg2, totalLength, &totalLength, &timeout);
	free(msg2);
	
	if (reply == 0)
		return;
	
	test = (void *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_MESSAGE, &IDPLen, 0);
	if (strcmp((char *) test, "KVR")!=0) {
		IDP = *(anaLabel_t *) test;
	} else {
		context.value = ".";
		context.valueLen = 2;
		service.value = type;
		service.valueLen =  strlen(type) + 1;

		IDP = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		if (IDP == 0) {
			anaPrint(ANA_EMERG, "service-type not found in local node!\n");
			return;
		}
	}
	
	anaL0_send(IDP, data, len);
	anaPrint(ANA_NOTICE, "subscribe message successfully forwarded\n");
}

/* receive data from node function */
void recvNodeData(char *data, int len, anaLabel_t input, void *aux) {
	
	/* parse the received xrp message */
	xrpCmd_t command = data;
	
	/* subscribe message received*/
        if (equalXRPCommands(command, XRP_CMD_SUBSCRIBE)) {
		anaPrint(ANA_NOTICE, "\n\n--> received subscribe message from local node\n\n");
		localSubscribe(data, len);
	}
	
	/* response message received */
        if (equalXRPCommands(command, XRP_CMD_RESPONSE)) {
		anaPrint(ANA_NOTICE, "\n\n--> received response from local node\n\n");
		localResponse(data, len);
	}
	
	/* data message received */
        if (equalXRPCommands(command, XRP_CMD_DATA)) {
		anaPrint(ANA_NOTICE, "\n\n--> received data from local node\n\n");
		dataMessage(data, len);
	}
}

/* receive data form ethernet function */
void recvEthData(char *data, int len, anaLabel_t input, void *aux) {
	int lenMsg, senderContextLen, senderTargetLen;
	void *message, *senderContext, *senderTarget;
	xrpCmd_t command; 
	
	lenMsg = anaL1_decResponse(data, XRP_CLASS_MESSAGE, 0, (void**) &message);
	senderContextLen = anaL1_decResponse(data, XRP_CLASS_SRC_CXT, 0, &senderContext);
       	senderTargetLen = anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, &senderTarget);
	
	/* parse the received xrp message */
	command = message;
	
	/* subscribe message received*/
        if (equalXRPCommands(command, XRP_CMD_SUBSCRIBE)) {
		anaPrint(ANA_NOTICE, "\n\n--> received subscribe message from ethernet\n\n");
		ethSubscribe(message, lenMsg, (char *) senderContext, senderContextLen);
	}
	
	/* response message received*/
        if (equalXRPCommands(command, XRP_CMD_RESPONSE)) {
		anaPrint(ANA_NOTICE, "\n\n--> received response from ethernet\n\n");
		ethResponse(message, lenMsg, (char *) senderContext, senderContextLen);
	}
	
	/* data message received */
        if (equalXRPCommands(command, XRP_CMD_DATA)) {
		anaPrint(ANA_NOTICE, "\n\n--> received data from from ethernet\n\n");
		dataMessage(message, lenMsg);
	}
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	
	anaL2_initDefault();
	anatimer_init();
	
	ftabLabel = 0;
	
	/* initializes mids list */
	mid = (struct mids *) malloc(sizeof(struct mids));
	mid_start = mid;
	mid_test = mid;
	mid->type = NULL;
	mid->id = NULL;
	mid->cown = 0;
	mid->sown = 0;
	mid->clabel = 0;
	mid->slabel = 0;
	mid->requester = NULL;
	mid->next = NULL;
	
	/* register data receive from ethernet callback function */
	recvEthIDP = anaL0_registerCallback((anaCallback_t) & recvEthData, NULL, NULL, NULL, 0, IDP_PERM, THREAD);
	if (recvEthIDP < 0) {
		anaPrint(ANA_EMERG,  "brick_start: unable to register recvEthIDP!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "brick_start: recvEthIDP registered\n");
	
	/* publish data IDP in the QREP of eth */
	context.value = ".";
	context.valueLen = 2;
	service.value = "wired";
	service.valueLen = strlen("wired")+1;

	ethLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	if (ethLabel == 0) {
		anaPrint(ANA_EMERG, "ethernet brick not found!\n");
		return -1;
	}
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_forw";
	service.valueLen = strlen("fbr_forw")+1;

	lenMsg = anaL1_encCompartmentPublish(msg, recvEthIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "brick_start: problem encoding publish request for recvEthData!\n");
		return -1;
	}
	ret =  anaL1_requestReply(ethLabel, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_EMERG, "brick_start: problem sending publish request for recvEthData!\n");
		return -1;
	}
	free(msg);

	/* register data receive from node callback function */
	recvNodeIDP = anaL0_registerCallback((anaCallback_t) & recvNodeData, NULL, NULL, NULL, 0, IDP_PERM, THREAD);
	if (recvNodeIDP < 0) {
		anaPrint(ANA_EMERG,  "brick_start: unable to register recvNodeIDP!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "brick_start: recvNodeIDP registered\n");
	
	/* publish ip along with the data IDP in the KVR */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_forw";
	service.valueLen = strlen("fbr_forw")+1;
	lenMsg = anaL1_encCompartmentPublish(msg, recvNodeIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "brick_start: problem encoding publish request for recvNodeIDP!\n");
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_EMERG, "brick_start: problem sending publish request for recvNodeIDP!\n");
		return -1;
	}
	free(msg);
	
	return 1;
}

