/* C/bricks/fbr/fbr_ftab.c
 *  
 * contains the forwarding table
 * runs in user space
 * KVR entries: fbr_ftab
 * 
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "anaLib2.h"
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "fbr_ftab";


/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "fbr.h"

anaLabel_t recvIDP;
struct tab *ft;
struct tab *ft_start;
struct tab *ft_test;

int garbage = GARBAGE_LIFES;

/* called upon leaving the fbr_ftab */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

        anaPrint(ANA_NOTICE, "FBR_FTAB HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex!= NULL\n");
	msg = anaL1_allocateMessage();
	
	/* unpublish the entries in the KVR */
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_ftab";
	service.valueLen = strlen("fbr_ftab")+1;

	len = anaL1_encCompartmentUnpublish(msg, recvIDP, &context, &service);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0){
		anaPrint(ANA_ERR, "problem encoding unpublish request for recvIDP!\n");
	}
	free(msg);
	
	anatimer_exit();
	
	anaL0_unregisterCallback(recvIDP, 0);
}

void showFwd(void) {
	anaLabel_t tmp;
		
	anaPrint(ANA_NOTICE, "+---------------------+---------------+\n");
	anaPrint(ANA_NOTICE, "| Service type\t| Label\t\t|\n");
	anaPrint(ANA_NOTICE, "+---------------------+---------------+\n");
	
	while (ft_test->type!=NULL) {
		if (ft_test->own==1) {
			tmp = NODE_LABEL;
		}
		else {
			tmp = ft_test->label;
		}
		if (strlen((char *) ft_test->type)<14) {
			anaPrint(ANA_NOTICE, "| %s\t\t| %i\t\t|\n", (char *) ft_test->type, tmp);
		}
		else {
			anaPrint(ANA_NOTICE, "| %s\t| %i\t\t|\n", (char *) ft_test->type, tmp);
		}
		anaPrint(ANA_NOTICE, "+---------------------+---------------+\n");
		ft_test = ft_test->next;
	}
	ft_test = ft_start;
}

void handleDiscover(void *data, int len) {
	char *type;
	int typeLen, labelLen, size, found = 0;
	anaLabel_t replyTo;
	xrpMsg_t replyMsg;
	anaLabel_t *replyToptr = NULL;

	type =  (char *) getXRPArg((char *) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	replyToptr =  (anaLabel_t*) getXRPArg((char *) data, XRP_CLASS_LABEL, &labelLen, 0);
	if(replyToptr != NULL)
		replyTo = *replyToptr;	

	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	
	while (ft_test->type!=NULL) {
		anaPrint(ANA_DEBUG, "route to %s\n", ft_test->type);
		if (strcmp(ft_test->type, type)==0) {
			found = 1;
			if (ft_test->own==0) {
				/* send answer */
				replyMsg = anaL1_allocateMessage();
			
				size = anaL1_encData(replyMsg, &ft_test->label, sizeof(anaLabel_t),  NULL, NULL, 0);
				anaL0_send(replyTo, replyMsg, size);
				anaPrint(ANA_NOTICE, "label sent\n");
				free(replyMsg);
				break;
			} else {
				/* send answer */
				replyMsg = anaL1_allocateMessage();
				size = anaL1_encData(replyMsg, (char *) "KVR", 4, NULL, NULL, 0);
				anaL0_send(replyTo, replyMsg, size);
				anaPrint(ANA_NOTICE, "label sent\n");
				free(replyMsg);
				break;
			}
		}
		ft_test = ft_test->next;
	}
	ft_test = ft_start;
	
	if (!found) {
		anaPrint(ANA_EMERG, "no route to %s specified!\n", type);
	}
}

void handleChange(void *data, int len) {
	int typeLen, labelLen, found = 0, dummy, show = 0;
	char *type;
	anaLabel_t *test;
	
	type =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	test =  (void *) getXRPArg((xrpMsg_t) data, XRP_CLASS_LABEL, &labelLen, 0);
	
	while (ft_test->type!=NULL) {
		if (strcmp(ft_test->type, type)==0) {
			found = 1;
			if (strcmp((char *) test, "KVR")==0) {
				if (ft_test->own!=1) {
					dummy = 1;
					memcpy(&ft_test->own, &dummy, 4);
					memcpy(&ft_test->age, &garbage, 4);
					ft_test->label = 0;
					show = 1;
					anaPrint(ANA_NOTICE, "entry successfully updated\n");
					break;
				} else {
					memcpy(&ft_test->age, &garbage, 4);
					anaPrint(ANA_NOTICE, "nothing new\n");
					show = 1;
					break;
				}
			} else {
				if (ft_test->own==0) {
					if (ft_test->label!=*(anaLabel_t *) test) {
						dummy = 0;
						memcpy(&ft_test->own, &dummy, 4);
						memcpy(&ft_test->age, &garbage, 4);
						ft_test->label = *(anaLabel_t *)test;
						show = 1;
						anaPrint(ANA_NOTICE, "entry successfully updated\n");
						break;
					} else {
						memcpy(&ft_test->age, &garbage, 4);
						anaPrint(ANA_NOTICE, "nothing new\n");
						show = 1;
						break;
					}
				} else {
					dummy = 0;
					memcpy(&ft_test->own, &dummy, 4);
					memcpy(&ft_test->age, &garbage, 4);
					ft_test->label = *(anaLabel_t *)test;
					show = 1;
					anaPrint(ANA_NOTICE, "entry successfully updated\n");
					break;
				}
			}
		}
		ft_test = ft_test->next;
	}
	ft_test = ft_start;
	
	if (!found) {
		ft->next = (struct tab *) malloc(sizeof(struct tab));
		ft->type = (char*) malloc(strlen(type) + 1);
		memcpy(ft->type, type, strlen(type) + 1);
		if (strcmp((char *) test, "KVR")==0) {
			ft->own = 1;
			ft->label = 0;
		} else {
			ft->own = 0;
			ft->label = *(anaLabel_t *)test;
		}
		ft->age = GARBAGE_LIFES;
		ft = ft->next;
		ft->type = NULL;
		ft->own = 0;
		ft->label = 0;
		ft->age = 0;
		ft->next = NULL;
		show = 1;
		anaPrint(ANA_NOTICE, "entry successfully added\n");
	}
	
	if (show)
		showFwd();
}

/* receive data function */
void fwd(char *data, int len, anaLabel_t input, void *aux) {
	
	/* parse the received xrp message */
	xrpCmd_t command = data;
	
	/* discover received */
	if (equalXRPCommands(command, XRP_CMD_DISCOVER)) {
		anaPrint(ANA_NOTICE, "\n\n--> received discover request\n\n");
		handleDiscover(data, len);
	}
	
	/* change received */
	if (equalXRPCommands(command, XRP_CMD_CHANGE)) {
		anaPrint(ANA_NOTICE, "\n\n--> received change request\n\n");
		handleChange(data, len);
	}
}

void garbageUpdate(void *unused) {
	int garbage_new, died = 0;
	struct tab *ft_back;
	
	ft_back = ft_start;
	while (ft_test->type!=NULL) {
		if (ft_test->age<=1) {
			if (ft_test->own==1) {
				if (ft_test==ft_start) {
					free(ft_test->type);
					ft_start = ft_test->next;
				} else {
					free(ft_test->type);
					if (ft_test->next==NULL) {
						ft_back->next = NULL;
					} else {
						ft_back->next = ft_test->next;
					}
				}
			} else {
				if (ft_test==ft_start) {
					free(ft_test->type);
					ft_start = ft_test->next;
				} else {
					free(ft_test->type);
					if (ft_test->next==NULL) {
						ft_back->next = NULL;
					} else {
						ft_back->next = ft_test->next;
					}
				}
			}
			died =  1;
		} else {
			garbage_new = ft_test->age - 1;
			memcpy(&ft_test->age, &garbage_new, 4);
		}
		if (ft_test!=ft_start)
			ft_back = ft_back->next;
		ft_test = ft_test->next;
	}
	ft_test = ft_start;
	
	if (died) {
		anaPrint(ANA_NOTICE, "\n\n--> route died\n\n");
		showFwd();
	}
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);
	
	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else
		if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

	anatimer_init();
	
	/* initializes tab list */
	ft = (struct tab *) malloc(sizeof(struct tab));
	ft_start = ft;
	ft_test = ft;
	ft->type = NULL;
	ft->own = 0;
	ft->label = 0;
	ft->age = 0;
	ft->next = NULL;
	
	/* register fwd callback function */
	recvIDP = anaL0_registerCallback((anaCallback_t) & fwd, NULL, NULL, NULL, 0, IDP_PERM, NOTHREAD);
	if (recvIDP < 0) {
		anaPrint(ANA_ERR,  "recvIDP: unable to register callback IDP to ANA minmex!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "registered recvIDP\n");
	
	/* publish in the KVR */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_ftab";
	service.valueLen = strlen("fbr_ftab")+1;

	lenMsg = anaL1_encCompartmentPublish(msg, recvIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_ERR, "problem encoding publish request for recvIDP in KVR!\n");
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_ERR, "problem sending publish request for recvIDP in KVR!\n");
		return -1;
	}
	
	showFwd();
	
	anatimer_add((1000 * GARBAGE_UPDATE_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)garbageUpdate, NULL, 0);
	
	free(msg);
	return 1;
}

