/* C/bricks/fbr/fbr_rtab.c
 *  
 * contains the routing table for field-based publish / subscribe
 * runs in user space
 * KVR entries: fbr_rtab
 * 
 * IMPORTANT: use minmex with -b option to extend default message size
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "anaLib2.h"
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "fbr_rtab";

/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "fbr.h"

anaLabel_t recvIDP;
anaLabel_t ftabLabel;
struct tab *rt;
struct tab *rt_start;
struct tab *rt_test;

void showTable(void);

int garbage = GARBAGE_LIFES;

/* called upon leaving the fbr_rtab */
void brick_exit() {
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&timeout, 0, sizeof(struct timespec));
	
	timeout.tv_sec = 1; 

        anaPrint(ANA_NOTICE, "FBR_RTAB HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex!= NULL\n");
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_rtab";
	service.valueLen = strlen("fbr_rtab")+1;

	anaL2_unpublish(NODE_LABEL, recvIDP, &context, &service, &timeout);
	
	anatimer_exit();

}

void sendToFwd(void *unused) {
	int totalLength, arglen, found = 0, dummy;
	int test_potential, new_potential;
	char *tmp;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	anaPrint(ANA_DEBUG, "sendToFwd\n");
	struct sendFwd {
		char *type;				/* service-type */
		int own;					/* stop forwarding */
		char *potential;			/* potential */
		anaLabel_t label;			/* label */
		struct sendFwd *next;
	} *fwd, *fwd_test, *fwd_start;
	
	memset(&timeout, 0, sizeof(struct timespec));
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	timeout.tv_sec = 1; 

	fwd = (struct sendFwd *) malloc(sizeof(struct sendFwd));
	fwd->potential = NULL;
	fwd_start = fwd;
	fwd_test = fwd;
	fwd->type = NULL;
	fwd->own = 0;
	fwd->potential = NULL;
	fwd->label = 0;
	fwd->next = NULL;
	
	while (rt_test->type!=NULL) {
		while(fwd_test->type!=NULL) {
			if (strcmp(fwd_test->type, rt_test->type)==0) {
				found = 1;
				if (strcmp(rt_test->potential, "infinite")!=0 && strcmp(fwd_test->potential, "infinite")!=0) {
					sscanf(rt_test->potential, "%i", &new_potential);
					sscanf(fwd_test->potential, "%i", &test_potential);
					if (new_potential>test_potential) {
						tmp = (char *) malloc(20);
						sprintf(tmp, "%i", new_potential);
						//fwd_test->potential = realloc(fwd_test->potential, strlen(tmp) + 1);
						if(fwd_test->potential != NULL)
							free(fwd_test->potential);
						fwd_test->potential = malloc(strlen(tmp) + 1);
						memcpy(fwd_test->potential, tmp, strlen(tmp) + 1);
						fwd_test->label = rt_test->label;
						free(tmp);
						found = 0;
						break;
					}
				} else {
					if (strcmp(fwd_test->potential, "infinite")==0 && rt_test->own==1) {
						dummy = 1;
						memcpy(&fwd_test->own, &dummy, 4);
						fwd_test->label = 0;
						found = 0;
						break;
					}
					if (strcmp(rt_test->potential, "infinite")==0 && strcmp(fwd_test->potential, "infinite")!=0) {
					//	fwd_test->potential = realloc(fwd_test->potential, strlen("infinite") + 1);
						if(fwd_test->potential != NULL)
							free(fwd_test->potential);
						fwd_test->potential = malloc(strlen("infinite") + 1);
						memcpy(fwd_test->potential, "infinite", strlen("infinite") + 1);
						if (rt_test->own==1) {
							dummy = 1;
							memcpy(&fwd_test->own, &dummy, 4);
							fwd_test->label = 0;
							found = 0;
							break;
						} else {
							dummy = 0;
							memcpy(&fwd_test->own, &dummy, 4);
							fwd_test->label = rt_test->label;
							found = 0;
							break;
						}
					}
				}
			}
			fwd_test = fwd_test->next;
		}
		fwd_test = fwd_start;
		if (!found) {
			fwd->next = (struct sendFwd *) malloc(sizeof(struct sendFwd));
			fwd->type = (char*) malloc(strlen(rt_test->type) + 1);
			memcpy(fwd->type, rt_test->type, strlen(rt_test->type) + 1);
			if (rt_test->own==0) {
				fwd->own = 0;
				fwd->label = rt_test->label;
			} else {
				fwd->own = 1;
				fwd->label = 0;
			}
			fwd->potential = (char*) malloc(strlen(rt_test->potential) + 1);
			memcpy(fwd->potential, rt_test->potential, strlen(rt_test->potential) + 1);
			fwd = fwd->next;
			fwd->type = NULL;
			fwd->own = 0;
			fwd->potential = NULL;
			fwd->label = 0;
			fwd->next = NULL;
			found = 0;
		}
		rt_test = rt_test->next;
	}
	rt_test = rt_start;
	
	while(fwd_test->type!=NULL) {
		msg = anaL1_allocateMessage();
		totalLength = fillXRPCommand(msg, XRP_CMD_CHANGE);
		if (totalLength<0) {
			anaPrint(ANA_EMERG, "sendToFwd: unable to fill command in message!\n");
			free(msg);
			return;
		}  
		arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, fwd_test->type, strlen(fwd_test->type) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "sendToFwd: unable to add service-type to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		if (fwd_test->own==1) {
			arglen = addXRPArg(msg, XRP_CLASS_LABEL, "KVR", 4);
		} else {
			arglen = addXRPArg(msg, XRP_CLASS_LABEL, &fwd_test->label, sizeof(anaLabel_t));
		}
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "sendToFwd: unable to add label to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
	
		if (ftabLabel == 0) {
			context.value = ".";
			context.valueLen = 2;
			service.value = "fbr_ftab";
			service.valueLen = strlen("fbr_ftab")+1;
			ftabLabel = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
			if (ftabLabel == 0) {
				anaPrint(ANA_EMERG, "forwarding table brick not found!\n");
				free(msg);
				return;
			}
		}
		anaL0_send(ftabLabel, msg, totalLength);
		anaPrint(ANA_NOTICE, "route update sent to forwarding table brick\n");
		free(msg);
		fwd_test = fwd_test->next;
	}
	fwd_test = fwd_start;
	
	while (fwd_test->type!=NULL) {
		if (fwd_test->own==1) {
			free(fwd_test->type);
			free(fwd_test->potential);
			fwd_start = fwd_test->next;
		} else {
			free(fwd_test->type);
			free(fwd_test->potential);
			fwd_start = fwd_test->next;
		}
		fwd_test = fwd_test->next;
	}
	fwd_test = fwd_start;
}

void handleUpdate(void *data, int len) {
	int typeLen, potLen, labelLen, type_known = 0, handled = 0, send = 0;
	void *test;
	char *type, *potential_m;
	struct tab *rt_back;
	
	type =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
	potential_m =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_POTENTIAL, &potLen, 0);
	test =  (void *) getXRPArg((xrpMsg_t) data, XRP_CLASS_LABEL, &labelLen, 0);
	
	anaPrint(ANA_NOTICE, "service-type: %s\n", type);
	anaPrint(ANA_NOTICE, "potential: %s\n", potential_m);
	
	if (strcmp(potential_m, "zero")==0) {
		rt_back = rt_start;
		while (rt_test->type!=NULL) {
			if (strcmp(rt_test->type, type)==0) {
				if (rt_test->own==1) {
					if (rt_test==rt_start) {
						free(rt_test->type);
						free(rt_test->potential);
						rt_start = rt_test->next;
					} else {
						free(rt_test->type);
						free(rt_test->potential);
						if (rt_test->next==NULL) {
							rt_back->next = NULL;
						} else {
							rt_back->next = rt_test->next;
						}
					}
				}
			}
			if (rt_test!=rt_start)
				rt_back = rt_back->next;
			rt_test = rt_test->next;
		}
		rt_test = rt_start;
		anaPrint(ANA_NOTICE, "entry successfully deleted\n");
		sendToFwd(NULL);
		showTable();
		return;
	}
	
	while (rt_test->type!=NULL) {
		if (strcmp(rt_test->type, type)==0) {
			type_known = 1;
			if (strcmp((char *) test, "KVR")==0 && rt_test->own==1) {
				memcpy(&rt_test->age, &garbage, 4);
				handled = 1;
				anaPrint(ANA_NOTICE, "nothing new\n");
				break;
			} else if (strcmp((char *) test, "KVR")!=0 && rt_test->own!=1 && *(anaLabel_t *)test == rt_test->label) {
				if (strcmp(rt_test->potential, potential_m)!=0) {
					memcpy(&rt_test->age, &garbage, 4);
					//rt_test->potential = realloc(rt_test->potential, strlen(potential_m) + 1);
					if(rt_test->potential != NULL)
						free(rt_test->potential);
					rt_test->potential = malloc(strlen(potential_m) + 1);
					memcpy(rt_test->potential, potential_m, strlen(potential_m) + 1);
					handled = 1;
					anaPrint(ANA_NOTICE, "entry successfully updated\n");
					send = 1;
					break;
				} else {
					memcpy(&rt_test->age, &garbage, 4);
					handled = 1;
					anaPrint(ANA_NOTICE, "nothing new\n");
					break;
				}
			}
		}
		rt_test = rt_test->next;
	}
	rt_test = rt_start;
	
	if (handled)
		showTable();
	
	if (!type_known || !handled) {
		rt->next = (struct tab *) malloc(sizeof(struct tab));
		rt->type = (char*) malloc(strlen(type) + 1);
		memcpy(rt->type, type, strlen(type) + 1);
		rt->potential = (char*) malloc(strlen(potential_m) + 1);
		memcpy(rt->potential, potential_m, strlen(potential_m) + 1);
		if (strcmp((char *) test, "KVR")==0) {
			rt->own = 1;
			rt->label = 0;
		} else {
			rt->own = 0;
			rt->label = *(anaLabel_t *)test;
		}
		rt->age = GARBAGE_LIFES;
		rt = rt->next;
		rt->type = NULL;
		rt->potential = NULL;
		rt->own = 0;
		rt->label = 0;
		rt->age = 0;
		rt->next = NULL;
		
		anaPrint(ANA_NOTICE, "entry successfully added\n");
		sendToFwd(NULL);
		showTable();
	}
	
	if (send)
		sendToFwd(NULL);
	
	return;
}

/* receive messages */
void recvData_rtab(struct anaL2_message *data) {
	int length;
	void *saddr, *message;
	xrpCmd_t command;
	length = data->dataLen;
	saddr = (char *) data->senderContext;
	message = (char *) data->data;
	
	/* parse the received xrp message */
	command = message;
	
	if (equalXRPCommands(command, XRP_CMD_ROUTEUP)) {
		anaPrint(ANA_NOTICE, "\n\n--> received route update\n\n");
		handleUpdate(message, length);
	}
}

void showTable(void) {
	anaLabel_t tmp;
		
	anaPrint(ANA_NOTICE, "+---------------------+-----------------------+---------------+\n");
	anaPrint(ANA_NOTICE, "| Service type\t| Potential\t\t| Label\t\t|\n");
	anaPrint(ANA_NOTICE, "+---------------------+-----------------------+---------------+\n");
	
	while (rt_test->type!=NULL) {
		if (rt_test->own==1) {
			tmp = NODE_LABEL;
		}
		else {
			tmp = rt_test->label;
		}
		if (strlen((char *) rt_test->type)<14) {
			anaPrint(ANA_NOTICE, "| %s\t\t| %s\t\t| %i\t\t|\n", (char *) rt_test->type, (char *) rt_test->potential, tmp);
		}
		else {
			anaPrint(ANA_NOTICE, "| %s\t| %s\t\t| %i\t\t|\n", (char *) rt_test->type, (char *) rt_test->potential, tmp);
		}
		anaPrint(ANA_NOTICE, "+---------------------+-----------------------+---------------+\n");
		rt_test = rt_test->next;
	}
	rt_test = rt_start;
}

void garbageUpdate(void *unused) {
	int garbage_new, died = 0;
	struct tab *rt_back;
	anaPrint(ANA_DEBUG, "garbageUpdate");	
	rt_back = rt_start;
	while (rt_test->type!=NULL) {
		if (rt_test->age<=1) {
			if (rt_test->own==1) {
				if (rt_test==rt_start) {
					free(rt_test->type);
					free(rt_test->potential);
					rt_start = rt_test->next;
				} else {
					free(rt_test->type);
					free(rt_test->potential);
					if (rt_test->next==NULL) {
						rt_back->next = NULL;
					} else {
						rt_back->next = rt_test->next;
					}
				}
			} else {
				if (rt_test==rt_start) {
					free(rt_test->type);
					free(rt_test->potential);
					rt_start = rt_test->next;
				} else {
					free(rt_test->type);
					free(rt_test->potential);
					if (rt_test->next==NULL) {
						rt_back->next = NULL;
					} else {
						rt_back->next = rt_test->next;
					}
				}
			}
			died =  1;
		} else {
			garbage_new = rt_test->age - 1;
			memcpy(&rt_test->age, &garbage_new, 4);
		}
		if (rt_test!=rt_start)
			rt_back = rt_back->next;
		rt_test = rt_test->next;
	}
	rt_test = rt_start;
	
	if (died) {
		anaPrint(ANA_NOTICE, "\n\n--> route died\n\n");
		sendToFwd(NULL);
		showTable();
	}
}

int brick_start() {
		
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&timeout, 0, sizeof(struct timespec));
	
	timeout.tv_sec = 1; 

	anaL2_initDefault();
	anatimer_init();
	
	ftabLabel = 0;
	
	/* initializes rtab list */
	rt = (struct tab *) malloc(sizeof(struct tab));
	rt_start = rt;
	rt_test = rt;
	rt->type = NULL;
	rt->potential = NULL;
	rt->own = 0;
	rt->label = 0;
	rt->age = 0;
	rt->next = NULL;
	
	/* publish in KVR */
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_rtab";
	service.valueLen = strlen("fbr_rtab")+1;

	recvIDP = anaL2_publish(NODE_LABEL, &context, &service, recvData_rtab, THREAD, &timeout);
	if (recvIDP == 0) {
		anaPrint(ANA_EMERG, "publish failed!\n");
		return -1;
	}
	
	showTable();
	
	anatimer_add((1000 * GARBAGE_UPDATE_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)garbageUpdate, NULL, 0);
	anatimer_add((1000 * ROUTE_UPDATE_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)sendToFwd, NULL, 0);
	
	return 1;
}

