 /* C/bricks/fbr/fbr_serv.c
 *  
 * sample service application
 * depends on: field-based publish / subscribe compartment
 * runs in user space
 * KVR entries: fbr_serv
 * 
 * use:
 *
 * fbr_serv -a "{service-type+capacity}"
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "anaLib2.h"
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */

char mymodename[] = "fbr_serv";


/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */
#include "fbr.h"

anaLabel_t recvIDP;
anaLabel_t potfLabel;
anaLabel_t forwLabel;

/* called upon leaving the fbr_serv */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;
	char *target, *tmp;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

        anaPrint(ANA_NOTICE, "FBR_SERV HALTED\n");
	anaPrint(ANA_NOTICE, "in minmex!= NULL\n");
	
	msg = anaL1_allocateMessage();
	/* unpublish service */
	target ="gateway";
	if (getAuxArg(0)!=NULL) {
		tmp = getAuxArg(0);
		target = strsep(&tmp, "+");
	}
	context.value = ".";
	context.valueLen = 2;
	service.value = target;
	service.valueLen = strlen(target)+1;

	len = anaL1_encCompartmentUnpublish(msg, recvIDP, &context, &service);
	ret = anaL0_send(potfLabel, msg, len);
	free(msg);
	
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_serv";
	service.valueLen = strlen("fbr_serv") + 1;

	len = anaL1_encCompartmentUnpublish(msg, recvIDP, &context, &service);
	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request in KVR!\n");
	}
	free(msg);
	
	anaL0_unregisterCallback(recvIDP, 0);
	anatimer_exit();
}

/* receive data function */
void recvData_serv(void *data, int len, anaLabel_t input, void *aux) {
	int typeLen, idLen, totalLength, arglen, messageLen;
	char *type, *id, *message;
	xrpMsg_t msg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;

	/* parse the received xrp message */
	xrpCmd_t command = data;

	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 
	
        if (equalXRPCommands(command, XRP_CMD_SUBSCRIBE)) {
		anaPrint(ANA_NOTICE, "\n\n--> subscribe message received\n\n");
		
		type =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SERVICETYPE, &typeLen, 0);
		id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGEID, &idLen, 0);
	
		anaPrint(ANA_NOTICE, "message-id: %s\n", id);
	
		msg = anaL1_allocateMessage();
		totalLength = fillXRPCommand(msg, XRP_CMD_RESPONSE);
		if (totalLength<0) {
			anaPrint(ANA_EMERG, "recvData: unable to fill command in message!\n");
			free(msg);
			return;
		}  
		arglen = addXRPArg(msg, XRP_CLASS_SERVICETYPE, type, strlen(type) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "recvData: unable to add service-type to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		arglen = addXRPArg(msg, XRP_CLASS_MESSAGEID, id, strlen(id) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "recvData: unable to add message-id to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		arglen = addXRPArg(msg, XRP_CLASS_MESSAGE, "Hi! I received your subscribe message!", strlen("Hi! I received your subscribe message!") + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "recvData: unable to add message to message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
	
		if (forwLabel == 0) {
			context.value = ".";
			context.valueLen = 2;
			service.value = "fbr_forw";
			service.valueLen = strlen("fbr_forw")+1;
			forwLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
			if (forwLabel == 0) {
				anaPrint(ANA_EMERG, "forwarding brick not found!\n");
				free(msg);
				return;
			}
		}
		anaL0_send(forwLabel, msg, totalLength);
		anaPrint(ANA_NOTICE, "response sent to forwarding brick\n");
		free(msg);
		
	} else if (equalXRPCommands(command, XRP_CMD_DATA)) {
		anaPrint(ANA_NOTICE, "\n\n--> data message received\n\n");
		
		id =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGEID, &idLen, 0);
		message =  (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_MESSAGE, &messageLen, 0);
	
		anaPrint(ANA_NOTICE, "message-id: %s\n", id);
		anaPrint(ANA_NOTICE, "message: %s\n", message);
		anaValidate("message %s\n", message);
	}
}

void updateService(void *unused) {
	int totalLength, arglen; 
	xrpMsg_t msg;
	char *target;
	
	if (getAuxArg(0)!=NULL && getAuxArg(1)!=NULL) {
		target = (char *) malloc(50);
		memcpy(target, getAuxArg(0), strlen(getAuxArg(0)));
		memcpy(target + strlen(getAuxArg(0)), "+", 1);
		memcpy(target + strlen(getAuxArg(0)) + 1, getAuxArg(1), strlen(getAuxArg(1)) + 1);
	} else 
		target = "gateway+12";
	
	/* update service */
	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_PUBLISHUP);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "updateService: unable to fill command in message!\n");
		free(msg);
		return;
	}  
	arglen = addXRPArg(msg, XRP_CLASS_DESCRIP, target, strlen(target) + 1);
	if (arglen < 0) {
		anaPrint(ANA_EMERG, "updateService: unable to add description to message!\n");
		free(msg);
		return;
	}
	totalLength += arglen;
	
	anaL0_send(potfLabel, msg, totalLength);
	anaPrint(ANA_NOTICE, "publish update sent\n");
	free(msg);
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

void handleLookupReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name;
	int labelSize;
	char *target;
	anaLabel_t *temp;
	struct context_s context;
	struct service_s service;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));

	anaPrint(ANA_NOTICE, "received lookup reply for field computing brick in the KVR\n");
	
	if (anaL1_isError(data)) { /* if ERROR command */
		return;
	}
	
	if (anaL1_decResponse(data, XRP_CLASS_NAME, 0, (void**) &name)>0) {
		labelSize =  anaL1_decResponse(data, XRP_CLASS_LABEL, 0, (void**) &temp);
		if (name!=NULL && temp != 0) {
			int lenMsg, ret; 
			xrpMsg_t msg;
			potfLabel = *temp;
			if (getAuxArg(0)!=NULL && getAuxArg(1)!=NULL) {
				target = (char *) malloc(50);
				memcpy(target, getAuxArg(0), strlen(getAuxArg(0)));
				memcpy(target + strlen(getAuxArg(0)), "+", 1);
				memcpy(target + strlen(getAuxArg(0)) + 1, getAuxArg(1), strlen(getAuxArg(1)) + 1);
			} else 
				target = "gateway+12";
			
			/* publish service */
			msg = anaL1_allocateMessage();
			context.value = "*";
			context.valueLen = 2;
			service.value = target;
			service.valueLen =  strlen(target) + 1;

			lenMsg = anaL1_encCompartmentPublish(msg, recvIDP, &context, &service, 1);
			ret =  anaL1_requestReply(potfLabel, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
			free(msg);
			anaPrint(ANA_NOTICE, "service published: %s\n", target);
			
			anatimer_init();
			anatimer_add((1000 * SERVICE_UPDATE_TIME), ANATIMER_PERIODIC, ANATIMER_ABSOLUTE, (fct_t)updateService, NULL, 0);
		}
	}
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	char *target;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));

	validate_init("fbr");

	forwLabel = 0;
	
	/* register data receive callback function */
	recvIDP = anaL0_registerCallback((anaCallback_t) & recvData_serv, NULL, NULL, NULL , 0, IDP_PERM, THREAD);
	if (recvIDP < 0) {
		return 1;
	}
	anaPrint(ANA_NOTICE,  "registered recvIDP\n");
	
	target ="gateway";
	if (getAuxArg(0)!=NULL && getAuxArg(1)!=NULL) {
		target = getAuxArg(0);
	}
	
	/* publish in the KVR */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = target;
	service.valueLen =  strlen(target) + 1;

	lenMsg = anaL1_encCompartmentPublish(msg, recvIDP, &context, &service, 1);
	if (lenMsg<0) {
		anaPrint(ANA_ERR, "problem encoding publish request for recvIDP in KVR!\n");
		free(msg);
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_ERR, "problem sending publish request for recvIDP in KVR!\n");
		free(msg);
		return -1;
	}
	free(msg);
	
	/* lookup the field computing brick by name */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value = "fbr_potf";
	service.valueLen = strlen("fbr_potf")+1;

	lenMsg = anaL1_encResolve(msg, &context, &service, 'u', &description);
	if (lenMsg<0) {
		return -1;
	}
	ret = anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handleLookupReply, NULL, NULL, NULL, 1); 
	if (ret<0) {
		return -1;
	}
	anaPrint(ANA_NOTICE, "lookup query sent\n");
	free(msg);
	return 1;
}

