/* C/shared/brick_template.h
 * 
 * this file has to be included in every brick
 * provides basic initialization of brick
 * attaches brick to minmex
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   Ariane Keller <ariane.keller@tik.ee.ethz.ch> 
 *
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#ifndef _BRICK_TEMPLATE_H
#define _BRICK_TEMPLATE_H

/* "standard" ANA include files for bricks */
/*#include "anaValidate.h"*/

#ifdef USER
#include <stdio.h>
#include <strings.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <dirent.h>

#endif

#ifdef KERNEL
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/random.h>
#include <linux/net.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/mm.h>
#include <linux/types.h>
#include <linux/netdevice.h>
#endif

/* ANA includes */
#include "anaLib0.h"
#include "anaCommon.h"
#include "anaLib2.h"
#include "anaValidate.h"

#ifdef KERNEL

/* "ANA default" kernel module parameters for bricks */

char *a[10];
int numArg = 0;
module_param_array(a, charp, &numArg, 0000);
MODULE_PARM_DESC(a, "array of up to 10 additional arguments that will be passed to the brick_start function");
char *D = NULL;
module_param(D, charp, 0);
MODULE_PARM_DESC(D,"log level for debugging information" );
#endif /*KERNEL */

anaHandle_t myHandle; /* brick handle */
anaLabel_t myNodeCompLabel;
int attached;

struct argList {
	char *arg;
	struct argList *next;
};

struct argList *auxList;

char* getAuxArg(int index);
void addAuxArg(char *auxArg);

int anaDebugLevel = ANA_DEBUG;

char* getAuxArg(int idx){
	int i=0;
	struct argList *tmp=auxList;
	while(i<idx && tmp != NULL){
		tmp = tmp->next;
		i++;	
	}
	if(tmp==NULL)
		return NULL;
	return tmp->arg;
}

void freeAuxArgs(void){
	struct argList *tmp=NULL;
	while(auxList != NULL){
		tmp = auxList->next;
		free(auxList->arg);
		free(auxList);
		auxList = tmp;
	}
}

void addAuxArg(char *auxArg){
	struct argList *tmp=auxList;
	struct argList *toAdd = (struct argList *) malloc(sizeof(struct argList));
	toAdd->arg = (char*) malloc(strlen(auxArg)+1);
	toAdd->next = NULL;
	strcpy(toAdd->arg,auxArg);
	
	if(auxList ==NULL)
		auxList = toAdd;
	else{
		while( tmp->next != NULL){
			tmp = tmp->next;
		}
		tmp->next = toAdd;
	}
}

unsigned long int anaThread(void * (*start_routine)(void *), void *arg, char *name){

	/* function enforcing that the brick handle is properly */
	/* passed to the function call (and recorded in the anathread list */
	
	/* must be defined here and not in anaThread.c such that myHandle */
	/* is properly linked */
	
	if (!start_routine)
		return 0;
		
	return _anaThread(myHandle, start_routine, arg, name);
}

#ifdef USER
static char ipCh[20] = "empty";
static char mCh[20] = "meaningless";
static char ethCh[20] = "empty";
static char gwCh[20] = "empty";

#ifdef GATES_MODE
#include "gatesConstants.h"

int main(int argc, char **argv) {

	int nodeDataMode = UNIX;
	char ctrlCh[115];
	char dataCh[115];
	char ndCtrlCh[115];
	char debug[12];

	char c;
       	int ndFlag=0, cFlag=0, dFlag=0, iFlag=0, mFlag=0, eFlag=0, gFlag=0;
	if (log_init() < 0)
		return -1;
	auxList =NULL;
	opterr = 0;
	while ((c = getopt (argc, argv, "c:d:n:D:a:i:m:e:g:h:")) != -1) /* n option is always needed*/
		switch (c) { 
			case 'c':{ 
				cFlag=1;
				strcpy(ctrlCh,optarg);
				break;
			}
			case 'd':{   
				dFlag=1;
				strcpy(dataCh,optarg);
				break;
			}
			case 'n':{   
				ndFlag=1;
				strcpy(ndCtrlCh,optarg);
				break;
			}
			case 'a':{   /*auxiliary argument*/
				addAuxArg(optarg);
				break;
			}
			case 'D':{ 
				strcpy(debug, optarg);
				if(!memcmp(debug, "ANA_NONE", 8))
					anaDebugLevel = ANA_NONE;
				if(!memcmp(debug, "ANA_EMERG", 9))
					anaDebugLevel = ANA_EMERG;
				if(!memcmp(debug, "ANA_ERR", 7))
					anaDebugLevel = ANA_ERR;
				if(!memcmp(debug, "ANA_NOTICE", 10))
					anaDebugLevel = ANA_NOTICE;
				if(!memcmp(debug, "ANA_DEBUG", 9))
					anaDebugLevel = ANA_DEBUG;
				if(!memcmp(debug, "ANA_TIMER", 9))
					anaDebugLevel = ANA_TIMER;
				if(!memcmp(debug, "ANA_VALIDATE", 12))
					anaDebugLevel = ANA_VALIDATE;
				break;

			}
			case 'i':{   
				iFlag=1;
				strcpy(ipCh,optarg);
				break;
			}
			case 'e':{   
				eFlag=1;
				strcpy(ethCh,optarg);
				break;
			}
			case 'm':{   
				mFlag=1;
				strcpy(mCh,optarg);
				break;
			}
			case 'g':{   
				gFlag=1;
				strcpy(gwCh,optarg);
				break;
			}

		}
	
	/* we check that the minmex control gate is provided */
	if(ndFlag==0){
		anaPrint(ANA_EMERG, "Please provide a minmex control gate URL\n");
		
		anaPrint(ANA_EMERG,  "Usage: %s [-d brick data URL] [-c brick control URL] -n 'minmex control URL'\n", argv[0]);
		exit(1);
	}
	
	/* we check that a valid config is given */
	if(!((iFlag==1 && mFlag==1 && eFlag==1 && gFlag==0) || (iFlag==0 && mFlag==0 && eFlag==0 && gFlag==1) || (iFlag==0 && mFlag==0 && eFlag==0 && gFlag==0))){
		anaPrint(ANA_EMERG, "Please provide a valid configuration order (either -i {ip address} -m {netmask} -e {ethernet brick} or -g {next hop})\n");
		exit(1);
	}
            
       	if(!cFlag){
       		int r;
		char c1,c2;
		memset(ctrlCh,0, 108);
		/*we generate 2 random chars from a-z*/
		srandom(time(0));
		r = (int) random();
		c1='a'+(r%26);
		r = (int) random();
		c2='a'+(r%26);
		sprintf(ctrlCh, "unix:///tmp/anaControl_%s_%c%c",mymodename, c1, c2);
       }
     
     	if(!dFlag){
       		int r;
		char c1,c2;
		memset(dataCh,0, 108);
		/*we generate 2 random chars from a-z*/
		srandom(time(0));
		r = (int) random();
		c1='a'+(r%26);
		r = (int) random();
		c2='a'+(r%26);
		sprintf(dataCh, "unix:///tmp/anaData_%s_%c%c",mymodename, c1, c2);
       }
        
	if(!memcmp(ndCtrlCh, "PIPES", 5)||!memcmp(ndCtrlCh, "pipes", 5))
		nodeDataMode = PIPES;
	if(!memcmp(ndCtrlCh, "UNIX", 4)||!memcmp(ndCtrlCh, "unix", 4))
		nodeDataMode = UNIX;
	if(!memcmp(ndCtrlCh, "GENETLINK", 9)||!memcmp(ndCtrlCh, "genetlink", 9))
		nodeDataMode = GENETLINK;
	if(!memcmp(ndCtrlCh, "UDP", 3)||!memcmp(ndCtrlCh, "udp", 3))
		nodeDataMode = UDP;

	anaPrint(ANA_DEBUG, "-----------------------nodeDataMode %i\n", nodeDataMode);

	myMinmex = anaL0_initMinmexSpecs(dataCh, ctrlCh, ndCtrlCh, nodeDataMode, 5);  /* timeout value 5*/

	if (myMinmex==NULL || anaL0_attach(myMinmex, 10)<0) {
		anaPrint(ANA_EMERG, "Unable to attach to ANA minmex \n");
		freeMinmexSpecs(myMinmex);
		return -1;
	}
	brick_start();
	
	pthread_exit(0);
	return 1;
}

#endif /* ifdef GATES_MODE*/

#ifdef PLUGIN_MODE

int __attribute__((visibility("default"))) so_brick_init(struct pluginit_s *pluginit);
char __attribute__((visibility("default"))) *so_brickname(void);
void __attribute__((visibility("default"))) *so_log_signal(pthread_t pid, int sig, char *brickname, char *threadname);

char* so_brickname(void) { return mymodename; }

void* so_log_signal(pthread_t pid, int sig, char *brickname, char *threadname)
{
	/* to log the segfault in the brick's log file */

	if (sig == SIGSEGV)
	{
		anaPrint(ANA_ERR, "Thread %lu (%s %s) doing segmentation fault!\n", 
			 pid, brickname, threadname);
	}
	else if (sig == SIGTERM)
	{
		anaPrint(ANA_ERR, "Thread %lu (%s %s) receiving SIGTERM!\n", 
			 pid, brickname, threadname);
	}
		 
	return NULL; /* for dlsym() and compiler to be happy */
}

int so_brick_init(struct pluginit_s *pluginit) {
	/* the commadline contains only the auxiliary arguments separated by a space*/
	char *token = NULL;
	struct commSpecs_s specs;
	char *commandline = NULL;
	struct attachRes_s *attachRes = NULL;
  	int iFlag=0, mFlag=0, eFlag=0, gFlag=0;;
	struct condLock0_s lock;

	auxList = NULL;
	log_init();	
	if (pluginit != NULL)
		commandline = pluginit->commandLine;
	
	if (commandline != NULL)
		token = strsep(&commandline, " ");
	anaPrint(ANA_DEBUG, "0 starting brick\n");	
	
	/* we expect the following syntax: i=... */
	while(token!=NULL){
		/* temporary hack to allow different debug levels for different bricks */
		switch(*token) {
			case 'i':{   
				iFlag=1;
				strcpy(ipCh,token+2);
				break;
			}
			case 'e':{   
				eFlag=1;
				strcpy(ethCh,token+2);
				break;
			}
			case 'm':{   
				mFlag=1;
				strcpy(mCh,token+2);
				break;
			}
			case 'g':{   
				gFlag=1;
				strcpy(gwCh,token+2);
				break;
			}
			case 'D':{
				anaDebugLevel = ANA_VALIDATE;
				break;
			}
			default: {
				addAuxArg(token);
			}
		}
		token = strsep(&commandline, " ");
	}
	
	/* we check that a valid config is given */
	if(!((iFlag==1 && mFlag==1 && eFlag==1 && gFlag==0) || (iFlag==0 && mFlag==0 && eFlag==0 && gFlag==1) || (iFlag==0 && mFlag==0 && eFlag==0 && gFlag==0))){
		anaPrint(ANA_EMERG, "Please provide a valid configuration order (either -i {ip address} -m {netmask} -e {ethernet brick} or -g {next hop})\n");
		exit(1);
	}
	anaPrint(ANA_DEBUG, "1 starting brick\n");	

	messageQueue = NULL;
	analock_init(&queueLock);
	analock_cond_init(&waitLock);
	/* now we attach to the minmex */
	memset(&specs, 0,  sizeof(struct commSpecs_s));
	specs.queueLock = &queueLock;  
	specs.msgQueue = &messageQueue; 
	specs.condLock = &waitLock;
	specs.pluginHandle = pluginit->pluginHandle;
	specs.pluginFlag = 1;
	mypluginHandle = pluginit->pluginHandle;
	anaPrint(ANA_DEBUG, "2 starting brick\n");	
	
	/* we need to have attached = 1 before we create the dataLoop!
 	 * otherwise it will die silently */	
	attached = 1;
	
	memset( &lock, 0, sizeof(struct condLock0_s));
	analock_init(&(lock.condLock));
	analock_cond_init(&(lock.cond));
	lock.flag = 0;

	if(pluginit != NULL)
		free(pluginit->commandLine);
	
	free(pluginit);
	anaPrint(ANA_DEBUG, "3 starting brick\n");	

	attachRes = handleAttach(mymodename, &specs);
	if (attachRes==NULL){
		anaPrint(ANA_EMERG, "Unable to attach to ANA minmex \n");
		attached = 0;
		return -1; 
	}
	
	myHandle = attachRes->handle;
	myNodeCompLabel = attachRes->nodeCompLabel;
	
	/* we free the returned attachRes*/
	free(attachRes);
	
	/* we start the main data receive loop for the thread */
	_anaThread(myHandle, (void *) dataLoop, &lock, "dataloop");
			
	/* we initialize the heartBeat timer */
	anatimer_init();
	anaPrint(ANA_DEBUG, " 4 starting brick\n");	
		
	if ( anatimer_add(HEARTBEAT_INTERVAL*1000, 1, 1, (fct_t)heartBeat, NULL, 0) == -1 ){
		anaPrint(ANA_EMERG, "Unable to set heartbeat time --> this brick will be soon kicked by the minmex \n");
		return -1; 
	}
	anaPrint(ANA_DEBUG, " 5 starting brick\n");	
	
	analock_lock(&(lock.condLock));
	/* we lock here in order to make sure that the data loop has started before we go any further (with brick_start)*/
	while(lock.flag==0)
		analock_cond_wait( &(lock.cond) , &(lock.condLock));
	

	/* we launch the brick_start function*/
	anaPrint(ANA_DEBUG, "starting brick\n");
	sleep(1);	
	_anaThread(myHandle, (void *) brick_start, NULL, "brick_start");
	
	/* probably we should exit this thread correctly after  brick_start */
	
	return 1;
}	

#endif
#endif /* USER */


#ifdef KERNEL

struct anaMinmexFunctions *myminmex;
analock_rw_t attachLock; /*protects the attached variable */

struct dataMessageQueue_s *recvDataQueue;
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
DECLARE_WORK(rcv_data, recvData);
#else
/*name, function, data*/
DECLARE_WORK(rcv_data, recvData, NULL); 
#endif
struct workqueue_struct *wq;
analock_t queueLock;
char *validatebuf;
int validateoffset = 0;

#ifdef withgates
static int __init MODULE_NAME_init(void){
	int nodeDataMode = KERN;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	EXPORT_NO_SYMBOLS;
#endif		
	anaDebugLevel = ANA_DEBUG ;
	if(!memcmp(debug, "ANA_NONE", 8))
		anaDebugLevel = ANA_NONE;
	if(!memcmp(debug, "ANA_EMERG", 9))
		anaDebugLevel = ANA_EMERG;
	if(!memcmp(debug, "ANA_ERR", 7))
		anaDebugLevel = ANA_ERR;
	if(!memcmp(debug, "ANA_NOTICE", 10))
		anaDebugLevel = ANA_NOTICE;
	if(!memcmp(debug, "ANA_DEBUG", 9))
		anaDebugLevel = ANA_DEBUG;
	if(!memcmp(debug, "ANA_TIMER", 9))
		anaDebugLevel = ANA_TIMER;
	if(!memcmp(debug, "ANA_VALIDATE", 12))
		anaDebugLevel = ANA_VALIDATE;


	if(!memcmp(cd, "KERN", 4)||!memcmp(cd, "kern", 4))
		nodeDataMode = KERN;
	if(!memcmp(cd, "GENETLINK", 9)||!memcmp(cd, "genetlink", 9))
		nodeDataMode = GENETLINK;
	if(!memcmp(cd, "UDP", 3)||!memcmp(cd, "udp", 3))
		nodeDataMode = UDP;


        anaPrint(ANA_DEBUG,  "Kernel module initialized\n");
        anaPrint(ANA_DEBUG,  "brick data url = %s\n", cd);
	anaPrint(ANA_DEBUG,  "brick control url = %s\n", cc);
        anaPrint(ANA_DEBUG,  "minmex control url = %s\n", nc);
        anaPrint(ANA_DEBUG,  "debug = %i\n", anaDebugLevel);
        anaPrint(ANA_DEBUG,  "number of aux arguments: %i\n", numArg);
	auxList = NULL;
        while(numArg > 0) {
                addAuxArg(a[numArg - 1]);
                numArg--;
        }

	myminmex = anaL0multi_initMinmexSpecs(cd, cc, nc, nodeDataMode , 5);  /* timeout value 5*/
	
	if(myminmex==NULL || anaL0multi_attach(myminmex, 10)<0) {
		anaPrint(ANA_EMERG, "Brick : Unable to attach to ANA minmex \n");
                freeMinmexSpecs(myminmex);
		return -14; 
	}
	#ifdef MULTIPLE_MINMEX
	brick_start(myminmex);
	#else
	brick_start();
	#endif
	return 1;
}
#else //withgates

/* we need a notification function to avoid that a brick
 * crashes when the minmex gets stopped before the brick
 * in the future this will be done by the notification 
 * system.
 */
void notificationTemp(void) {
	analock_write_lock(&attachLock);
	attached = 0;
	analock_write_unlock(&attachLock);

}

static int __init MODULE_NAME_init(void){
	int i = 0;
	struct attachRes_s *attachRes = NULL;
	struct anaBrickFunctions *myFunctions = NULL;
	recvDataQueue = NULL;
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	EXPORT_NO_SYMBOLS;
#endif		
	anaDebugLevel = ANA_DEBUG ;
	if (D != NULL) {
		if(!memcmp(D, "ANA_NONE", 8))
			anaDebugLevel = ANA_NONE;
		if(!memcmp(D, "ANA_EMERG", 9))
			anaDebugLevel = ANA_EMERG;
		if(!memcmp(D, "ANA_ERR", 7))
			anaDebugLevel = ANA_ERR;
		if(!memcmp(D, "ANA_NOTICE", 10))
			anaDebugLevel = ANA_NOTICE;
		if(!memcmp(D, "ANA_DEBUG", 9))
			anaDebugLevel = ANA_DEBUG;
		if(!memcmp(D, "ANA_TIMER", 9))
			anaDebugLevel = ANA_TIMER;
		if(!memcmp(D, "ANA_VALIDATE", 12))
			anaDebugLevel = ANA_VALIDATE;
	}
        anaPrint(ANA_DEBUG,  "Kernel module initialized\n");
        anaPrint(ANA_DEBUG,  "debuglevel = %i\n", anaDebugLevel);
        anaPrint(ANA_DEBUG,  "number of aux arguments: %i\n", numArg);
	auxList = NULL;
        while(i < numArg) {
                addAuxArg(a[i]);
                i++;
        }

	/* for Linux kernel 2.6.12 and older the name can be nolonger than 10! */
	if(strlen(THIS_MODULE->name) >= 10) {
		char tmp[11];
		memset(tmp, 0, 11);
		memcpy(tmp, THIS_MODULE->name,10);
		wq = create_workqueue(tmp);
	}
	else
		wq = create_workqueue(THIS_MODULE->name);
	
	/* the minmex registers itself with its name (THIS_MODULE->name),
 	 * if nothing is changed this is anaMinmex
 	 */
	if((myminmex = getMinmexFunctions("anaMinmex"))== NULL){
		anaPrint(ANA_ERR, "couldn't attach to minmex\n");
		return -1;
	}            
	/*****************************************************************/
	recvDataQueue = NULL;
	analock_init(&queueLock);
	myFunctions = malloc(sizeof(struct anaBrickFunctions));
        myFunctions->wq = wq;				/* the work queue */
        myFunctions->work = &rcv_data;			/* which function we are submitting to the work queue */
	myFunctions->recvDataQueue = &recvDataQueue;	/* the message queue where to add new messages */
	myFunctions->queueLock = &queueLock;		/*this lock protects the recvDataQueue */
	myFunctions->notification = notificationTemp;

   	attachRes = myminmex->attach(THIS_MODULE->name, myFunctions);
	if(attachRes==NULL){
		anaPrint(ANA_EMERG, "Unable to attach to ANA minmex \n");
		return -1; 
	}
	
	myHandle = attachRes->handle;
	myNodeCompLabel = attachRes->nodeCompLabel;
	
	/* we free the returned attachRes*/
	free(attachRes);

	analock_rw_init(&attachLock);
	analock_write_lock(&attachLock);
	attached = 1;
	analock_write_unlock(&attachLock);

	
	/* we initialize the heratBeat timer */
	anatimer_init();
		
	if ( anatimer_add(HEARTBEAT_INTERVAL*1000, 1, 1, (fct_t)heartBeat, NULL, 0) == -1 ){
		anaPrint(ANA_EMERG, "Unable to set heartbeat time --> this brick will be soon kicked by the minmex \n");
		return -1; 
	}
	/* we launch the brick_start function*/	
	return brick_start();
	return 1;
}
#endif
	
static void __exit MODULE_NAME_exit( void )
{
	brick_exit();
	analock_write_lock(&attachLock);
        if(attached)
       		myminmex->detach(myHandle);
	
	attached = 0;
	analock_write_unlock(&attachLock);
	anatimer_exit();

	if( wq ) {
                flush_workqueue(wq);
                destroy_workqueue(wq);
        }

	analock_rw_destroy(&attachLock);
	if (anaDebugLevel == ANA_VALIDATE) {
		validate_exit();
	}

}

module_init(MODULE_NAME_init);
module_exit(MODULE_NAME_exit);

#endif /* KERNEL */

#endif /* _BRICK_TEMPLATE_H */
