/* Thanks to Martin Burkhart for providing this TUI!
 *
 * Contains methods for the textual user interface (TUI) of the chat client. The interface is
 * based on ncurses. Method comments are found in the header file.
 * 
 * See http://tldp.org/HOWTO/NCURSES-Programming-HOWTO/ for an introduction to ncurses. 
 */

#include <ncurses.h>
#include "client-tui.h"

const int INPUT_WIN_HEIGHT = 7;
const int BOARD_WIN_BOTTOM_SPACE = 5;

WINDOW *input_win, *board_win;

void init_tui() {

    initscr();			/* Start curses mode 		*/
    noecho();
    cbreak();			/* Line buffering disabled      */
    keypad(stdscr, TRUE);	/* I need that nifty F1 	*/
 //   keypad(input_win, TRUE);
 //   keypad(board_win, TRUE);
    timeout(50);		/* Don't block on input 	*/
    refresh();

    // Input window
    input_win = newwin(INPUT_WIN_HEIGHT, COLS, LINES-INPUT_WIN_HEIGHT, 0);
    reset_input_win(input_win);

    // Message board
    board_win = newwin(LINES-INPUT_WIN_HEIGHT, COLS, 0, 0);
    scrollok(board_win, TRUE);

    refresh_tui();
}

void del_char(){
	int x, y;
	getyx(input_win, y, x);
	mvwdelch(input_win, y, x-1);
	refresh_tui();	
}

void reset_input_win() {
    wclear(input_win);
    mvwprintw(input_win,0,0, "-----------[ANA Demo IP - Chat Application]---------------------------");
    mvwprintw(input_win,1,0, "> ");
    wrefresh(input_win);
}

void refresh_tui() {
    wrefresh(board_win);
    wrefresh(input_win);
}

void close_tui() {
    endwin();
}

void post_message(char *msg) {
    wprintw(board_win, "%s\n", msg);
    wrefresh(board_win);
    wrefresh(input_win);
}

void echo_character(char c) {
    waddch(input_win, c);
    wrefresh(input_win);
}
