/** \file
 * IP encapsulation brick
 */

/* C/bricks/ip/ip_enc.c
 *  
 * implementation of the ip compartment
 * depends on: dgrmEthBrick
 * runs in user space
 * KVR entries: ip+internet+inet+compartment
 * 
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	   Ariane Keller <arkeller@ee.ethz.ch> Adaption for release 0.2
 *
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"

/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

/* !!! Change the two variables below to the name of your brick !!! */
char mymodename[] = "ip_enc";

/*--------------------------------End default brick header code --------------------------*/

#include "ip.h"
struct fullTarget;
struct QREP dataRep;
anaLabel_t ip_recvLLIDP;
anaLabel_t ip_recvHLIDP;
anaLabel_t ethLabel;
anaLabel_t sumIDP = 0;
anaLabel_t fwdLabel = 0;
struct devAddr *dA;
//struct devAddr *dA_test;
struct devAddr *dA_start;

struct resolveRequestList *requestList;

anaLabel_t cfinder, regexp; /* for publication of IP regexp in compartment finder */

/* wrapper for L2 messages: added by Christophe for compartment finder code */
void anaL2_msg(struct anaL2_message *msg); /* simply calls ip_recvHLIDPrecvHLir() */

void ip_recvLL(char *data, int len, anaLabel_t input, void *aux);
void ip_recvHL(char *data, int len, anaLabel_t input, void *aux);
void makeHeaderAndSend(char *data, int len, anaLabel_t input, struct fullTarget *aux);

struct fullTarget *cloneFullTarget(struct fullTarget *ft);
void freeFullTarget(struct fullTarget *ft);
static void freeValue(void *value);
char * copy(char *aux);
anaLabel_t * cloneLabel_ipenc(anaLabel_t *l);
static void freeLabel_ipenc(anaLabel_t *l);


struct fullTarget *cloneFullTarget(struct fullTarget *ft) {
	struct fullTarget *ft2 = NULL;
	ft2 = (struct fullTarget *) malloc(sizeof(struct fullTarget));
	ft2->address = (char *) malloc(strlen(ft->address) + 1);
	memcpy(ft2->address, ft->address, strlen(ft->address) + 1);
	ft2->label = ft->label;
	if (ft->destDescrip!=NULL) {
		ft2->destDescrip = (char *) malloc(strlen(ft->destDescrip) + 1);
		memcpy(ft2->destDescrip, ft->destDescrip, strlen(ft->destDescrip) + 1);
	}
	else
		ft2->destDescrip = NULL;
	ft2->source = (char *) malloc(strlen(ft->source) + 1);
	memcpy(ft2->source, ft->source, strlen(ft->source) + 1);
	ft2->type = ft->type;
	if (ft->srcDescrip!=NULL) {
		ft2->srcDescrip = (char *) malloc(strlen(ft->srcDescrip) + 1);
		memcpy(ft2->srcDescrip, ft->srcDescrip, strlen(ft->srcDescrip) + 1);
	}
	else
		ft2->srcDescrip = NULL;
	return ft2;
}

void freeFullTarget(struct fullTarget *ft) {
	free(ft->address);
	free(ft->destDescrip);
	free(ft->source);
	free(ft->srcDescrip);
}

char *copy(char *aux) {
	char *copy;
	copy = (char *) malloc(strlen(aux) + 1);
	memcpy(copy, aux, strlen(aux) + 1);
	return copy;
}
//! A function to generate a next header code for the IP next header field.
    /*!
      \return New random next header code
    */
uint8_t  generateNextHeader(void) {
	uint8_t res;
	int exists = 1;
	while (exists) {
		res = (uint8_t) random();
		/* we check that the header is different */
		if (res!=0 && res!=IP_CMD  &&  getQREPEntryByName(&dataRep, (char *) &res, sizeof(uint8_t))==NULL)
			exists = 0;
	}
	return res;
}

/* called upon leaving ip_enc */
void  brick_exit() {
#ifdef boub
	xrpMsg_t msg;
	int ret = 0, len = 0;
	
        anaPrint(ANA_NOTICE, "IP_ENC HALTED\n");
	
	msg = anaL1_allocateMessage();
	/* unpublish the entries in the QREP of eth */	
	len = anaL1_encCompartmentUnpublish(msg, ip_recvHLIDP, (void *) "ip+internet+inet", strlen("ip+internet+inet") + 1);
	/* this is not correct! ethLabel was never initialized and we need support for multiple eth bricks */	
/*	ret = anaL0_send(ethLabel, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request for \"ip\"!\n");
	}
*/	
	/* unpublish the entries in the KVR */
/*	ret = anaL0_send(NODE_LABEL, msg, len);
	if (ret<0) {
		anaPrint(ANA_ERR, "brick_exit: problem sending unpublish request for \"ip\" in KVR!\n");
	}
	free(msg);
*/	
	anaL0_unregisterCallback(ip_recvLLIDP, NULL);
	anaL0_unregisterCallback(ip_recvHLIDP, NULL);
	freeQRepository(&dataRep, freeValue);
#endif
	/* remove compartment finder entry (if any) */
	if (cfinder && regexp)
	{
		struct service_s srv;
		struct context_s ctx;
		
		ctx.value = ".";
		ctx.valueLen = 2;
		srv.value = IPv4_REGEXP;
		srv.valueLen = strlen(srv.value)+1;
		
		anaL2_unpublish(cfinder, regexp, &ctx, &srv, NULL);
	}
}

void  dummy(char *data, int len, anaLabel_t input, anaLabel_t aux) {
	return;
}

/* will be called after a publish message was sent, checks whether an error occured */
void  handlePublishReplyConfig(char *data, int len, anaLabel_t input, anaLabel_t *aux) {
	char *name = NULL;
	xrpMsg_t replyMsg;
	int totalSize, arglen;
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		
		replyMsg = anaL1_allocateMessage();
		totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
		anaL0_send(*aux, replyMsg, totalSize);
		free(replyMsg);
	}
	else {
		if (name!=NULL)
			anaPrint(ANA_NOTICE, "Config: publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
		
		replyMsg = anaL1_allocateMessage();
		totalSize = fillXRPCommand(replyMsg, XRP_CMD_PUBLISH);
		arglen = addXRPArg(replyMsg, XRP_CLASS_NAME, name, strlen(name) + 1);
		if (arglen<0) {
			anaPrint(ANA_ERR, "handlePublishReplyConfig: problem encoding reply message!\n");
			free(replyMsg);
			return;
		}
		totalSize += arglen;
		anaPrint(ANA_NOTICE, "sent reply message mmmm %i %s %i\n", totalSize, replyMsg, *aux);
		if (anaL0_send(*(aux), replyMsg, totalSize) < 0)
			anaPrint(ANA_ERR, "could not send msg\n");
		anaPrint(ANA_NOTICE, "sent reply message %i\n", totalSize);
		free(replyMsg);
	}
}

void  updateResolveReply(char *ID, struct iphdr *ip_header, void *response, int responseLen) {
	struct resolveRequestList *request = NULL;
	struct resolveRequestList *predecessor = NULL;
	struct fullTarget *ft;
	int found = 0, totalLength, arglen, IDPLen, nhopLen;
	anaLabel_t dataIDP, IDP, *IDPptr, ethIDP;
	uint8_t *header;
	xrpMsg_t replyMsg = NULL, msg2;
	void *reply;
	char *nhop;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

	/* we check if the requestId is still in the queue */
	request = requestList;
	while (!found && request!=NULL) {
		if (!memcmp(ID, request->resId, RES_ID_SIZE))
			found = 1;
		else {
			predecessor = request;
			request = request->next;
		}
	}
	
	/* if not found means that the response comes too late (timer expired) */
	if (!found)
		return;
	
	/* now we stop the timer */
	anatimer_del(request->timerId);
	
	replyMsg = anaL1_allocateMessage();
	
	ft = (struct fullTarget *) malloc(sizeof(struct fullTarget));
	ft->address = (char *) malloc(strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1);
	memcpy(ft->address, hexToDottedQuad(htonl(ip_header->saddr)), strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1);
	ft->destDescrip = NULL;
	ft->source = (char *) malloc(strlen(hexToDottedQuad(htonl(ip_header->daddr))) + 1);
	memcpy(ft->source, hexToDottedQuad(htonl(ip_header->daddr)), strlen(hexToDottedQuad(htonl(ip_header->daddr))) + 1);
	
	/* get IDP of fwd brick */
	if (fwdLabel == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "ip_fwd";
		service.valueLen = strlen("ip_fwd")+1;
		fwdLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	}
	if (fwdLabel == 0) {
		anaPrint(ANA_ERR, "updateResolveReply: fwd brick not found!\n");
		return;
	}
	
	/* lookup for netnumber in fwd table */
	msg2 = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg2, XRP_CMD_DISCOVER);
	if (totalLength<0) {
		anaPrint(ANA_ERR, "updateResolveReply: problem encoding discover message!\n");
		free(msg2);
		return;
	}
	arglen = addXRPArg(msg2, XRP_CLASS_IPADDRESS, hexToDottedQuad(htonl(ip_header->saddr)), strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1);
	if (arglen<0) {
		anaPrint(ANA_ERR, "updateResolveReply: unable to add address to discover message!\n");
		free(msg2);
		return;
	}
	totalLength += arglen;
	reply = anaL2_requestReply(fwdLabel, msg2, totalLength, &totalLength, &timeout);
	free(msg2);
	if(reply == NULL) {
		anaPrint(ANA_ERR, "updateResolveReply: requestReply failed\n");
		return;
	}	
	IDPptr = (anaLabel_t *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_MESSAGE, &IDPLen, 0);
	nhop = (char *) getXRPArg((char *) reply, XRP_CLASS_SRC_CXT, &nhopLen, 0);
	IDP = *IDPptr;	/* since we use XRP_CLASS_MESSAGE it htonl was not called */
	if (IDP != 0) {
		/* create data channel */
		if (strcmp(nhop, "none")) {
			/* we are sending to a gateway */
			context.value = "*";
			context.valueLen = 2;
			service.value = nhop;
			service.valueLen = nhopLen;
			ethIDP = anaL2_resolve(IDP, &context, &service, 'u', &description, &timeout);
		}
		else {
			context.value = "*";
			context.valueLen = 2;
			service.value =  hexToDottedQuad(htonl(ip_header->saddr));
			service.valueLen = strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1;

			ethIDP = anaL2_resolve(IDP, &context, &service, 'u', &description, &timeout);
		}
		if (ethIDP == 0) {
			anaPrint(ANA_ERR, "%s not found!\n", hexToDottedQuad(htonl(ip_header->saddr)));
			return;
		}
		ft->label = ethIDP;
	}
	else
		return;
	header = (uint8_t *) response;
	if (*header!=0) {
		ft->type = *header;
		if (request->description!=NULL) {
			ft->srcDescrip= (void *) malloc(request->descLen);
			memcpy(ft->srcDescrip, request->description, request->descLen);
		}
		else
			ft->srcDescrip = NULL;
		
		/* we create an IDP mapped to makeHeaderAndSend with the right aux argument */
		if ((dataIDP = anaL0_registerCallback((anaCallback_t) &  makeHeaderAndSend, ft, 
			(copyAuxFunction) cloneFullTarget, (freeAuxFunction) freeFullTarget, 0, 0, THREAD)) < 0) {
			int totalSize;
			anaPrint(ANA_ERR, "updateResolveReply: unable to register an IDP for resolve response!\n");
			totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
            		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
           		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
			anaL0_send(request->replyTo, replyMsg, totalSize);
            		free(replyMsg);
		}
		else { /* we send a successfull response */
			int totalSize;
			
			anaPrint(ANA_NOTICE, "IDP for resolve response registered\n");
			totalSize = fillXRPCommand(replyMsg, XRP_CMD_RESOLVE);
            		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
           		totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
           		totalSize += addXRPArg(replyMsg, XRP_CLASS_LABEL, &dataIDP, sizeof(anaLabel_t));
			anaL0_send(request->replyTo, replyMsg, totalSize);
            		free(replyMsg);
		}
	}
	else {
		int totalSize;
		anaPrint(ANA_ERR, "updateResolveReply: no next header field in response!\n");
		totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
        	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
        	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
		anaL0_send(request->replyTo, replyMsg, totalSize);
        	free(replyMsg);
	}
	
	if (predecessor==NULL)
		requestList = request->next;
	else
		predecessor->next = request->next;
	free(request->context);
	free(request->target);
	free(request->description);
	free(request);
}

/* receive data function from eth */
void  ip_recvLL(char *data, int len, anaLabel_t input, void *aux) {
	struct iphdr *ip_header = NULL;
	int dataLen, newLen, offset, size, resultLen, totalLength, arglen, IDPLen, subnetLen, messageLen, myAddress = 0, nhopLen, msgLen;
	struct QREPKeyObjectList *res = NULL;
	void *msg = NULL, *result = NULL, *reply, *index;
	uint8_t *opCode, ip_rsp;
	uint16_t *chksum, newttl, check, newcheck, destDescLen, srcDescLen;
	uint32_t tmp, temp, daddr;
	anaLabel_t IDP, *IDPptr, ethIDP;
	xrpMsg_t chkMsg, msg2;
	char *subnet, *message_data, *nhop = NULL, *srcDesc, *sAddr = NULL;
 	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	struct devAddr *dA_test = dA_start;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 
	anaPrint(ANA_ERR, "received data on %i\n", input);
	if (data==NULL || len<0)
		return;
	
	/* message received is encoded as encData */
	messageLen = anaL1_decResponse(data, XRP_CLASS_MESSAGE, 0, (void **) &message_data);
	
	ip_header = (struct iphdr *) message_data;
	offset = sizeof(struct iphdr);
	anaPrint(ANA_ERR, "\n\n--> receivd from ethernet: %i bytes\n\n", len);
	
	anaPrint(ANA_NOTICE, "from %s to %s\n", hexToDottedQuad(htonl(ip_header->saddr)), hexToDottedQuad(htonl(ip_header->daddr)));
	
	/* first we check the checksum */
	if (sumIDP == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "checksum";
		service.valueLen = strlen("checksum")+1;
		sumIDP = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
	}
	if (sumIDP != 0) {
		chkMsg = anaL1_allocateMessage();
		size = anaL1_encData(chkMsg, ip_header, sizeof(struct iphdr),  NULL, NULL, 0);
		result = anaL2_requestReply(sumIDP, chkMsg, size, &size, &timeout);
		if(result == NULL) {
			anaPrint(ANA_DEBUG, "ip_recvLL requestReply failed\n");
			return;
		}
		chksum = (uint16_t *) getXRPArg((xrpMsg_t)result, XRP_CLASS_MESSAGE, &resultLen, 0);
		if (*chksum!=0) {
			anaPrint(ANA_ERR, "checksum wrong!\n");
			return;
		}
		else
			anaPrint(ANA_NOTICE, "checksum correct\n");
	}
	
	/* then, we check if we have to deliver local or forward */
	while (dA_test->addr!=NULL) {
		temp = dottedQuadToHex(dA_test->addr);
		daddr = htonl(ip_header->daddr);
		if (memcmp(&daddr, &temp, 4)==0 || (daddr | dottedQuadToHex(dA_test->mask))==0xFFFFFFFF) {
			myAddress = 1;
			break;
		}
		dA_test = dA_test->next;
	}
	dA_test = dA_start;
	tmp = dottedQuadToHex("255.255.255.255");
	
	if (!myAddress && memcmp(&ip_header->daddr, &tmp, 4)!=0) { /* forward */
		
		if (ip_header->saddr==0x00000000)
			return;
		
		anaPrint(ANA_NOTICE, "ip packet is not for me -> try to forward\n");

		/* get IDP of fwd brick */
		if (fwdLabel == 0) {
			context.value = ".";
			context.valueLen = 2;
			service.value = "ip_fwd";
			service.valueLen = strlen("ip_fwd")+1;
			fwdLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
		}
		if (fwdLabel == 0) {
			anaPrint(ANA_ERR, "ip_recvLL: fwd brick not found!\n");
			return; /* drop */
		}
		
		/* lookup for netnumber in fwd table */
		msg2 = anaL1_allocateMessage();
		totalLength = fillXRPCommand(msg2, XRP_CMD_DISCOVER);
		if (totalLength<0) {
			anaPrint(ANA_ERR, "ip_recvLL: problem encoding discover message!\n");
			free(msg);
			return;
		}
		arglen = addXRPArg(msg2, XRP_CLASS_IPADDRESS, hexToDottedQuad(htonl(ip_header->daddr)), strlen(hexToDottedQuad(ip_header->daddr)) + 1);
		if (arglen<0) {
			anaPrint(ANA_ERR, "ip_recvLL: unable to add address to discover message!\n");
			free(msg);
			return;
		}
		totalLength += arglen;
		reply = anaL2_requestReply(fwdLabel, msg2, totalLength, &totalLength, &timeout);
		free(msg2);
		if(reply == NULL) {
			anaPrint(ANA_DEBUG, "ip_recvLL requestReply failed (2)\n");
			return;
		}

		
		IDPptr = (anaLabel_t *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_MESSAGE, &IDPLen, 0);
		IDP = *IDPptr;
		subnet = (char *) getXRPArg((char *) reply, XRP_CLASS_SRC_SRV, &subnetLen, 0);
		nhop = (char *) getXRPArg((char *) reply, XRP_CLASS_SRC_CXT, &nhopLen, 0);
		
		/* resolve data channel */
		if (!strcmp(nhop, "none")) {
			context.value = "*";
			context.valueLen = 2;
			service.value = hexToDottedQuad(htonl(ip_header->daddr));
			service.valueLen =  strlen(hexToDottedQuad(ip_header->daddr)) + 1;
			ethIDP = anaL2_resolve(IDP, &context, &service, 'u', &description, &timeout);
			if (ethIDP != 0)
				anaPrint(ANA_NOTICE, "forward to %s\n", hexToDottedQuad(htonl(ip_header->daddr)));
			else {
				anaPrint(ANA_ERR, "%s not found!\n", hexToDottedQuad(htonl(ip_header->daddr)));
				return;
			}
		}
		else {
			context.value = "*";
			context.valueLen = 2;
			service.value =  nhop;
			service.valueLen =  strlen(nhop) + 1;
			ethIDP = anaL2_resolve(IDP, &context, &service, 'u', &description, &timeout);
			if (ethIDP != 0)
				anaPrint(ANA_NOTICE, "forward to %s\n", nhop);
			else {
				anaPrint(ANA_ERR, "%s not found!\n", nhop);
				return;
			}
		}
		
		/* decrease TTL */
		if (ip_header->ttl<=1) {
			anaPrint(ANA_ERR, "TTL is over!\n");
			return; /* drop */
		}
		else {
			newttl = ip_header->ttl - 1;
			memcpy(&ip_header->ttl, &newttl, 1);
			check = ntohs(ip_header->check) + 0x0100;
			if ((check & 0xFF00)==0)
				check++; /* carry overflow */
			newcheck = htons(check);
			memcpy(&ip_header->check, &newcheck, 2);
			anaPrint(ANA_NOTICE, "new TTL: %i\n", newttl);
			anaPrint(ANA_NOTICE, "new checksum: 0x%X\n", htons(newcheck));
		}
		
		/* send */
		anaL0_send(ethIDP, message_data, messageLen);
		anaPrint(ANA_NOTICE, "ip packet successfully forwarded\n");
		return;
		
	}
	
	switch (ip_header->protocol) {
		case IP_CMD: {
			opCode = (uint8_t *) message_data + offset;
			if (*opCode==IP_DLV) {
				uint16_t *tmpLen;  
				void *destDesc = NULL;
				anaPrint(ANA_NOTICE, "local deliver ip packet\n");
				
				offset += sizeof(uint8_t);
				
				tmpLen = (uint16_t *) (message_data + offset);
				destDescLen = ntohs(*tmpLen);
				offset += sizeof(uint16_t);
				destDesc = message_data + offset;
				offset += destDescLen;
				
				anaPrint(ANA_NOTICE, "target description: %s\n", (char *) destDesc);
				
				/* lookup the target description in the QREP */
				if (QREPfindByQuery(&dataRep, destDesc, &res)<0)
					anaPrint(ANA_ERR, "could not find an entry matching the target description!\n");
				if (res==NULL)
					return; /* no matching target */
				
				tmpLen = (uint16_t *) (message_data + offset);
				srcDescLen = ntohs(*tmpLen);
				offset += sizeof(uint16_t);
				srcDesc = message_data + offset;
				offset += srcDescLen;

				dataLen = messageLen - offset; /* length of the data contained in the received message */
	
				msg = anaL1_allocateMessage();
			
				context.value = hexToDottedQuad(htonl(ip_header->saddr));
				context.valueLen = strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1;
				service.value =  srcDesc;
				service.valueLen =  srcDescLen;
				msgLen=anaL1_encData(msg, message_data + offset, dataLen, &context, &service, 0);
				if (msgLen<0) {
					anaPrint(ANA_ERR, "ip_recvLL: problem encoding data message!\n");
					free(msg);
					return;
				}
				
				/* now we send the message to the target bricks: IMPORTANT / TO CHANGE: if many bricks fit this description, they would receive a copy as well! */	
				while (res != NULL) {
					anaPrint(ANA_DEBUG, "send msg to idp %i\n", *(anaLabel_t *)res->entry->value);
					anaL0_send(*(anaLabel_t*) res->entry->value, msg, msgLen);
					res = res->next;
				}
				QREPfreeQueryRes(res);
				free(msg);
			}
			if (*opCode==IP_RES) {
				char *resId = NULL;
				uint16_t *tmpLen;
				void *destDesc = NULL;
				
				offset += sizeof(uint8_t);
				
				tmpLen = (uint16_t *) (message_data + offset);
				destDescLen = ntohs(*tmpLen);
				offset += sizeof(uint16_t);
				destDesc = message_data + offset;
				offset += destDescLen;
			
				anaPrint(ANA_NOTICE, "received remote resolve request for %s\n", (char *) destDesc);
				
				/* lookup the target description in the QREP */
				if (QREPfindByQuery(&dataRep, destDesc, &res)<0)
					anaPrint(ANA_ERR, "could not find an entry matching the target description!\n");
				
				tmpLen = (uint16_t *) (message_data + offset);
				srcDescLen = ntohs(*tmpLen);
				offset += sizeof(uint16_t);
				if (srcDescLen) {
					srcDesc = message_data + offset;
				}
				else
					srcDesc = NULL;
				offset += srcDescLen;
				
				/* we need to find the resolve queryidentifier */
				resId  = message_data + offset;
		
				if (res==NULL) {	
					return;
				}
				else {
					struct iphdr tmphed;
					struct QREPKeyObjectList *tmp = NULL;
					int nbRes = 0;
					void *response = NULL;
					int offset = 0;
					uint8_t *tmpP = NULL;
					uint8_t tmpI;
					
					/* first we check how many matches we have */
					tmp = res;
					while (tmp!=NULL) {
						nbRes++;
						tmp=tmp->next;
					}
				
					/* then we allocate a data buffer to hold the requestId + nbRes*headers */
					response = (void *) malloc((sizeof(uint8_t)*nbRes) + RES_ID_SIZE);
					memcpy(response, resId, RES_ID_SIZE);
					offset += RES_ID_SIZE;
			
					while (res!=NULL) {
						tmpP = (uint8_t *) (res->entry->name);
						tmpI = htons(*tmpP);	
						tmpP = (uint8_t *) ((char *) response + offset);
						*tmpP = tmpI;
						memcpy((char *) response + offset, res->entry->name, sizeof(uint8_t));
						offset += sizeof(uint8_t);
						res = res->next; 
					}
			
					QREPfreeQueryRes(res);
					
					tmphed.daddr = ip_header->saddr;
					
					/* get IDP of fwd brick */
					if (fwdLabel == 0) {
						context.value = ".";
						context.valueLen = 2;
						service.value = "ip_fwd";
						service.valueLen = strlen("ip_fwd")+1;
						fwdLabel = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
					}
					if (fwdLabel != 0) {
						anaPrint(ANA_ERR, "ip_recvLL: fwd brick found\n");
						
						msg = anaL1_allocateMessage();
						totalLength = fillXRPCommand(msg, XRP_CMD_DISCOVER);
						if (totalLength<0) {
							anaPrint(ANA_ERR, "ip_recvLL: problem encoding discover message!\n");
							free(msg);
							return;
						}  
						arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, hexToDottedQuad(htonl(ip_header->saddr)), strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1);
						if (arglen<0) {
							anaPrint(ANA_ERR, "ip_recvLL: unable to add address to discover message!\n");
							free(msg);
							return;
						}
						totalLength += arglen;
						reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
						if (reply==NULL) {
							anaPrint(ANA_ERR, "ip_recvLL: requestReply failed (3)!\n");
							return;
						}
						IDPptr = (anaLabel_t*) getXRPArg((xrpMsg_t) reply, XRP_CLASS_MESSAGE, &IDPLen, 0);
						IDP = *IDPptr;
						nhop = (char *) getXRPArg((char *) reply, XRP_CLASS_SRC_CXT, &nhopLen, 0);
					}
					else {
						/* get IDP from eth01 */
						context.value = ".";
						context.valueLen = 2;
						service.value = "eth01";
						service.valueLen = strlen("eth01")+1;
						IDP = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
						anaPrint(ANA_ERR, "received label of eth01 %i\n", IDP); 

					}
					/* take corresponding source ip address */
					while (dA_test->dev != 0) {
						if (dA_test->dev == IDP) {
							sAddr = dA_test->addr;
							break;
						}
						dA_test = dA_test->next;
					}
					dA_test = dA_start;
					if (sAddr==NULL)
						sAddr="0.0.0.0";
					tmphed.saddr = htonl(dottedQuadToHex(sAddr));
					tmphed.version = 0x4; /* ipv4 */
					tmphed.ihl = 0x5;
					tmphed.tos = 0x00;
					tmphed.tot_len = htons(tmphed.ihl * 4 + sizeof(uint8_t)*nbRes + RES_ID_SIZE + sizeof(uint8_t));
					tmphed.id = 0x0000;
					tmphed.frag_off = 0x0040;
					tmphed.ttl = 0x40; /* default time to live 64 */
					tmphed.protocol = IP_CMD; /* next header */
					tmphed.check = 0x0000;
	
					/* computing checksum */
					if (sumIDP == 0) {
						context.value = ".";
						context.valueLen = 2;
						service.value = "checksum";
						service.valueLen = strlen("checksum")+1;
						sumIDP = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &description, &timeout);
					}
					if (sumIDP != 0) {
						chkMsg = anaL1_allocateMessage();
						size = anaL1_encData(chkMsg, &tmphed, sizeof(struct iphdr),  NULL, NULL, 0);
						result = anaL2_requestReply(sumIDP, chkMsg, size, &size, &timeout);
						if (result==NULL) {
							anaPrint(ANA_ERR, "ip_recvLL: requestReply failed (4)!\n");
							return;
						}
						chksum = (uint16_t *) getXRPArg((xrpMsg_t) result, XRP_CLASS_MESSAGE, &resultLen, 0);
						tmphed.check = htons(*chksum);
						anaPrint(ANA_NOTICE, "checksum successfully added: 0x%X\n", htons(tmphed.check));
					}
					
					newLen = sizeof(struct iphdr) + sizeof(uint8_t) + sizeof(uint8_t)*nbRes + RES_ID_SIZE;
					msg = (char *) malloc(newLen);
					index = msg;
					memcpy(index, &tmphed, sizeof(struct iphdr));
					index += sizeof(struct iphdr);
					ip_rsp = IP_RSP;
					memcpy(index, &ip_rsp, sizeof(uint8_t));
					index += sizeof(uint8_t);
					memcpy(index, response, sizeof(uint8_t)*nbRes + RES_ID_SIZE);
					index += sizeof(uint8_t)*nbRes + RES_ID_SIZE;
					
					/* create data channel */
					if (strcmp(nhop, "none")) {
						/* we are sending to a gateway */
						context.value = "*";
						context.valueLen = 2;
						service.value = nhop;
						service.valueLen = nhopLen;
						ethIDP = anaL2_resolve(IDP, &context, &service, 'u', &description, &timeout);
					}
					else {
						context.value = "*";
						context.valueLen = 2;
						service.value = hexToDottedQuad(htonl(ip_header->saddr));
						service.valueLen = strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1;

						ethIDP = anaL2_resolve(IDP, &context, &service, 'u', &description, &timeout);
					}
		
					if (ethIDP == 0) {
						anaPrint(ANA_ERR, "%s not found!\n", hexToDottedQuad(htonl(ip_header->saddr)));
						return;
					}
					
					anaL0_send(ethIDP, msg, newLen);
					anaPrint(ANA_NOTICE, "resolve reply sent to ethernet\n");
					free(msg);
					free(response);	
				}
			}
			if (*opCode==IP_RSP) { /* resolve response */
				char *resId;
				int responseLen;
				anaPrint(ANA_NOTICE, "received resolve response\n");
				offset += sizeof(uint8_t);
			
				/* lookup the target description in the QREP */
				resId  = message_data + offset;
				offset += RES_ID_SIZE;
				responseLen = messageLen - offset;
				
				updateResolveReply(resId, ip_header, message_data + offset, responseLen); 
			}
			break;
		}
		default: { /* here we use the next header field */
			struct QREPEntry *entry = NULL;
			uint16_t *tmp;
			uint16_t tmpI;
			anaLabel_t dest;
			anaPrint(ANA_NOTICE, "local deliver ip packet\n");
			tmp = (uint16_t *) (message_data + offset);
			srcDescLen = ntohs(*tmp);
			offset += sizeof(uint16_t);
			srcDesc = message_data + offset;
			offset += srcDescLen;
			
			tmpI = ip_header->protocol;
			anaPrint(ANA_NOTICE, "next header: 0x%X\n", tmpI);
			entry = getQREPEntryByName(&dataRep, (char*) &tmpI, sizeof(uint8_t));	
			
			if (entry==NULL) {
				anaPrint(ANA_ERR, "could not find an entry matching the target description!\n");
				return; /* no matching target */
			}
			dataLen = messageLen - offset; /* length of the data contained in the received message */
	
			msg = anaL1_allocateMessage();
			msgLen = 0;

			context.value = hexToDottedQuad(htonl(ip_header->saddr));
			context.valueLen = strlen(hexToDottedQuad(htonl(ip_header->saddr))) + 1;
			service.value =  srcDesc;
			service.valueLen =  srcDescLen;

			msgLen=anaL1_encData(msg, message_data + offset, dataLen, &context, &service, 0);
			if (msgLen<0) {
				anaPrint(ANA_ERR, "ip_recvLL: problem encoding data message for next header!\n");
				free(msg);
				return;
			}
			dest = *(anaLabel_t *)entry->value;
			anaPrint(ANA_ERR, "sending data to %i\n", dest);
			anaL0_send(dest, msg, msgLen);	
			free(msg);
		}
	}
}

/* this function is called from the higher level brick (e.g. ip_usr) to 
 * publish it's description and its IDP in the QREP of this Brick (e.g. "ip_usr01")
 * a confirmation might be sent back if there is an answerIDP supplied
 */
void  handlePublish(void *msg, int len) {
	anaLabel_t IDP, *IDPptr, *replyToptr = NULL;
	anaLabel_t replyTo;
        int labelLen, dummy, descLen;
     	char *description = NULL, *keyWord, *desc;
        struct QREPEntry *entry;
	struct QREP *rep = (struct QREP *) &dataRep;
	uint8_t nextHeader;

        IDPptr =  (anaLabel_t *) getXRPArg((xrpMsg_t) msg, XRP_CLASS_LABEL, &labelLen, 0);
 	description =  (char*) getXRPArg((xrpMsg_t) msg, XRP_CLASS_SRC_SRV, &descLen, 0);
	IDP = *IDPptr;
	desc = description;
	
	if (description == 0 || IDP == 0) {
                anaPrint(ANA_ERR, "handlePublish: unable to identify name or label!\n");
                return;
        }

        /* the replyTo label is optional, we look for an argument with class label
 	 * at the second occurence => the first label is the one associated to the entry
 	 */
        replyToptr = (anaLabel_t *)getXRPArg((xrpMsg_t) msg, XRP_CLASS_LABEL, &dummy, 1);
 	replyTo = *replyToptr; 
	/* we generate a unique next header field for this service (that will be stored as its name in QREP) */
	nextHeader = generateNextHeader();
	
	keyWord = strsep(&description, "+");

	if ((keyWord==NULL) || (entry = allocateQREPEntry((char *) &nextHeader, sizeof(uint8_t), &IDP, labelLen))==NULL) {
	        /* we send an error message, if a replyTo label was specified */
                if (replyToptr != NULL) {
    			fillXRPCommand((xrpMsg_t) msg, XRP_CMD_ERROR);
                        anaL0_send(replyTo, msg, len);
                }
                return;
        }

        while (keyWord!= NULL) {
		QREPaddKeywordToEntry(entry, keyWord);
                keyWord = strsep(&description, "+");
        }
	
        if (addQREPEntry(rep, entry)<0) {
                if (replyToptr != NULL) {
                        fillXRPCommand((xrpMsg_t) msg, XRP_CMD_ERROR);
			anaL0_send(replyTo, msg, len);
		}
 		freeQREPEntry(rep, entry, freeValue);
        }
 	else { /* everything went well */
		if (replyToptr != NULL) { /* as a confirmation we send the same received message */
                        anaL0_send(replyTo, msg, len);
			anaPrint(ANA_NOTICE, "publish of %s successful\n", desc);
                }
        }
}

/* this function is called upon an unpublish request,
 * looks for an entry with the name given in data and frees it
 */
void  handleUnpublish(char *data, int len) {
	char *query;
	anaLabel_t *labelptr, *replyToptr = NULL;
	anaLabel_t label = 0, replyTo = 0;
	int lenQuery, lenLabel, dummy;
	struct QREPKeyObjectList *res = NULL;
	struct QREPKeyObjectList *tmp = NULL;
	
	query =  (char *) getXRPArg(data, XRP_CLASS_SRC_SRV, &lenQuery, 0);
	labelptr =  (anaLabel_t *) getXRPArg(data, XRP_CLASS_LABEL, &lenLabel, 0);
	replyToptr =  (anaLabel_t *)getXRPArg(data, XRP_CLASS_LABEL, &dummy, 1);
	
	anaPrint(ANA_NOTICE, "unpublish description: %s\n", query);
	
	if (query == NULL) {
		anaPrint(ANA_ERR, "no query specified\n");
		return;
	}

	if (labelptr == NULL)
		anaPrint(ANA_ERR, "handleUnpublish: no label!\n");
	
	if (replyToptr == NULL)
		anaPrint(ANA_ERR, "handleUnpublish: no reply label!\n");
	
	if(labelptr != NULL)	
		label = *labelptr;
	if(replyToptr != NULL)
		replyTo = *replyToptr;

	if (QREPfindByQuery(&dataRep, query, &res)<0) {
		anaPrint(ANA_ERR, "handleUnpublish: could not find an entry matching the target description!\n");
		if (replyToptr != NULL) {
			/* we send an error message */
			fillXRPCommand(data, XRP_CMD_ERROR);
			anaL0_send(replyTo, data, len);
		}		
		return;
	}
	
	tmp = res;
	while (tmp!=NULL) {
		if (label == *(anaLabel_t *)tmp->entry->value) {
			if (freeQREPEntry(&dataRep, tmp->entry, freeValue)==0) {
				anaPrint(ANA_NOTICE, "unpublish of %s successful\n", query);
				if (replyToptr != NULL) {	
					/* we send a confirmation message: same one received */
					anaL0_send(replyTo, data, len);
				}
			}
			else {
				if (replyToptr != NULL) {	
					/* we send an error message: same one received but with XRP_CMD_ERROR */
					fillXRPCommand(data, XRP_CMD_ERROR);
					anaL0_send(replyTo, data, len);
				}
			}
			QREPfreeQueryRes(res);
			return;
		}
		tmp = tmp->next;
	}
	
	/* if we arrive here, means no entry with same IDP + descript */	
	anaPrint(ANA_ERR, "handleUnpublish: could not find an entry matching the target description and label!\n");
	QREPfreeQueryRes(res);
	
	if (replyToptr != NULL) {	
		/* we send an error message: same one received but with XRP_CMD_ERROR */
		fillXRPCommand(data, XRP_CMD_ERROR);
		anaL0_send(replyTo, data, len);
	}
	return;
}

void  checkResolveReply(char *ID) {
	struct resolveRequestList *request = NULL;
	struct resolveRequestList *predecessor = NULL;
	int found=0, totalSize=0;
	xrpMsg_t replyMsg=NULL;
	
	/* we check if the requestId is still in the queue */
	request = requestList;
	while (!found && request!=NULL) {
		if (!memcmp(ID, request->resId, RES_ID_SIZE))
			found=1;
		else {
			predecessor = request;
			request = request->next;
		}
	}

	/* if not found means that the request was answered */
	if (!found) {
		anaPrint(ANA_ERR, "checkResolveReply: received answer\n");
		return;
	}
	anaPrint(ANA_ERR, "checkResolveReply: answer not received\n");
	
	replyMsg = anaL1_allocateMessage();
	
	/* we send an error message to the querier */
	totalSize = fillXRPCommand(replyMsg, XRP_CMD_ERROR);
     	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_CXT, request->context, request->contextLen);
     	totalSize += addXRPArg(replyMsg, XRP_CLASS_DST_SRV, request->target, request->targetLen);
    					
	anaL0_send(request->replyTo, replyMsg, totalSize);
	free(replyMsg);
	
	/* now we free the request and unqueue it */
	if (predecessor==NULL)
		requestList = request;
	else
		predecessor->next = request->next;

	free(request->context);
	free(request->target);
	free(request->description);
	free(request);
}

int  sendResolveRequest(char *context, void *target, int targetLen, 
			void *desc, int descLen, char chanType, anaLabel_t replyTo) {
	struct resolveRequestList *resRequest = NULL;	
	int timerId = -1, i, totalLength, arglen, contextLen, IDPLen, subnetLen, nhopLen, size, resultLen, newLen;
	anaLabel_t IDP, ethIDP;
	anaLabel_t *IDPptr = NULL;
	char *subnet, *nhop = NULL, *sAddr = NULL, *index = 0;
	xrpMsg_t msg, chkMsg;
	void *reply, *result;
	uint32_t localb = 0;
	uint16_t *chksum, *tmp;
	uint8_t ip_res;
	struct iphdr tmphed;
	struct devAddr *dA_test = dA_start;
	struct timespec timeout;
	struct context_s mycontext;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&mycontext, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 4; 

	contextLen = strlen(context) + 1;

	/* if it is not a broadcast */	
	if (strcmp(context, "255.255.255.255") && strcmp(context, "*")) {
		 /* get IDP of fwd brick */
		if (fwdLabel == 0) {
				mycontext.value = ".";
				mycontext.valueLen = 2;
				service.value = "ip_fwd";
				service.valueLen = strlen("ip_fwd")+1;
				fwdLabel = anaL2_resolve(NODE_LABEL, &mycontext, &service, 'u', &description, &timeout);
		}
		if (fwdLabel == 0) {
			anaPrint(ANA_NOTICE, "sendResolveRequest: fwd brick not found!\n");
			mycontext.value = ".";
			mycontext.valueLen = 2;
			service.value = "eth01";
			service.valueLen = strlen("eth01")+1;
			IDP = anaL2_resolve(NODE_LABEL, &mycontext, &service, 'u', &description, &timeout);
			anaPrint(ANA_ERR, "received IPD\n");
		}
		else { /* lookup for netnumber in fwd table */
			msg = anaL1_allocateMessage();
			totalLength = fillXRPCommand(msg, XRP_CMD_DISCOVER);
			if (totalLength<0) {
				anaPrint(ANA_ERR, "sendResolveRequest: problem encoding discover message!\n");
				free(msg);
				return -1;
			}  
			arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, context, contextLen);
			if (arglen<0) {
				anaPrint(ANA_ERR, "sendResolveRequest: unable to add address to discover message!\n");
				free(msg);
				return -1;
			}
			totalLength += arglen;
			reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
			if (reply==NULL) {
				anaPrint(ANA_ERR, "sendResolveRequest: requestReply failed\n");
				return -1;
			}
			IDPptr = (anaLabel_t *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_MESSAGE, &IDPLen, 0);
			IDP = *IDPptr;
			subnet = (char *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_SRC_SRV, &subnetLen, 0);
			nhop = (char *) getXRPArg((char *) reply, XRP_CLASS_SRC_CXT, &nhopLen, 0);
		}
		anaPrint(ANA_ERR, "2 received IPD\n");
		
		/* take corresponding source ip address */
		while (dA_test->dev != 0) {
			if (dA_test->dev == IDP) {
				sAddr = dA_test->addr;
				localb = dottedQuadToHex(dA_test->mask) | dottedQuadToHex(context);
				break;
			}
			dA_test = dA_test->next;
		}
		anaPrint(ANA_ERR, "3 received IPD\n");

		dA_test = dA_start;
		if (sAddr==NULL)
			sAddr="0.0.0.0";
		anaPrint(ANA_ERR, "4 received IPD\n");
		
		/* create data channel for ip user */
		if (localb == 0xFFFFFFFF) { /* local broadcast */
			anaPrint(ANA_NOTICE, "local broadcast\n");
			mycontext.value = "*";
			mycontext.valueLen = 2;
			service.value = "ip";
			service.valueLen = strlen("ip")+1;

			ethIDP = anaL2_resolve(IDP, &mycontext, &service, 'b', &description, &timeout);
		}
		else if ((nhop != NULL) && strcmp(nhop, "none")) {
			/* we are sending to a router */
			anaPrint(ANA_ERR, "send to router\n");
			mycontext.value = "*";
			mycontext.valueLen = 2;
			service.value = nhop;
			service.valueLen = nhopLen;
			ethIDP = anaL2_resolve(IDP, &mycontext, &service, 'u', &description, &timeout);
		}
		else {
			mycontext.value = "*";
			mycontext.valueLen = 2;
			service.value =  context;
			service.valueLen = contextLen;
			ethIDP = anaL2_resolve(IDP, &mycontext, &service, 'u', &description, &timeout);
			anaPrint(ANA_ERR, "end\n");
		}
		
		if (ethIDP == 0) {
			anaPrint(ANA_ERR, "--%s not found!\n", context);
			return -1;
		}
	
		tmphed.daddr = htonl(dottedQuadToHex(context));
		tmphed.saddr = htonl(dottedQuadToHex(sAddr));
		tmphed.version = 0x4; /* ipv4 */
		tmphed.ihl = 0x5;
		tmphed.tos = 0x00;
		tmphed.tot_len = htons(tmphed.ihl * 4 + sizeof(uint8_t) + targetLen + descLen + RES_ID_SIZE + 2 * sizeof(uint16_t));
		tmphed.id = 0x0000;
		tmphed.frag_off = 0x0040;
		tmphed.ttl = 0x40; /* default time to live 64 */
		tmphed.protocol = IP_CMD; /* next header */
		tmphed.check = 0x0000;
	
		/* computing checksum */
		if (sumIDP == 0) {
			mycontext.value = ".";
			mycontext.valueLen = 2;
			service.value = "checksum";
			service.valueLen = strlen("checksum")+1;
			sumIDP = anaL2_resolve(NODE_LABEL, &mycontext, &service, 'u', &description, &timeout);
		}
		if (sumIDP != 0) {
			chkMsg = anaL1_allocateMessage();
			size = anaL1_encData(chkMsg, &tmphed, sizeof(struct iphdr),  NULL, NULL, 0);
			result = anaL2_requestReply(sumIDP, chkMsg, size, &size, &timeout);
			if (result==NULL) {
				anaPrint(ANA_ERR, "sendResolveRequest: requestReply failed (2)\n");
				return -1;
			}

			chksum = (uint16_t *) getXRPArg((xrpMsg_t) result, XRP_CLASS_MESSAGE, &resultLen, 0);
			tmphed.check = htons(*chksum);
			anaPrint(ANA_NOTICE, "checksum successfully added: 0x%X\n", htons(tmphed.check));
		}
		
		resRequest = (struct resolveRequestList *) malloc(sizeof(struct resolveRequestList));
		for (i=0; i<RES_ID_SIZE; i++)
			resRequest->resId[i] = (int) random()%256;
		resRequest->replyTo = replyTo;
		resRequest->context = (char *) malloc(strlen(context) + 1);
		memcpy(resRequest->context, context, strlen(context) + 1);
		resRequest->target = (char *) malloc(targetLen);
		memcpy(resRequest->target, target, targetLen);
		resRequest->targetLen = targetLen;
		resRequest->contextLen = contextLen;
		if (desc!=NULL) {
			resRequest->description = (void*) malloc(descLen);
			memcpy(resRequest->description, desc, descLen);
			resRequest->descLen = descLen;
		}
		else {
			resRequest->description = NULL;
			resRequest->descLen = 0;
		}
		resRequest->chanType = chanType;
		resRequest->next = requestList;
		requestList = resRequest;	
	
		timerId = anatimer_add(RESOLVE_TIMEOUT, 0, 0, (fct_t) &  checkResolveReply, resRequest->resId, RES_ID_SIZE);
		if (timerId==-1)
			anaPrint(ANA_ERR, "sendResolveRequest: problem adding timer!\n");
		resRequest->timerId = timerId;
		
		newLen = sizeof(uint8_t) + sizeof(struct iphdr) + targetLen + descLen + RES_ID_SIZE + 2 * sizeof(uint16_t);
		msg = (char *) malloc(newLen);
		index = msg;
		memcpy(index, &tmphed, sizeof(struct iphdr));
		index += sizeof(struct iphdr);
		ip_res = IP_RES;
		memcpy(index, &ip_res, sizeof(uint8_t));
		index += sizeof(uint8_t);
		tmp = (uint16_t *) index;
		*tmp = htons(targetLen);
		index += sizeof(uint16_t);
		memcpy(index, target, targetLen);
		index += targetLen;
		tmp = (uint16_t *) index;
		*tmp = htons(descLen);
		index += sizeof(uint16_t);
		if (descLen) {
			memcpy(index, desc, descLen);
			index += descLen;
		}
		memcpy(index, resRequest->resId, RES_ID_SIZE);
		index += RES_ID_SIZE;
	
		anaL0_send(ethIDP, msg, newLen);
		anaPrint(ANA_NOTICE, "resolve request sent to ethernet\n");
		free(msg);
		return 0;
	}
	else {
		anaPrint(ANA_NOTICE, "broadcast on all interfaces\n");
		context = "255.255.255.255";
		contextLen = strlen("255.255.255.255") + 1;
		resRequest = (struct resolveRequestList *) malloc(sizeof(struct resolveRequestList));
		for (i=0; i<RES_ID_SIZE; i++)
			resRequest->resId[i] = (int) random()%256;
		resRequest->replyTo = replyTo;
		resRequest->context = (char *) malloc(contextLen);
		memcpy(resRequest->context, context, contextLen);
		resRequest->target = (char *) malloc(targetLen);
		memcpy(resRequest->target, target, targetLen);
		resRequest->targetLen = targetLen;
		resRequest->contextLen = contextLen;
		if (desc!=NULL) {
			resRequest->description = (void*) malloc(descLen);
			memcpy(resRequest->description, desc, descLen);
			resRequest->descLen = descLen;
		}
		else {
			resRequest->description = NULL;
			resRequest->descLen = 0;
		}
		resRequest->chanType = chanType;
		resRequest->next = requestList;
		requestList = resRequest;	
	
		timerId = anatimer_add(RESOLVE_TIMEOUT, 0, 0, (fct_t) &  checkResolveReply, resRequest->resId, RES_ID_SIZE);
		if (timerId==-1)
			anaPrint(ANA_ERR, "sendResolveRequest: problem adding timer!\n");
		resRequest->timerId = timerId;
		
		while (dA_test->dev != 0) {
			
			/* take corresponding source ip address */
			sAddr = dA_test->addr;
			if (sAddr==NULL)
				sAddr="0.0.0.0";
			/* create data channel for ip user */
			mycontext.value = ".";
			mycontext.valueLen = 2;
			service.value = "ip";
			service.valueLen = strlen("ip")+1;
			ethIDP = anaL2_resolve(dA_test->dev, &mycontext, &service, 'b', &description, &timeout);
			if (ethIDP == 0)
				break;

			tmphed.daddr = htonl(dottedQuadToHex("255.255.255.255"));
			tmphed.saddr = htonl(dottedQuadToHex(sAddr));
			tmphed.version = 0x4; /* ipv4 */
			tmphed.ihl = 0x5;
			tmphed.tos = 0x00;
			tmphed.tot_len = htons(tmphed.ihl * 4 + sizeof(uint8_t) + targetLen + descLen + RES_ID_SIZE + 2 * sizeof(uint16_t));
			tmphed.id = 0x0000;
			tmphed.frag_off = 0x0040;
			tmphed.ttl = 0x40; /* default time to live 64 */
			tmphed.protocol = IP_CMD; /* next header */
			tmphed.check = 0x0000;
	
			/* computing checksum */
			if (sumIDP == 0){
				mycontext.value = ".";
				mycontext.valueLen = 2;
				service.value = "checksum";
				service.valueLen = strlen("checksum")+1;
				sumIDP = anaL2_resolve(NODE_LABEL, &mycontext, &service, 'u', &description, &timeout);
			}
			if (sumIDP != 0) {
				chkMsg = anaL1_allocateMessage();
				size = anaL1_encData(chkMsg, &tmphed, sizeof(struct iphdr),  NULL, NULL, 0);
				result = anaL2_requestReply(sumIDP, chkMsg, size, &size, &timeout);
				if (result==NULL) {
					anaPrint(ANA_ERR, "sendResolveRequest: requestReply failed (2)\n");
					return -1;
			}

				chksum = (uint16_t *) getXRPArg((xrpMsg_t) result, XRP_CLASS_MESSAGE, &resultLen, 0);
				tmphed.check = htons(*chksum);
				anaPrint(ANA_NOTICE, "checksum successfully added: 0x%X\n", htons(tmphed.check));
			}
			
			newLen = sizeof(uint8_t) + sizeof(struct iphdr) + targetLen + descLen + RES_ID_SIZE + 2 * sizeof(uint16_t);
			msg = (char *) malloc(newLen);
			index = msg;
			memcpy(index, &tmphed, sizeof(struct iphdr));
			index += sizeof(struct iphdr);
			ip_res = IP_RES;
			memcpy(index, &ip_res, sizeof(uint8_t));
			index += sizeof(uint8_t);
			tmp = (uint16_t *) index;
			*tmp = htons(targetLen);
			index += sizeof(uint16_t);
			memcpy(index, target, targetLen);
			index += targetLen;
			tmp = (uint16_t *) index;
			*tmp = htons(descLen);
			index += sizeof(uint16_t);
			if (descLen) {
				memcpy(index, desc, descLen);
				index += descLen;
			}
			memcpy(index, resRequest->resId, RES_ID_SIZE);
			index += RES_ID_SIZE;
	
			anaL0_send(ethIDP, msg, newLen);
			anaPrint(ANA_NOTICE, "resolve request sent to ethernet\n");
			free(msg);
			
			dA_test = dA_test->next;
		}
		dA_test = dA_start;
		return 0;
	}
	return -1;
}

void nodeLocal(char *data, int len, anaLabel_t input, struct fullTarget *aux) {
	int srcDescLen, msgLen;
	xrpMsg_t msg;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));

	anaPrint(ANA_NOTICE, "\n\n--> received %i bytes from myself to myself\n\n", len);
	anaPrint(ANA_NOTICE, "destination description: %s\n", aux->destDescrip);
	anaPrint(ANA_NOTICE, "source description: %s\n", aux->srcDescrip);
	
	if (aux->srcDescrip!=NULL)
		srcDescLen = strlen(aux->srcDescrip) + 1;
	else {
		srcDescLen = 0;
		aux->srcDescrip = "dummy";
	}
	
	msg = anaL1_allocateMessage();
	context.value = aux->source;
	context.valueLen = strlen(aux->source) + 1;
	service.value =  aux->srcDescrip;
	service.valueLen =  srcDescLen;
	msgLen = anaL1_encData(msg, data, len, &context, &service, 0);
	if (msgLen<0) {
		anaPrint(ANA_ERR, "nodeLocal: problem encoding data message!\n");
		free(msg);
		return;
	}
	anaL0_send(aux->label, msg, msgLen);	
	free(msg);

}

void  handleResolve(void *data, int len) {
	xrpMsg_t msg;
	struct fullTarget *ft;
	/* replyTo needs to be a pointer, since we receive from the xrpMsg */
	anaLabel_t *replyToptr, *IDPptr = NULL;
	anaLabel_t replyTo;
	anaLabel_t IDP = 0, userIDP = 0, ethIDP = 0, broadcastIDP;
	int labelLen, targetLen, IDPLen, totalLength, arglen, subnetLen, contextLen, ret, nhopLen, own, descLen, chanLen;
	char *target = NULL, *subnet, *context, *sAddr = NULL, *nhop = NULL, *description, *chanType;
	void *reply = NULL;
	uint32_t localb = 0;
	struct QREPKeyObjectList *res = NULL;
	struct devAddr *dA_test = dA_start;
	struct timespec timeout;
	struct context_s mycontext;
	struct service_s service;
	struct service_s mydescription;
	memset(&service, 0, sizeof(struct service_s));
	memset(&mycontext, 0, sizeof(struct context_s));
	memset(&mydescription, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

	if (data==NULL || len<0)
       		return;
	
	replyToptr = (anaLabel_t *)getXRPArg((xrpMsg_t)data, XRP_CLASS_LABEL, &labelLen, 0);
	if (replyToptr == NULL ) { 
                anaPrint(ANA_ERR, "handleResolve: no reply label!\n");
                return;
        }
	replyTo = *replyToptr;
	target = (char *) getXRPArg((char *) data, XRP_CLASS_DST_SRV, &targetLen, 0);
	context = (char *) getXRPArg((char *) data, XRP_CLASS_DST_CXT, &contextLen, 0);
	description = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_SRC_SRV, &descLen, 0);
	chanType = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_CHANTYPE, &chanLen, 0);
	
	if (target==NULL || context==NULL)
		return; 
	
	anaPrint(ANA_NOTICE, "destination address: %s\n", context);
	anaPrint(ANA_NOTICE, "destination description: %s\n", target);
	
	/* node local */
	own = 0;
	if (!strcmp(context, "127.0.0.1") || !strcmp(context, "localhost") || !strcmp(context, ".")) {
		own = 1;
	}
	else {
		while (dA_test->dev != 0) {
			if (!strcmp(context, dA_test->addr)) {
				own = 1;
				break;
			}
			dA_test = dA_test->next;
		}
		dA_test = dA_start;
	}
	if (own==1) {
		anaPrint(ANA_NOTICE, "resolve for localhost\n");
		if (QREPfindByQuery(&dataRep, target, &res)<0) {
			anaPrint(ANA_ERR, "could not find an entry matching the target description!\n");
			fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
                        anaL0_send(replyTo, data, len);
			return;		
		}
		else {
			ft = (struct fullTarget *) malloc(sizeof(struct fullTarget));
			ft->address = (char *) malloc(contextLen);
			memcpy(ft->address, context, contextLen);
			ft->label = *(anaLabel_t *)res->entry->value;
			ft->destDescrip = (char *) malloc(targetLen);
			memcpy(ft->destDescrip, target, targetLen);
			ft->source = (char *) malloc(strlen("localhost") + 1);
			memcpy(ft->source, "localhost", strlen("localhost") + 1);
			ft->type = 0;
			if (description!=NULL) {
				ft->srcDescrip = (char *) malloc(strlen(description) + 1);
				memcpy(ft->srcDescrip, description, strlen(description) + 1);
			}
			else
				ft->srcDescrip = NULL;
			userIDP = anaL0_registerCallback((anaCallback_t) &  nodeLocal, ft, 
							(copyAuxFunction) cloneFullTarget, (freeAuxFunction) freeFullTarget, 
							0, 0, THREAD);
			if (userIDP<0) {
				anaPrint(ANA_ERR, "handleResolve: unable to create an IDP!\n");
				fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
				anaL0_send(replyTo, data, len);
				return;
			}
			anaPrint(ANA_NOTICE, "data channel sent to user 1\n");
			*replyToptr = userIDP;
			anaL0_send(replyTo, data, len);					
			QREPfreeQueryRes(res);	
			return;	
		}
	}
	if (chanType[0]=='m' || chanType[0]=='b') { /* in this case we use deliver to send the message */
		if (strcmp(context, "255.255.255.255") && strcmp(context, "*")) {
			/* get IDP of fwd brick */
			if (fwdLabel == 0){
				mycontext.value = ".";
				mycontext.valueLen = 2;
				service.value = "ip_fwd";
				service.valueLen = strlen("ip_fwd")+1;
				fwdLabel = anaL2_resolve(NODE_LABEL, &mycontext, &service, 'u', &mydescription, &timeout);
			}
			if (fwdLabel == 0) {
				anaPrint(ANA_ERR, "handleResolve: fwd brick not found!\n");
				mycontext.value = ".";
				mycontext.valueLen = 2;
				service.value = "eth01";
				service.valueLen = strlen("eth01")+1;
				IDP = anaL2_resolve(NODE_LABEL, &mycontext, &service, 'u', &mydescription, &timeout);
			}
			else { /* lookup for netnumber in fwd table */
				msg = anaL1_allocateMessage();
				totalLength = fillXRPCommand(msg, XRP_CMD_DISCOVER);
				if (totalLength<0) {
					anaPrint(ANA_ERR, "handleResolve: problem encoding discover message!\n");
					free(msg);
					fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
					anaL0_send(replyTo, data, len);
					return;
				}  
				arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, context, contextLen);
				if (arglen<0) {
					anaPrint(ANA_ERR, "handleResolve: unable to add address to discover message!\n");
					free(msg);
					fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
					anaL0_send(replyTo, data, len);
					return;
				}
				totalLength += arglen;
				reply = anaL2_requestReply(fwdLabel, msg, totalLength, &totalLength, &timeout);
				if (reply == 0) {
					fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
					anaL0_send(replyTo, data, len);
					return;
				}
				IDPptr = (anaLabel_t *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_MESSAGE, &IDPLen, 0);
				IDP = *IDPptr;
				subnet = (char *) getXRPArg((xrpMsg_t) reply, XRP_CLASS_SRC_SRV, &subnetLen, 0);
				nhop = (char *) getXRPArg((char *) reply, XRP_CLASS_SRC_CXT, &nhopLen, 0);
			}
		
			/* take corresponding source ip address */
			while (dA_test->dev != 0) {
				if (dA_test->dev == IDP) {
					sAddr = dA_test->addr;
					localb = dottedQuadToHex(dA_test->mask) | dottedQuadToHex(context);
					break;
				}
				dA_test = dA_test->next;
			}
			dA_test = dA_start;
			if (sAddr==NULL)
				sAddr="0.0.0.0";
		
			/* create data channel for ip user */
			if (localb==0xFFFFFFFF) { /* local broadcast */
				anaPrint(ANA_NOTICE, "local broadcast\n");
				mycontext.value = "*";
				mycontext.valueLen = 2;
				service.value = "ip";
				service.valueLen = strlen("ip") + 1;
				ethIDP = anaL2_resolve(IDP, &mycontext, &service, 'b', &mydescription, &timeout);
			}
			else if (strcmp(nhop, "none")) {
				/* we are sending to a gateway */
				mycontext.value = "*";
				mycontext.valueLen = 2;
				service.value = nhop;
				service.valueLen = nhopLen;
				ethIDP = anaL2_resolve(IDP, &mycontext, &service, 'b', &mydescription, &timeout);
			}
			else {
				mycontext.value = "*";
				mycontext.valueLen = 2;
				service.value = context;	/*Ariane: I hope this variable is missnamed! */
				service.valueLen = contextLen;
				ethIDP = anaL2_resolve(IDP, &mycontext, &service, 'b', &mydescription, &timeout);
			}
		
			ft = (struct fullTarget *) malloc(sizeof(struct fullTarget));
			ft->address = (char *) malloc(contextLen);
			memcpy(ft->address, context, contextLen);
			if (ethIDP != 0)
				ft->label = ethIDP;
			else {
				anaPrint(ANA_ERR, "%s not found!\n", context);
				fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
				anaL0_send(replyTo, data, len);
				return;
			}
			ft->destDescrip = (char *) malloc(targetLen);
			memcpy(ft->destDescrip, target, targetLen);
			ft->source = (char *) malloc(strlen(sAddr) + 1);
			memcpy(ft->source, sAddr, strlen(sAddr) + 1);
			ft->type = IP_CMD;
			if (description!=NULL) {
				ft->srcDescrip = (char *) malloc(strlen(description) + 1);
				memcpy(ft->srcDescrip, description, strlen(description) + 1);
			}
			else
				ft->srcDescrip = NULL;
	
			userIDP = anaL0_registerCallback((anaCallback_t) &  makeHeaderAndSend, ft, 
				(copyAuxFunction) cloneFullTarget, (freeAuxFunction) freeFullTarget, 
				0, 0, THREAD);
			if (userIDP == 0) {
				anaPrint(ANA_ERR, "handleResolve: unable to create an IDP!\n");
				fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
				anaL0_send(replyTo, data, len);
				return;
			}
			anaPrint(ANA_NOTICE, "data channel sent to user 2\n");
			*replyToptr = userIDP; 
			anaL0_send(replyTo, data, len);
			return;
		}
		else { /* we have a global broadcast */
			context = "255.255.255.255";
			contextLen = strlen("255.255.255.255") + 1;
			anaPrint(ANA_NOTICE, "broadcast on all interfaces\n");
			userIDP = anaL0_registerCallback((anaCallback_t) &  dummy, NULL, NULL, NULL, 0, 0, NOTHREAD);
			if (userIDP == 0) {
				anaPrint(ANA_ERR, "handleResolve: unable to create an IDP!\n");
				fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
				anaL0_send(replyTo, data, len);
				return;
			}
			while (dA_test->dev != 0) {
				int ret = -123;
				/* take corresponding source ip address */
				sAddr = dA_test->addr;
				if (sAddr==NULL)
					sAddr="0.0.0.0";
				/* create data channel for ip user */

				mycontext.value = "*";
				mycontext.valueLen = 2;
				service.value = "ip";
				service.valueLen = strlen("ip")+1;
				ethIDP = anaL2_resolve(dA_test->dev,  &mycontext, &service, 'b', &mydescription, &timeout);
		
				ft = (struct fullTarget *) malloc(sizeof(struct fullTarget));
				ft->address = (char *) malloc(contextLen);
				memcpy(ft->address, context, contextLen);
				if (ethIDP != 0)
					ft->label = ethIDP;
				else {
					anaPrint(ANA_ERR, "handleResolve: unable to create a broadcast IDP!\n");
					fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
					anaL0_send(replyTo, data, len);
					return;
				}
				ft->destDescrip = (char *) malloc(targetLen);
				memcpy(ft->destDescrip, target, targetLen);
				ft->source = (char *) malloc(strlen(sAddr) + 1);
				memcpy(ft->source, sAddr, strlen(sAddr) + 1);
				ft->type = IP_CMD;
				if (description!=NULL) {
					ft->srcDescrip = (char *) malloc(strlen(description) + 1);
					memcpy(ft->srcDescrip, description, strlen(description) + 1);
				}
				else
					ft->srcDescrip = NULL;
	
				broadcastIDP = anaL0_registerCallback((anaCallback_t) &  makeHeaderAndSend, ft, 
						(copyAuxFunction) cloneFullTarget, (freeAuxFunction) freeFullTarget, 
						0, 0, THREAD);
				if (broadcastIDP == 0) {
					anaPrint(ANA_ERR,  "handleResolve: unable to register broadcastIDP!\n");
					fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
					anaL0_send(replyTo, data, len);
					return;
				}
				ret = anaL0_forkData(userIDP, 0, broadcastIDP, 0);
				anaPrint(ANA_DEBUG, "try to fork: %i %i ret: %i\n", userIDP, broadcastIDP, ret);
				dA_test = dA_test->next;
			}
			dA_test = dA_start;
			
			anaPrint(ANA_NOTICE, "data channel sent to user 3\n");
			
			*replyToptr = userIDP;
			anaL0_send(replyTo, data, len);
			return;
		}
	}
	else { /* the chanType is either unicast or anycast => we send a request */
		anaPrint(ANA_NOTICE, "send resolve request\n");
		ret = sendResolveRequest(context, target, targetLen, description, descLen, chanType[0], replyTo);
		if (ret<0) {
			anaPrint(ANA_ERR, "handleResolve: unable to send resolve request!\n");
			fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR);
                       	anaL0_send(replyTo, data, len);
		}
		return;
	}
	return;
}

void  handleConfig(void *data, int len) {
	xrpMsg_t msg;
	/*pointer, since received from XRP message */
	anaLabel_t *replyToptr, *labelptr;
	anaLabel_t replyTo, label;
	int labelLen, addressLen, lenMsg, ret, maskLen;
	char *address = NULL, *ip, *mask, *tmp;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));

	if (data==NULL || len<0)
       		return;

	labelptr = (anaLabel_t *) getXRPArg((xrpMsg_t) data, XRP_CLASS_LABEL, &labelLen, 0);
	replyToptr = (anaLabel_t *) getXRPArg((xrpMsg_t) data, XRP_CLASS_LABEL, &labelLen, 1);
	address = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_IPADDRESS, &addressLen, 0);
	mask = (char *) getXRPArg((xrpMsg_t) data, XRP_CLASS_IPMASK, &maskLen, 0);
	if(replyToptr != NULL)
		replyTo = *replyToptr;
	else {
		anaPrint(ANA_ERR, "no replyTo\n");
		return;
	}	
	label = *labelptr;
	/* publish ip along with the data IDP in the QREP of eth */
	msg = anaL1_allocateMessage();

	ip = (char *) malloc(20);
	strcat(ip, "ip+");
	tmp = strcat(ip, address);
	context.value = "*";
	context.valueLen = 2;
	service.value = tmp;
	service.valueLen = strlen(tmp)+1;

	lenMsg = anaL1_encCompartmentPublish(msg, ip_recvLLIDP, &context, &service, IDP_PERM);
	if (lenMsg<0) {
		anaPrint(ANA_ERR, "handleConfig: problem encoding publish request for ip_recvLL!\n");
		anaValidate("handleConfig: problem encoding publish request for ip_recvLL!\n");
	}

	ret =  anaL1_requestReply(label, msg, lenMsg, (anaCallback_t) &  handlePublishReplyConfig, 
				  &replyTo, (copyAuxFunction) &cloneLabel_ipenc, (freeAuxFunction) &freeLabel_ipenc, 0);
	if (ret<0) {
		anaPrint(ANA_ERR, "handleConfig: problem sending publish request for ip_recvLL!\n");
		anaValidate("handleConfig: problem encoding publish request for ip_recvLL!\n");
	}
	free(msg);

	dA->next = (struct devAddr *) malloc(sizeof(struct devAddr));
	dA->dev = label;
	dA->addr = (char*) malloc(strlen(address) + 1);
	memcpy(dA->addr, address, strlen(address) + 1);
	dA->mask = (char*) malloc(strlen(mask) + 1);
	memcpy(dA->mask, mask, strlen(mask) + 1);
	dA = dA->next;
	dA->dev = 0;
	dA->addr = NULL;
	dA->mask = NULL;
	dA->next = NULL;
	
	anaPrint(ANA_NOTICE, "address %s with mask %s configured\n", address, mask);
	anaValidate("address %s with mask %s configured\n", address, mask);

}

void  makeHeaderAndSend(char *data, int len, anaLabel_t input, struct fullTarget *aux) {
	char *index = 0;
	struct iphdr tmphed;
	void *result = NULL;
	uint8_t ip_dlv;
	uint16_t *chksum, *tmp;
	int targetLen, newLen, resultLen, size = 0, srcLen, typeLen;
	xrpMsg_t msg, chkMsg;
	struct timespec timeout;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 1; 

	if (data==NULL || len<0 ||aux == NULL)
		return;
	
	anaPrint(ANA_NOTICE, "\n\n--> received data from higher layer for ethernet\n\n");
	
	if (aux->destDescrip!=NULL) {
		targetLen = strlen(aux->destDescrip) + 1;
	}
	else {
		targetLen = 0;
	}
	if (aux->srcDescrip!=NULL) {
		srcLen = strlen(aux->srcDescrip) + 1;
	}
	else {
		srcLen = 0;
	}
	if (aux->type==IP_CMD) {
		typeLen = sizeof(uint8_t);
		anaPrint(ANA_NOTICE, "from: %s / %s to %s / %s\n", aux->source, aux->srcDescrip, aux->address, aux->destDescrip);
	}
	else {
		typeLen = 0;
		anaPrint(ANA_NOTICE, "from: %s / %s to %s / 0x%X (next header)\n", aux->source, aux->srcDescrip, aux->address, aux->type);
	}
	
	tmphed.daddr = htonl(dottedQuadToHex(aux->address));
	tmphed.saddr = htonl(dottedQuadToHex(aux->source));
	tmphed.version = 0x4; /* ipv4 */
	tmphed.ihl = 0x5;
	tmphed.tos = 0x00;
	tmphed.tot_len = htons(tmphed.ihl * 4 + len + targetLen + srcLen + typeLen + 2 * sizeof(uint16_t));
	tmphed.id = 0x0000;
	tmphed.frag_off = 0x0040;
	tmphed.ttl = 0x40; /* default time to live 64 */
	tmphed.protocol = aux->type; /* next header field */
	tmphed.check = 0x0000;
	
	/* computing checksum */
	if (sumIDP == 0) {
		context.value = ".";
		context.valueLen = 2;
		service.value = "checksum";
		service.valueLen = strlen("checksum")+1;
		sumIDP = anaL2_resolve(NODE_LABEL,  &context, &service, 'u', &description, &timeout);
	}
	if (sumIDP != 0) {
		chkMsg = anaL1_allocateMessage();
		size = anaL1_encData(chkMsg, &tmphed, sizeof(struct iphdr),  NULL, NULL, 0);
		result = anaL2_requestReply(sumIDP, chkMsg, size, &size, &timeout);
		if (result==NULL) {
			anaPrint(ANA_ERR, "sendResolveRequest: requestReply failed (2)\n");
			return;
		}

		chksum = (uint16_t *) getXRPArg((xrpMsg_t) result, XRP_CLASS_MESSAGE, &resultLen, 0);
		tmphed.check = htons(*chksum);
		anaPrint(ANA_NOTICE, "checksum successfully added: 0x%X\n", htons(tmphed.check));
	}

	newLen = len + sizeof(struct iphdr) + targetLen + srcLen + typeLen + 2 * sizeof(uint16_t);
	msg = (char *) malloc(newLen);
	index = msg;
	memcpy(index, &tmphed, sizeof(struct iphdr));
	index += sizeof(struct iphdr);
	if (aux->type==IP_CMD) {
		ip_dlv = IP_DLV;
		memcpy(index, &ip_dlv, sizeof(uint8_t));
		index += sizeof(uint8_t);
	}
	if (targetLen) {
		tmp = (uint16_t *) index;
		*tmp = htons(targetLen);
		index += sizeof(uint16_t);
		memcpy(index, aux->destDescrip, targetLen);
		index += targetLen;
	}
	tmp = (uint16_t *) index;
	*tmp = htons(srcLen);
	index += sizeof(uint16_t);
	if (srcLen) {
		memcpy(index, aux->srcDescrip, srcLen);
		index += srcLen;
	}
	memcpy(index, data, len);
	index += len;
	
	anaL0_send(aux->label, msg, newLen);
	anaPrint(ANA_NOTICE, "data message sent to ethernet\n");
	free(msg);

}

/* returns a list of all ipaddresses */
void  handleGet(void *data, int len) {
	xrpMsg_t msg;
	anaLabel_t *replyToptr;
	anaLabel_t replyTo;
	int labelLen;
	int totalLength;
	struct devAddr *dA_test = dA_start;

	if (data==NULL || len<0)
       		return;
	
	replyToptr = (anaLabel_t *) getXRPArg((xrpMsg_t) data, XRP_CLASS_LABEL, &labelLen, 0);
	replyTo = *replyToptr;

/*	char * labelstr = allocLabelStr();
	label2str(labelstr, replyTo);
	anaPrint(ANA_DEBUG, "anaL2_requestReply----------replyTo %s \n", labelstr);
	free(labelstr);
*/

	msg = anaL1_allocateMessage();
	totalLength = fillXRPCommand(msg, XRP_CMD_GET);
	if (totalLength<0) {
		anaPrint(ANA_EMERG, "handleRip: unable to fill command in message!\n");
		free(msg);
	}
	
	dA_test = dA_start;
	while (dA_test->addr!=NULL) {
		int arglen = addXRPArg(msg, XRP_CLASS_IPADDRESS, dA_test->addr, strlen(dA_test->addr) + 1);
		if (arglen < 0) {
			anaPrint(ANA_EMERG, "handleRip: unable to add address to message!\n");
			fillXRPCommand((xrpMsg_t) data, XRP_CMD_ERROR); /* send an error msg back*/
			anaL0_send(replyTo, data, len);
			free(msg);
		}
		else {
			anaPrint(ANA_DEBUG, "addedd address %s\n", dA_test->addr);
		}
		totalLength += arglen;
		dA_test = dA_test->next;
	}
	anaPrint(ANA_DEBUG, "send answer %s with length %i\n", msg, totalLength);
	anaL0_send(replyTo, msg, totalLength);
	dA_test = dA_start;
}

/* added by Christophe with compartment finder code */

void anaL2_msg(struct anaL2_message *msg) {

	if (!msg || !msg->data)
		return;

	ip_recvHL(msg->data, msg->dataLen, msg->idp, NULL);
}

/* receive data function from higher layers */
void  ip_recvHL(char *data, int len, anaLabel_t input, void *aux) {
	
	/* parse the received xrp message */
	xrpCmd_t command = data;
	
	/* publish data IDP message */
        if (equalXRPCommands(command, XRP_CMD_PUBLISH)) {
		anaPrint(ANA_NOTICE, "\n\n--> received publish request\n\n");
		handlePublish(data, len);
	}
	
	/* resolve received */
	if (equalXRPCommands(command, XRP_CMD_RESOLVE)) {
		anaPrint(ANA_NOTICE, "\n\n--> received resolve request\n\n");
		handleResolve(data, len);
	}
	
	/* unpublish received */
	if (equalXRPCommands(command, XRP_CMD_UNPUBLISH)) {
		anaPrint(ANA_NOTICE, "\n\n--> received unpublish request\n\n");
		handleUnpublish(data, len);
	}
	
	/* config received */
	if (equalXRPCommands(command, XRP_CMD_CONFIG)) {
		anaPrint(ANA_NOTICE, "\n\n--> received config request --1--\n\n");
		handleConfig(data, len);
		anaPrint(ANA_NOTICE, "handle config finished\n");
		anaValidate("handle config finished\n");
	}
	
	/* get information, for now a list of the local addresses */
	if (equalXRPCommands(command, XRP_CMD_GET)) {
		anaPrint(ANA_NOTICE, "\n\n--> received get request\n\n");
		handleGet(data, len);
	}

}

/* will be called after a publish message was sent, checks whether an error occured */
void  handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);

	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

int  brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));
	anatimer_init();

	if (validate_init("ip_local") < 0)
		anaPrint(ANA_ERR, "couldn't start validation\n");
	
	/* create the data repository */
	initQRepository(&dataRep, QREP_ENTRYTAB_DEFAULTMASK, QREP_KEYTAB_DEFAULTMASK);
	
	/* initializes devAddr list */
	dA = (struct devAddr *) malloc(sizeof(struct devAddr));
	dA_start = dA;
	dA->dev = 0;
	dA->addr = NULL;
	dA->mask = NULL;
	dA->next = NULL;
	
	requestList = NULL;
	sumIDP = 0;

	/* register data receive callback function */
	ip_recvLLIDP = anaL0_registerCallback((anaCallback_t) &  ip_recvLL, NULL, NULL, NULL, 0, IDP_PERM, THREAD);
	if (ip_recvLLIDP < 0) {
		anaPrint(ANA_EMERG,  "brick_start: unable to register ip_recvLL!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "brick_start: ip_recvLL registered\n");
	
	/* register callback function for higher layer bricks */
	ip_recvHLIDP = anaL0_registerCallback((anaCallback_t) &  ip_recvHL, NULL, NULL, NULL, 0, IDP_PERM, THREAD);
	if (ip_recvHLIDP < 0) {
		anaPrint(ANA_EMERG,  "brick_start: unable to register ip_recvHL!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE,  "brick_start: ip_recvHL registered\n");
	
	/* publish ip along with the data IDP in the KVR */
	msg = anaL1_allocateMessage();
	context.value = ".";
	context.valueLen = 2;
	service.value =  "ip+internet+inet+compartment";
	service.valueLen = strlen("ip+internet+inet+compartment")+1;

	lenMsg = anaL1_encCompartmentPublish(msg, ip_recvHLIDP, &context, &service, IDP_PERM);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "brick_start: problem encoding publish request for ip_recvHL!\n");
		return -1;
	}
	ret =  anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) &  handlePublishReply, NULL, NULL, NULL, 0);
	if (ret<0) {
		anaPrint(ANA_EMERG, "brick_start: problem sending publish request for ip_recvHL!\n");
		return -1;
	}
	free(msg);

	/* added by Christophe for publication in compartment finder */

	anaL2_initDefault();
	service.value = "cfinder";
	service.valueLen = strlen(service.value)+1;	
	cfinder = anaL2_resolve(NODE_LABEL, &context, &service, 'u', NULL, NULL);
	if (!cfinder)
		anaPrint(ANA_ERR, "cannot find cfinder brick in KVR\n");
	else
	{
		service.value = IPv4_REGEXP;
		service.valueLen = strlen(service.value)+1;
		regexp = anaL2_publish(cfinder, &context, &service, anaL2_msg, NOTHREAD, NULL);
		
		if (!regexp)
			anaPrint(ANA_ERR, "cannot publish IPv4 regexp in cfinder\n");
		else
			anaPrint(ANA_DEBUG, "published IPv4 regexp in cfinder\n");
	}

	anaValidate("ip brick_start function finished\n");
	return 1;
}

/* free function which has to be provided for the QREP
 * to free all the entries
 */
static void freeValue(void *value) {
	free(value);
}

/* compatibility */
static void freeLabel_ipenc(anaLabel_t *l) {
	return;
}
/*compatibility */
anaLabel_t *cloneLabel_ipenc(anaLabel_t *aux) {
	anaLabel_t *new = malloc(sizeof(anaLabel_t));
	*new = *aux;
	return new;
}

