 /* C/bricks/ip/ip_usr.c
 *  
 * sample ip user application, sends a message to other ip_usr
 * depends on: ip_enc 
 * runs in user space
 * KVR entries: none
 * 
 * use:
 *
 * ip_usr -a {context}
 *
 * Author: Stephan Dudler <dudlers@ee.ethz.ch> 
 *	   Ariane Keller <arkeller@ee.ethz.ch> Adaption for release 0.2
 *	
 * Copyright (c) 2008, Stephan Dudler, Swiss Federal Institute of Technology
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"
/* Update the fields below with your own name + MINMEX-brick gates specifications */

#ifdef KERNEL
MODULE_LICENSE("GPL");
MODULE_AUTHOR("Stephan Dudler");
#endif /* KERNEL */

char mymodename[] = "ip_usr";


/*-------------------------------------------------------------------------------------*/

/* brick specific code starts here */

anaLabel_t recvDataIDP;
anaLabel_t ipLabel;

#include "ip.h"
struct context_s publcontext;
struct service_s publservice;

/* called upon leaving the ip_usr */
void brick_exit() {
	xrpMsg_t msg;
	int ret = 0, len = 0;

        anaPrint(ANA_NOTICE, "IP_USR HALTED\n");
	if (getAuxArg(0) != NULL) {
		msg = anaL1_allocateMessage();
		/* unpublish the the entries in the QREP of the ipBrick */
		len = anaL1_encCompartmentUnpublish(msg, recvDataIDP, &publcontext, &publservice);
		ret = anaL0_send(ipLabel, msg, len);
		if (ret<0) {
			anaPrint(ANA_ERR, "problem sending unpublish request for recvData IDP!\n");
		}
		free(msg);
	}	
	anaL0_unregisterCallback(recvDataIDP, 0);

}

/* receive Data function from ip */
void recvData_usr(void *data, int len, anaLabel_t input, void *aux) {
       	int lenMsg, senderContextLen, senderTargetLen;
	void *message, *senderContext, *senderTarget;
	
       	if (data==NULL || len<0)
       		return;
	
	lenMsg = anaL1_decResponse(data, XRP_CLASS_MESSAGE, 0, (void**) &message);
	senderContextLen = anaL1_decResponse(data, XRP_CLASS_SRC_CXT, 0, &senderContext);
       	senderTargetLen = anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, &senderTarget);
	
	anaPrint(ANA_NOTICE, "--> %i bytes received from: %s at %s\n", lenMsg, (char *) senderTarget, (char *) senderContext);
	anaPrint(ANA_NOTICE, "### %s ###\n", (char *) message);
}

/* sends a message back to the IDP
 * we only expect one IDP
 */
void sendMessage(anaLabel_t ipLabel) {
	struct timespec timeout;
	char *message ="Hi there! How are you doing?";
	char *context = "*";
	anaLabel_t targetIDP = 0;

	if (getAuxArg(0)!=NULL)
		context = getAuxArg(0);
	
	if (ipLabel == 0)
		return;
	
	/* lookup target */
	publcontext.value = context;
	publcontext.valueLen = strlen(context) + 1;
	publservice.value = "ip_usr";
	publservice.valueLen = strlen("ip_usr") + 1;
	memset(&timeout, 0, sizeof(struct timespec));
	timeout.tv_sec = 10; 
	anaPrint(ANA_ERR, "start resolve for %s\n", (char *)publcontext.value);
	targetIDP = anaL2_resolve(ipLabel, &publcontext, &publservice, 'u', &publservice, &timeout);
	
	/* we send our message */
	if (targetIDP == 0 )
		anaPrint(ANA_ERR, "targetIDP == 0\n");
	else if(anaL0_send(targetIDP, message, strlen(message) + 1) < 0 )
		anaPrint(ANA_ERR, "unable to send message!\n");
	else
		anaPrint(ANA_NOTICE, "message sent to %s\n", context);	
}

/* will be called after a publish message was sent, checks whether an error occured */
void handlePublishReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name = NULL;

	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**) &name);
	
	if (anaL1_isError(data)) { /* if ERROR command */
		if (name!=NULL)
			anaPrint(ANA_ERR, "publish of entry %s failed!\n", name);
		else
			anaPrint(ANA_ERR, "publish error!\n");
		return;
	}
	else
		if (name!=NULL)
			anaPrint(ANA_NOTICE, "publish of entry %s succeeded\n", name);
		else
			anaPrint(ANA_NOTICE, "publish success\n");
}

void handleLookupReply(char *data, int len, anaLabel_t input, void *aux) {
	char *name;
	int labelSize;
	anaLabel_t *temp;

	anaPrint(ANA_NOTICE, "received lookup reply for ip in the KVR\n");
	
	if (anaL1_isError(data)) { /* if ERROR command */
		anaPrint(ANA_EMERG, "lookup for ip failed -> cannot work without ip!\n");
		return;
	}
	if (anaL1_decResponse(data, XRP_CLASS_NAME, 0, (void**) &name)>0) {
		labelSize =  anaL1_decResponse(data, XRP_CLASS_LABEL, 0, (void**) &temp);
		if (name!=NULL && temp != 0) {
			int lenMsg, ret; 
			xrpMsg_t msg;
			struct context_s context;
			struct service_s service;

			ipLabel = *temp;
	
			/* publish "ip_usr" along with the data IDP in the QREP of ip */
			msg = anaL1_allocateMessage();
			context.value = "*";
			context.valueLen = 2;
			service.value = "ip_usr";
			service.valueLen = strlen("ip_usr")+1;

			lenMsg = anaL1_encCompartmentPublish(msg, recvDataIDP, &context, &service, 1);
        		if (lenMsg<0) {
                		anaPrint(ANA_ERR, "problem encoding publish request for recvData IDP in QREP of ip!\n");
   	     		}
			
			ret =  anaL1_requestReply(ipLabel, msg, lenMsg, (anaCallback_t) & handlePublishReply, NULL, NULL, NULL, 0);
			if (ret<0) {
                		anaPrint(ANA_ERR, "problem sending publish request for recvData IDP in QREP of ip!\n");
        		}
        	
			/* send a message */
			
			if (getAuxArg(0)==NULL) {
				anaPrint(ANA_ERR, "only listen for messages\n");
			}
			else
				sendMessage(ipLabel);
			
			free(msg);
		}
	}
}

int brick_start() {
	int ret = 0, lenMsg;
	xrpMsg_t msg;
	struct context_s context;
	struct service_s service;
	struct service_s description;
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&description, 0, sizeof(struct service_s));

	/* register data receive callback function */
	recvDataIDP = anaL0_registerCallback((anaCallback_t) &recvData_usr, NULL, NULL, NULL, 0, IDP_PERM, NOTHREAD);
	if (recvDataIDP<0) {
		anaPrint(ANA_ERR,  "recvDataIDP: unable to register callback IDP to ANA minmex!\n");
		return 1;
	}
	anaPrint(ANA_NOTICE,  "registered recvData IDP\n");
		
	/* lookup the ip brick by name */
	msg = anaL1_allocateMessage();
	
	context.value = ".";
	context.valueLen = 2;
	service.value = "ip";
	service.valueLen = strlen("ip")+1;

	lenMsg = anaL1_encResolve(msg, &context, &service, 'u', &description);
	if (lenMsg<0) {
		anaPrint(ANA_EMERG, "unable to format lookup request for \"ip\" -> cannot work without ip!\n");
		return -1;
	}
	ret = anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) & handleLookupReply, NULL, NULL, NULL, 1); 
	if (ret<0) {
		anaPrint(ANA_EMERG, "send lookup msg for \"ip\" failed -> cannot work without ip!\n");
		return -1;
	}
	anaPrint(ANA_NOTICE, "lookup query sent\n");
	
	free(msg);
	return 1;
}
