#ifndef _IP2ANA_H
#define _IP2ANA_H

/* C/bricks/ip2ana/common/ip2ana.h
 *  
 * ANA brick for ip2ana adaptation layer 
 * 
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <linux/ip.h>
#include <netinet/in.h>
#include <netinet/ip6.h>
#include <arpa/inet.h>

#include "anaCommon.h"
#include "anaLib1.h"
#include "anaLib2.h"

#define DNS_REQ_A           1
#define DNS_REQ_AAAA       28
#define DNS_REQ_PTR        12

#define IPT_PORT 4201
#define BRICK_PORT 4203
#define I2A_PORT 4204
#define A2I_PORT 4202

struct udp_ports {
	
	/* UDP port numbers to communicate with ip2ana netfilter module */

	u_int16_t ipt_port;   /* in network order */
	u_int16_t brick_port; /* in network order */
	u_int16_t i2a_port;   /* in network order */
	u_int16_t a2i_port;   /* in network order */
};

extern struct udp_ports uports; /* communication with ip2ana netfilter module */

struct ip2ana_entry {

	/* some state to map ip addresses with context/service tuples */

	char inet_family;
	union {
	  struct in_addr ip4;   /* ipv4 address */
	  struct in6_addr ip6;  /* ipv6 address */
	} ip;
	anaLabel_t s_idp;    /* idp for service */
	anaLabel_t c_idp;    /* idp for compartment */
	struct context_s ctx;
	struct service_s srv;
	struct ip2ana_entry *next;
};

extern struct ip2ana_entry *pub_list; /* list of published ip addresses */
extern struct ip2ana_entry *res_list; /* list of resolved ip addresses */

/* function prototypes */

int add_ip2ana_entry(struct ip2ana_entry **list, char inet_family, void *ip, anaLabel_t s_idp, 
		     anaLabel_t c_idp, struct context_s *ctx, struct service_s *srv);
void free_ip2ana_list(struct ip2ana_entry **list);
struct ip2ana_entry* lookup_ip2ana_entry(struct ip2ana_entry *list, char family, void *ip);

int check_anahosts(char *query, char *qtype, char *result);
void publish_listen(void);
int get_args(void);

void fwd_to_ana(char *buf, int len);

void dnsptr2ip4(char *cp, struct in_addr *addr);
void dnsptr2ip6(char *ptr, struct in6_addr *addr);

#ifdef USER
int user_socket_init(struct udp_ports *ports);
int user_file_init(char *path);
void user_close_handlers(void);
int user_loop(void);
#endif /* USER */

#endif /* IP2ANA_H */
