/* C/bricks/templates/brickAL2_template_server.c
 * 
 * template file which shows how to use API Level 2
 * It publishes a service to the MINMEX
 * For more information please refer to the doxygen docu or
 * to the ana-core documentation.
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   Christophe Jelger <christophe.jelger@unibas.ch> 
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*----------------------- Default brick header code -----------------------*/
/* brick_template.h contains the functions that initialize the bricks */
#include "brick_template.h"

/* Update the fields below with the name of your brick.
 * This name in combination with the process ID 
 * will be used for the log files */
char mymodename[]="brick2_server";

/* Standard ANA code uses the ISC licence.
 * To be able to load our brick to the kernel we use
 * the compatible "Dual BSD/GPL" license
 */
#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif
/*----------------------- End default brick header code -------------------*/

/*----------------------- Brick specific code starts here------------------*/
/* Your code MUST always provide both 
 * the brick_start and brick_exit functions
 */

static anaLabel_t serviceIDP;

static struct service_s service;
static struct context_s context;
static struct timespec timeout;




void rcvData(struct anaL2_message *msg){
	anaPrint(ANA_DEBUG, "received message : %s with len %i\n", (char *)msg->data, msg->dataLen);
	if(msg->senderContext)
		anaPrint(ANA_DEBUG, "sender context is : %s with len %i\n", (char *)msg->senderContext->value, msg->senderContext->valueLen);
	if(msg->senderService)
		anaPrint(ANA_DEBUG, "sender service is: %s with len %i\n", (char *)msg->senderService->value, msg->senderService->valueLen);

	anaPrint(ANA_DEBUG, "an IDP to send an answer: %i\n", msg->responseLabel);
	
	anaValidate("received message on firstIDP\n");
//	anaValidate("received message %s with len %i\n", (char *)msg->data, msg->dataLen);

}

/* The brick_exit function is called before a brick is unloaded
 * It should do all kind of clean up
 */
void brick_exit() {

	if (anaL2_unpublish(NODE_LABEL, serviceIDP, &context, &service, &timeout) < 0)
		anaPerror(ANA_ERR, "unpublish failed");
	else
		anaPrint(ANA_DEBUG, "unpublish success\n");

	anaPrint(ANA_DEBUG, "Brick EXIT\n");
}

/* The brick_start function is the main function of each brick */
int brick_start() {
	
	validate_init(mymodename);
	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	memset(&timeout, 0, sizeof(struct timespec));

	timeout.tv_sec = 1; 

	/* node local context */
	context.value = ".";
	context.valueLen = 2;
	
	/* our service is called "learn ANA" */
	service.value = "learn ANA server";
	service.valueLen = strlen(service.value) + 1;
	
	/* we publish the service with a callback function */

	serviceIDP = anaL2_publish(NODE_LABEL, &context,  &service, (AL2Callback_t) &rcvData, NOTHREAD, &timeout);
	
	if(serviceIDP == 0)
		anaPerror(ANA_ERR, "publish failed");
	else
		anaPrint(ANA_NOTICE, "publish success \n");	

	return 1;

}
