/* C/bricks/templates/brickAL2_template_client.c
 * 
 * template file which shows how to use API Level 2
 * It publishes a service to the MINMEX
 * For more information please refer to the doxygen docu or
 * to the ana-core documentation.
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   Christophe Jelger <christophe.jelger@unibas.ch> 
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*----------------------- Default brick header code -----------------------*/
/* brick_template.h contains the functions that initialize the bricks */
#include "brick_template.h"

/* Update the fields below with the name of your brick.
 * This name in combination with the process ID 
 * will be used for the log files */
char mymodename[]="brick2_client";

/* Standard ANA code uses the ISC licence.
 * To be able to load our brick to the kernel we use
 * the compatible "Dual BSD/GPL" license
 */
#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif
/*----------------------- End default brick header code -------------------*/

/*----------------------- Brick specific code starts here------------------*/
/* Your code MUST always provide both 
 * the brick_start and brick_exit functions
 */

/* The brick_exit function is called before a brick is unloaded
 * It should do all kind of clean up
 */
void brick_exit() {

	anaPrint(ANA_DEBUG, "Brick EXIT\n");
}


/* The brick_start function is the main function of each brick */
int brick_start() {
	struct context_s context;
	struct service_s service;
	struct service_s myservice;
	struct timespec timeout;
	anaLabel_t serviceIDP;	
	int ret = 0;

	validate_init(mymodename);
	
	memset(&context, 0, sizeof(struct context_s));
	memset(&service, 0, sizeof(struct service_s));
	memset(&myservice, 0, sizeof(struct service_s));
	memset(&timeout, 0, sizeof(struct timespec));
		
	context.value = ".";
	context.valueLen = 2;
	
	service.value = "learn ANA server";
	service.valueLen = strlen(service.value) + 1;
	
	myservice.value = "learn ANA client";
	myservice.valueLen = strlen(myservice.value) + 1;

	timeout.tv_sec = 1; 
		
	serviceIDP = anaL2_resolve(NODE_LABEL, &context, &service, 'u', &myservice, &timeout); 

	if (serviceIDP == 0) {
		anaPerror(ANA_ERR, "Unable to resolve service");
		return -1;
	}
	else
		anaPrint(ANA_DEBUG, "resolve SUCCESS serviceIDP= %d \n", serviceIDP);
	
	/* say hello to the server */
	ret = anaL0_send(serviceIDP, "hello", strlen("hello") + 1);
	anaValidate("sent message to server, returned %i\n", ret);	

	return 1;
}



