/* C/bricks/templates/brick_template_misc.c
 * 
 * shows how to use ana_timer, ana_theads, and ana_locks
 * For more information please refer to the doxygen docu or
 * to the ana-core documentation.
 * 
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   Ghazi Bouabene <ghazi.bouabene@unibas.ch> 
 * 	   Christophe Jelger <christophe.jelger@unibas.ch> 
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*----------------------- Default brick header code -----------------------*/
/* brick_template.h contains the functions that initialize the bricks */
#include "brick_template.h"

/* Update the fields below with the name of your brick.
 * This name in combination with the process ID 
 * will be used for the log files */
char mymodename[]="brick_misc";

/* Standard ANA code uses the ISC licence.
 * To be able to load our brick to the kernel we use
 * the compatible "Dual BSD/GPL" license
 */
#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif
/*----------------------- End default brick header code -------------------*/

/*----------------------- Brick specific code starts here------------------*/
/* Your code MUST always provide both 
 * the brick_start and brick_exit functions
 */

analock_t lock1;

void brick_exit() {
	freeAuxArgs();
	return;
}

struct arg {
	int a;
	int c;
};

/* an endless running thread
 * the anaThreadShouldStop is required 
 * if the brick runs as a Linux kernel module
 */
void * thread(void *argument) {
	int i = 0;
	anaValidate("thread started\n");
	while(!anaThreadShouldStop()) {
		anaPrint(ANA_DEBUG, "loop %i\n", i);
		i++;
		sleep(1);
	}
	return NULL;
}

void timer (void *argument) {
	struct arg **tmpptr = (struct arg **)argument;
	struct arg *tmp = *tmpptr;
	anaPrint(ANA_DEBUG, "timer called %i times\n", tmp->a);
	tmp->a++;
	if(tmp->a == 4) {
		anatimer_del(tmp->c);
		analock_unlock(&lock1);
	}
}


int brick_start() {
	int ret = 0;
	struct arg *argument;
	char * auxarg = NULL;
	int i = 0;

	/* we initialize the validation environment */
	validate_init(mymodename);

	/* the command line arguments for the brick */
	while((auxarg = getAuxArg(i)) != NULL) {
		anaPrint(ANA_DEBUG, "command line argument %i: %s\n", i, auxarg);
		anaValidate("command line argument %i: %s\n", i, auxarg);
		i++;
	}


	argument = malloc(sizeof(struct arg));
	argument->a = 0;

	/* ana provides also condition locks and read/write locks */
	analock_init(&lock1);
	
	/* the last argument is the name of the thread */
	anaThread(thread, argument, "thread1");

	analock_lock(&lock1);

	/* there are periodic and non periodic timers */
	ret = anatimer_add(1000, 		/* miliseconds */
			   ANATIMER_PERIODIC,   /* periodic timer */
			   ANATIMER_ABSOLUTE,   /* see: anacore-doc.pdf */
			   timer, 		/* the timer function */
			   &argument, 		/* aux arg for timer */
			   sizeof(void *));     /* sizeof aux arg,
						since it is passed by value, 
						we give the address of the pointer */
	argument->c = ret;	/* the timer id, needed to delete this timer */
	
	/* wait until timer has finished */
	analock_lock(&lock1);
	anaPrint(ANA_DEBUG, "timer has terminated after firing %i times\n", argument->a);
	anaValidate("timer has terminated after firing %i times\n", argument->a);
	
	return 1;
}

