/* test_ana/src/validate_al0.c
 * 
 * validation file for AL0
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"


/* Update the fields below with your own name + MINMEX-brick gates specifications */
char mymodename[]="validate_al0";

#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif

/*--------------------------------End default brick header code --------------------------*/

anaLabel_t callBackIDP;
anaLabel_t peerIDP;
anaLabel_t thirdIDP;
analock_t lock;

void callBackFct2(char *data, int len, anaLabel_t input,  void *aux) {
        char *ch = (char *) aux;
	sleep(1);
	anaValidate("callBackFct2 aux: %s, data %s, len %i\n", ch, data, len);
}

void callBackFct(char *data, int len, anaLabel_t input,  void *aux) {
        char *ch = (char *) aux;
	anaValidate("callBackFct aux: %s, data %s, len %i\n", ch, data, len);
	analock_unlock(&lock);
}


void * copyData(void * data){
	char * new_data = (char *)malloc(strlen((const char *)data)+1);
	memcpy(new_data, (char *)data, strlen((const char *)data));
	new_data[strlen((const char *)data)]='\0';
	return (char *)new_data;
}

void freeData(void * data){
	free(data);
}

void brick_exit() {
/*	int ret = -1;	
	ret = anaL0_unregisterCallback(thirdIDP, 0);
	anaValidate("anaL0_unregisterCallback returned %i\n", ret);

	ret = anaL0_unregisterCallback(peerIDP, 0);
	anaValidate("anaL0_unregisterCallback returned %i\n", ret);

	ret = anaL0_unregisterCallback(callBackIDP, 0);
	anaValidate("anaL0_unregisterCallback returned %i\n", ret);

	anaPrint(ANA_DEBUG, "Brick EXIT\n");
*/
}


/******************************/
int brick_start() {
	int ret = 0;
	analock_init(&lock);
	validate_init(mymodename);
	anaValidate("Brick %s\n", mymodename);
	
	callBackIDP = anaL0_registerCallback((anaCallback_t)& callBackFct, (void *) "firstAux", copyData, freeData, 0, IDP_PERM, 1);
	if ( callBackIDP == 0) {
		anaPrint(ANA_DEBUG, "1\n");
		anaValidate("anaL0_registerCallback failed: %i\n", callBackIDP);
	        return -1;
	}
	else
		anaValidate("anaL0_registerCallback success\n");

	analock_lock(&lock);	
	ret = anaL0_send(callBackIDP, "***first message***", strlen("***first message***")+1);
	analock_lock(&lock);
	anaValidate("anaL0_send returned %i\n", ret);

	ret = anaL0_changeLabelAux(callBackIDP, "secondAux", copyData, freeData);
	anaValidate("anaL0_changeLabelAux returned %i\n", ret);
	
	ret = anaL0_send(callBackIDP, "***second message***", strlen("***second message***")+1);
	analock_lock(&lock);
	anaValidate("anaL0_send returned %i\n", ret);

	peerIDP = anaL0_registerCallback((anaCallback_t)& callBackFct, (void *) "peerAux", copyData, freeData, 0, IDP_PERM, NOTHREAD);
	if (peerIDP == 0) {
		anaValidate("anaL0_registerCallback failed: %i\n", callBackIDP);
	        return -1;
	}
	else
		anaValidate("anaL0_registerCallback success\n");
	
	ret = anaL0_redirect(callBackIDP, 0, peerIDP, 0);
	anaValidate("anaL0_redirect returned %i\n", ret);
	
	ret = anaL0_send(callBackIDP, "***third message***", strlen("***third message***")+1);
	analock_lock(&lock);
	anaValidate("anaL0_send returned %i\n", ret);

	/* this will start the callback function in a dedicated thread and sleep 1 seconds in the thread
 	 * this ensures that for the fork *always* the peerIDP is fist prints the message.
 	 */ 
	thirdIDP = anaL0_registerCallback((anaCallback_t)& callBackFct2, (void *) "thirdAux", copyData, freeData, 0, IDP_PERM, THREAD);
	if (thirdIDP == 0) {
		anaValidate("anaL0_registerCallback failed: %i\n", callBackIDP);
	        return -1;
	}
	else
		anaValidate("anaL0_registerCallback success\n");
	

	ret = anaL0_forkData(peerIDP, 0, thirdIDP, 0);
	anaValidate("anaL0_forkData returned %i\n", ret);
	ret = anaL0_send(peerIDP, "***4. message***", strlen("***4. message***")+1);
	analock_lock(&lock);
	anaValidate("anaL0_send returned %i\n", ret);
	analock_unlock(&lock);
	return 1;
}

