/* C/bricks/test_ana/validate_al2.c
 * 
 * validation file for AL2
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"


/* Update the fields below with your own name + MINMEX-brick gates specifications */
char mymodename[]="validate_al2";

#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif

/*--------------------------------End default brick header code --------------------------*/

#define XRP_CMD_GET "get"

anaLabel_t serviceIDP;
analock_t lock;

void brick_exit() {
/*	int ret = -1;	
	ret = anaL0_unregisterCallback(thirdIDP, 0);
	anaValidate("anaL0_unregisterCallback returned %i\n", ret);

	ret = anaL0_unregisterCallback(peerIDP, 0);
	anaValidate("anaL0_unregisterCallback returned %i\n", ret);

	ret = anaL0_unregisterCallback(callBackIDP, 0);
	anaValidate("anaL0_unregisterCallback returned %i\n", ret);

	anaPrint(ANA_DEBUG, "Brick EXIT\n");
*/
}

void recvData3(struct anaL2_message *msg){
	anaLabel_t dest;
	struct service_s service;
	struct context_s context;
	struct timespec timeout;
	char * reply;
	int replylen = 0;
	int len = 0;
	xrpMsg_t msgFwd;
	anaValidate("received message : %s\n", (char *)msg->data);
	
	context.value = ".";
	context.valueLen = strlen(".")+1;
	service.value = "service2";
	service.valueLen = strlen("service2")+1;
	timeout.tv_sec = 1;
	dest = anaL2_resolve(NODE_LABEL, &context,  &service, 'u', NULL, &timeout);

	if(dest == 0)
		anaValidate("recvData2 resolve failed : %i\n", dest);
	else
		anaValidate("recvData2 resolve success\n");


	msgFwd = anaL1_allocateMessage();
	len = fillXRPCommand((xrpMsg_t) msgFwd, XRP_CMD_GET);
	if (len < 0) {
		anaValidate("unable to fill command in message!\n");
		free(msgFwd);
		return;
	}
	reply = anaL2_requestReply(dest, msgFwd, len, &replylen, &timeout);
	if(replylen <= 0)
		anaValidate("requestReply failed\n");
	else
		anaValidate("requestReply: %i %s\n", replylen, reply);
	analock_unlock(&lock);
}

void recvData2(struct anaL2_message *msg){
	anaLabel_t *replyTo;
	int len;

	replyTo = (anaLabel_t *) getXRPArg((xrpMsg_t) msg->data, XRP_CLASS_LABEL, &len, 0);
        if (!replyTo) {
                anaValidate("handleRipreq: unable to find replyTo label!\n");
                return;
        }
	anaL0_send(*replyTo, "hello world", 12);
	return;
	
}

void recvData(struct anaL2_message *msg){
	int ret;
	struct service_s service;
	struct context_s ctxt;
	struct timespec timeout;
	anaValidate("received message : %s\n", (char *)msg->data);
	
	ctxt.value = ".";
	ctxt.valueLen = strlen(".")+1;
	service.value = "service";
	service.valueLen = strlen("service")+1;
	timeout.tv_sec = 1;

	ret = anaL2_unpublish(NODE_LABEL, serviceIDP, &ctxt, &service, &timeout);
	if(ret<0)
		anaValidate("unpublish failed : %i\n", ret);
	else
		anaValidate("unpublish success: %i\n", ret);

	analock_unlock(&lock);
}


int brick_start() {
	struct service_s service;
	struct context_s context;
	struct timespec timeout;

	memset(&service, 0, sizeof(struct service_s));
	memset(&context, 0, sizeof(struct context_s));
	
	analock_init(&lock);
	anaL2_initDefault();

	validate_init(mymodename);
	anaValidate("Brick %s\n", mymodename);

	context.value = ".";
	context.valueLen = 2;
	
	service.value = "service";
	service.valueLen = strlen("service") + 1;
	timeout.tv_sec = 1;	
	/* we publish a service with a callback function */
	serviceIDP = anaL2_publish(NODE_LABEL, &context,  &service, (AL2Callback_t) &recvData, THREAD, &timeout);
	
	if(serviceIDP == 0)
		anaValidate("--->  first publish failed");
	else
		anaValidate("---> first publish ok\n");	

	/* resolve it with a different context -> should fail */
	context.value = "blub";
	context.valueLen = 5;

	serviceIDP = anaL2_resolve(NODE_LABEL, &context,  &service, 'u', NULL, &timeout);
	
	if(serviceIDP == 0)
		anaValidate("--->  first resolve failed\n");
	else
		anaValidate("---> first resolve ok\n");	


	serviceIDP = -1;

	/* resolve it with correct context */
	context.value = ".";
	context.valueLen = 2;

	serviceIDP = anaL2_resolve(NODE_LABEL, &context,  &service, 'u', NULL, &timeout);
	
	if(serviceIDP == 0)
		anaValidate("--->  second resolve failed\n");
	else
		anaValidate("---> second resolve ok \n");	


	anaL0_send(serviceIDP, "hello world", 12);
	analock_lock(&lock);
	/* the callback function has unregistered the service
 	 * try to resolve it again -> should fail */

	serviceIDP = anaL2_resolve(NODE_LABEL, &context,  &service, 'u', NULL, &timeout);
	
	if(serviceIDP == 0)
		anaValidate("--->  3. resolve failed\n");
	else
		anaValidate("---> 3. resolve ok\n");	


	/* publish 2 callback functions 
 	 * used to verify message sending between different callback functions
 	 */
	serviceIDP = anaL2_publish(NODE_LABEL, &context,  &service, (AL2Callback_t) &recvData2, NOTHREAD, &timeout);
	
	if(serviceIDP == 0)
		anaValidate("--->  second publish failed");
	else
		anaValidate("---> second publish ok\n");	

	/* the requestReply in the callback function should fail
 	 * since we have NOTHREAD specified
 	 */
	serviceIDP = anaL2_publish(NODE_LABEL, &context,  &service, (AL2Callback_t) &recvData3, NOTHREAD, &timeout);
	
	if(serviceIDP == 0)
		anaValidate("--->  second publish failed");
	else
		anaValidate("---> second publish ok\n");	

	analock_lock(&lock);

	/* publish it again this time with THREAD specified
 	 * -> requestReply should work 
 	 */
	serviceIDP = anaL2_publish(NODE_LABEL, &context,  &service, (AL2Callback_t) &recvData3, THREAD, &timeout);
	
	if(serviceIDP == 0)
		anaValidate("--->  3. publish failed");
	else
		anaValidate("---> 3. publish ok\n");	

	analock_lock(&lock);

	

	return 1;

}

