/* test_ana/src/validate_al0.c
 * 
 * validation file for AL0
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch>
 * 	   	
 * Copyright (c) 2008, Ariane Keller, ETH Zurich
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/*--------------------------- Default brick header code -------------------------------*/

/* !!! Your code MUST always provide both the brick_start and brick_exit functions !!! */
#include "brick_template.h"


/* Update the fields below with your own name + MINMEX-brick gates specifications */
char mymodename[]="validate_misc";

#ifdef KERNEL
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("ANA Project");
#endif

/*--------------------------------End default brick header code --------------------------*/

analock_t lock1;
analock_t lock2;
analock_t lock3;


void brick_exit() {

}

struct arg {
	int a;
	char *b;
	int c;
};

void * thread1 (void * argument) {
	int i = 0;
	while(i < 5) {
		analock_lock(&lock1);
		anaValidate("thread1: i = %i\n", i);
		analock_unlock(&lock2);
		i++;
	}
	return NULL;
}

void * thread2 (void * argument){
	int i = 0;
	while(i < 5) {
		analock_lock(&lock2);
		anaValidate("thread2: i = %i\n", i);
		
		analock_unlock(&lock1);
		i++;
	}
	return NULL;
}

void * thread3(void *argument) {
	anaValidate("thread3 started\n");
	while(!anaThreadShouldStop()) {
		sleep(1);
	}
	anaValidate("thread3 terminates\n");
	return NULL;
}

void timer1 (void *argument) {
	struct arg **tmpptr = (struct arg **)argument;
	struct arg * tmp = *tmpptr;
	anaValidate("timer1 called %i times\n", tmp->a);
	tmp->a++;
	if(tmp->a == 4) {
		anatimer_del(tmp->c);
		analock_unlock(&lock3);
	}
}


/******************************/
int brick_start() {
	int ret = 0;
	struct arg *arg1;
	struct arg *arg2;

	validate_init(mymodename);
	anaValidate("Brick %s\n", mymodename);

	arg1 = malloc(sizeof(struct arg));
//	free(arg1);
	arg1->b = malloc(strlen("thread1")+1);
	memcpy(arg1->b, "thread1", strlen("thread1")+1);
	arg1->a = 1;

	arg2 = malloc(sizeof(struct arg));
	arg2->b = malloc(strlen("thread2")+1);
	memcpy(arg2->b, "thread2", strlen("thread2")+1);
	arg2->a = 1;

	analock_init(&lock1);
	analock_init(&lock2);
	analock_init(&lock3);
	
	analock_lock(&lock1);
	
	
	anaThread(thread1, arg1, "thread1");
	anaThread(thread2, arg2, "thread2");
	
	analock_lock(&lock3);
	ret = anatimer_add(1000, 1, 0, timer1, &arg1, sizeof(void *));
	arg1->c = ret;
	
	/* wait until timer has finished */
	analock_lock(&lock3);
	anaThread(thread3, NULL, "thread3");

	return 1;
}

