/* C/bricks/tools/quickRep/QREPparser.c
 *   
 * functions for internal use of Qrep 
 * 
 * Author:  Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */

/* temporary */
#ifndef NS2
#include "quickRepository.h"
#endif

void  qrep_freeToken(struct qrep_token *tk){
	if(tk!=NULL){
		free(tk->lexicalVal);
		free(tk);
	}
}

int  qrep_reduceParentheses(struct qrep_token *tkChain){ 
	struct qrep_token *openPar=NULL;
	struct qrep_token *closePar=NULL;
	struct qrep_token *tmpTok=tkChain;
	struct qrep_token *tmpTok2 = NULL;
	struct qrep_token *next = NULL;
	struct qrep_token *next2 = NULL;
	
	while(tmpTok!=NULL){
		next=tmpTok->next;	
		if(tmpTok->type==QREP_OPERATOR && tmpTok->lexicalVal[0]=='('){
			int found =0;
			int intermediateOpar = 0;			
			openPar=tmpTok;
			/* we look for the closing parentheses */
			
			tmpTok2=tmpTok->next;
			while(!found && tmpTok2!= NULL){
				
				next2=tmpTok2->next;
				if(tmpTok2->type==QREP_OPERATOR && tmpTok2->lexicalVal[0]=='(')
					intermediateOpar++;	
				
				if(tmpTok2->type==QREP_OPERATOR && tmpTok2->lexicalVal[0]==')') {
					if(intermediateOpar>0)
						intermediateOpar--;
					else {	/* we found the closing parenthesis*/
						closePar = tmpTok2;
						found = 1;
						/*we need to cut the expression included in parenthesis form the chain*/
						
						if(closePar->prev == openPar)
							return -1; /* we do not allow () */
						else
							closePar->prev->next = NULL;
						
											
						if(qrep_evalTokenChain(openPar->next)<0)
							return -1;
						/* the expression between parentheses has been reduced
						inside the first token after openPar, we need to move it in OPAR token */
						openPar->type = QREP_OPERAND;
						openPar->semanticVal = openPar->next->semanticVal;
						
						qrep_freeToken(openPar->next);
						
						openPar->next=closePar->next;
						next = closePar->next;
						if(closePar->next!=NULL)
							closePar->next->prev=openPar;
						
						qrep_freeToken(closePar);
					}
				}
				
				tmpTok2=next2;
			}
			if(!found){
				anaPrint(ANA_ERR, "Missing closing parenthesis\n");
				return -1;
			}
		}
		tmpTok=next;
	}
	return 0;
}


/* this function is always called on tokenChains that do notcontain parentheses*/
int  qrep_reduceAND(struct qrep_token *tkChain){ 
	struct qrep_token *next=NULL;
	while(tkChain!=NULL){
		next=tkChain->next;
		if(tkChain->type==QREP_OPERATOR && tkChain->lexicalVal[0]=='*'){
					
			/* first we check if previous and next tokens are Operands
			  if not we return an error	
			*/	
			if(tkChain->prev==NULL || tkChain->prev->type!=QREP_OPERAND ||tkChain->next==NULL || tkChain->next->type!=QREP_OPERAND ){
				anaPrint(ANA_ERR, "invalid operands for AND operator \n");
				return -1;
			}
			
			tkChain->prev->semanticVal = qrep_intersectObjLists((struct QREPKeyObjectList * )tkChain->prev->semanticVal, (struct QREPKeyObjectList * )tkChain->next->semanticVal);
			tkChain->prev->next = tkChain->next->next;
			next = tkChain->next->next;
			if(next!=NULL)
				next->prev = tkChain->prev;
			
			qrep_freeToken(tkChain->next);
			qrep_freeToken(tkChain);
		}	
		tkChain = next;
	}
	return 0;
}

/* this function is always called on tokenChains that do notcontain parentheses nor AND tokens*/
int  qrep_reduceORMINUS(struct qrep_token *tkChain){ 
	struct qrep_token *next=NULL;
	while(tkChain!=NULL){
		next=tkChain->next;
		
		if(tkChain->type==QREP_OPERATOR && (tkChain->lexicalVal[0]=='+' || tkChain->lexicalVal[0]=='-')){
			struct QREPKeyObjectList * res = NULL;	
			/* first we check if previous and next tokens are Operands
			  if not we return an error	
			*/	
			if(tkChain->prev==NULL || tkChain->prev->type!=QREP_OPERAND ||tkChain->next==NULL || tkChain->next->type!=QREP_OPERAND ){
				anaPrint(ANA_ERR, "invalid operands %c operator\n", tkChain->lexicalVal[0] );
				return -1;
			}
			
			if( tkChain->lexicalVal[0] == '-')
				res = qrep_minusObjLists((struct QREPKeyObjectList*)tkChain->prev->semanticVal, (struct QREPKeyObjectList*)tkChain->next->semanticVal);
			if( tkChain->lexicalVal[0] == '+')
				res = qrep_uniteObjLists((struct QREPKeyObjectList*)tkChain->prev->semanticVal, (struct QREPKeyObjectList*)tkChain->next->semanticVal);
			
			tkChain->prev->semanticVal = res;
			tkChain->prev->next = tkChain->next->next;
			next=tkChain->next->next;
			if(next!=NULL)
				next->prev = tkChain->prev;	
			
			qrep_freeToken(tkChain->next);
			qrep_freeToken(tkChain);
		}
		tkChain = next;
	}
	return 0;
}

int  qrep_evalTokenChain(struct qrep_token *tkChain){ 

	anaPrint(ANA_ERR, "evalTkenChain : evaluating chain %s\n", tkChain->lexicalVal);
	if(tkChain==NULL)
		return -1;
	/* first we reduce the expressions between parentheses */
	if (qrep_reduceParentheses(tkChain)<0){
		anaPrint(ANA_ERR, "parentheses problem\n");
		return -1;
	}	
	
	/* then we reduce the AND(*) rule*/
	if (qrep_reduceAND(tkChain)<0){
		anaPrint(ANA_ERR, "AND problem\n");
		return -1;
	}

	/* and finally we reduce the OR, MINUS rules */
	if (qrep_reduceORMINUS(tkChain)<0){
		anaPrint(ANA_ERR, "OR,MINUS problem\n");
		return -1;
	}
	
	return 0;
}


struct qrep_token *  qrep_tokenize(struct QREP *rep, char *expr){
	int offset=0;
	int bufferOffset=0;
	struct qrep_token *tokList=NULL;
	struct qrep_token *predecessor=NULL;
	int exprLen=0;
	char buf[QREP_BUFSIZE];
	char c=1;

	if(expr==NULL)
		return NULL;
	
	exprLen=strlen(expr)+1;	
	memset(buf,0,QREP_BUFSIZE);
	while(offset<exprLen){
		c = expr[offset];
		
		if(c == 0){
			struct qrep_token *tmpTok = NULL;
			if(bufferOffset>0){ /*we create an operand token*/
				tmpTok = (struct qrep_token *) malloc(sizeof(struct qrep_token));
				tmpTok->type = QREP_OPERAND;
				tmpTok->lexicalVal = (char *) malloc(bufferOffset+1);
				tmpTok->lexicalVal[bufferOffset] = 0;
				strncpy(tmpTok->lexicalVal, buf, bufferOffset);
				
				/* and the semantic one */
				tmpTok->semanticVal = (struct QREPKeyObjectList *) qrep_getCopyObjList(rep, tmpTok->lexicalVal);
				
				if(predecessor==NULL)
					tokList = tmpTok;
				else {
					predecessor->next = tmpTok;
					tmpTok->prev = predecessor;
				}
				predecessor = tmpTok;
				/*we clear the buffer and reset the offset to 0*/
				memset(buf,0,QREP_BUFSIZE);
				bufferOffset=0;
			}
			return tokList;
		}
				
		if(c =='+' || c =='-' || c =='*' || c =='(' || c ==')' ) {
			
				struct qrep_token *tmpTok = NULL;
				struct qrep_token *tmpTok2 = NULL;
				if(bufferOffset>0){ /*we create an operand token*/
					tmpTok = (struct qrep_token *) malloc(sizeof(struct qrep_token));
					tmpTok->type = QREP_OPERAND;
					/* we set the lexical value*/
					tmpTok->lexicalVal = (char *) malloc(bufferOffset+1);
					tmpTok->lexicalVal[bufferOffset] = 0;
					strncpy(tmpTok->lexicalVal, buf, bufferOffset);
					
					/* and the semantic one */
					tmpTok->semanticVal = (struct QREPKeyObjectList *) qrep_getCopyObjList(rep, tmpTok->lexicalVal);
											
					if(predecessor==NULL)
						tokList = tmpTok;
					else {
						predecessor->next = tmpTok;
						tmpTok->prev = predecessor;
					}
					predecessor = tmpTok;
					/*we clear the buffer and reset the offset to 0*/
					memset(buf,0,QREP_BUFSIZE);
					bufferOffset=0;
				}
				
				tmpTok2 = (struct qrep_token *) malloc(sizeof(struct qrep_token));
				tmpTok2->type = QREP_OPERATOR;
				tmpTok2->lexicalVal = (char *) malloc(1);
				tmpTok2->lexicalVal[0] = c;
				tmpTok2->semanticVal=NULL;
				
				if(predecessor==NULL)
					tokList = tmpTok2;
				else {
					predecessor->next = tmpTok2;
					tmpTok2->prev = predecessor;
				}
				
				predecessor = tmpTok2;	
		}			
		else
			if(bufferOffset<QREP_BUFSIZE){
				buf[bufferOffset]=c;
				bufferOffset++;
			}
			
		offset++;
	}
	/* if we end up here the expression is too long, we retunr NULL*/
	return NULL;
}	

void  qrep_displayTokenChain(struct qrep_token *tokenChain){
	struct QREPKeyObjectList *tmp;
	while(tokenChain!=NULL){
		if(tokenChain->type==QREP_OPERAND)
			anaPrint(ANA_DEBUG, "%s [ ", tokenChain->lexicalVal);
		else
			anaPrint(ANA_DEBUG, "%c [ ", tokenChain->lexicalVal[0]);	
		
		tmp = (struct QREPKeyObjectList*)tokenChain->semanticVal;
		while(tmp!=NULL){
			tmp=tmp->next;
		}
		
		tokenChain = tokenChain->next;
		anaPrint(ANA_DEBUG, "] --> ");
	
	}
	anaPrint(ANA_ERR, "\n");
}


struct QREPKeyObjectList *  qrep_getCopyObjList( struct QREP *rep , char *keyword){
	
	struct QREPKeyObjectList *original= QREPfindByKey(rep, keyword); 
	struct QREPKeyObjectList *copy=NULL;
	struct QREPKeyObjectList *predec=NULL;
	struct QREPKeyObjectList *tmp=NULL;
	/* mow we make a copy */
	predec = NULL;
	
	while(original!=NULL){
		tmp = (struct QREPKeyObjectList *) malloc(sizeof(struct QREPKeyObjectList));
		tmp->entry = original->entry;
		if(predec==NULL)
			copy = tmp;
		else{
			predec->next=tmp;
		}
		predec = tmp;		
		original = original->next;
	}
	
	
	return copy;
}


/* this function gives back a list containing elements common to list and list2

ATTention : this function frees list1 and list2 at the end (since it is a reduction(YACC)), therefore list1 and list2 must be
copies of the findByKey(struct QREP *rep, char *keyword) query

 */
struct QREPKeyObjectList *  qrep_intersectObjLists(struct QREPKeyObjectList *list1 , struct QREPKeyObjectList *list2) {
	
	struct QREPKeyObjectList *tmp1;
	struct QREPKeyObjectList *tmp2;
	struct QREPKeyObjectList *tmp2_predec;
	struct QREPKeyObjectList *tmp1_next;
	struct QREPKeyObjectList *tmp2_next;
	struct QREPKeyObjectList *res = NULL;
	int added;
	
	if(list1==NULL || list2 == NULL)
		return NULL;
	tmp1 = list1;
	while(tmp1!=NULL){
		tmp2=list2;
		tmp2_predec=NULL;
		added=0;
		tmp1_next = tmp1->next;
		while(tmp2!=NULL && !added){
			if(tmp1->entry==tmp2->entry){
				tmp2_next=tmp2->next;
				/* we add tmp1 to res*/
				added = 1;
				tmp1->next = res;
				res = tmp1;
				/*we free tmp2 */
				if(tmp2_predec == NULL)
					list2 = tmp2->next;
				else
					tmp2_predec->next = tmp2->next;
				free(tmp2);
				
			}
			else {
				tmp2_predec = tmp2;
				tmp2=tmp2->next;
			}		
		}
		/*if not added we free tmp1*/
		if(!added)
			free(tmp1);
		
		tmp1=tmp1_next;	
	}
	
	/*now we free the elements of list2 that didn't match with tmp1 */
	while(list2!=NULL){
		tmp1=list2->next;
		free(list2);
		list2=tmp1;
	}
	return res;	
}

/* list1 minus list2 i.e  the order of parameters is important here*/
struct QREPKeyObjectList *  qrep_minusObjLists(struct QREPKeyObjectList *list1 , struct QREPKeyObjectList *list2) {
	struct QREPKeyObjectList *tmp1;
	struct QREPKeyObjectList *tmp2;
	struct QREPKeyObjectList *tmp1_predec;
	struct QREPKeyObjectList *tmp1_next;
	struct QREPKeyObjectList *tmp2_predec;
	int removed;
	
	if(list2 == NULL)
		return list1;
	
	tmp1=list1;
	tmp1_predec=NULL;
	while(tmp1!=NULL){
		tmp2=list2;
		tmp2_predec=NULL;
		tmp1_next=tmp1->next;
		removed = 0;
		while(tmp2!=NULL && !removed){
			if(tmp1->entry==tmp2->entry){
				/* we remove tmp1 */
				if(tmp1_predec==NULL)
					list1=tmp1->next;
				else
					tmp1_predec->next=tmp1->next;
				
				/*we also remove tmp2*/
				if(tmp2_predec==NULL)
					list2=tmp2->next;
				else
					tmp2_predec->next = tmp2->next;
				free(tmp2);	
				
				removed=1;							
			}
			else{
				tmp2_predec=tmp2;	
				tmp2=tmp2->next;
			}
		}
		if(!removed)
			tmp1_predec=tmp1;
		tmp1=tmp1_next;
	
	}
	return list1;
}


/* this function unites the 2 lists without duplicates, 
 I didn't use minus in this function since it frees list2 and it shoud stay that way*/

struct QREPKeyObjectList *  qrep_uniteObjLists(struct QREPKeyObjectList *list1 , struct QREPKeyObjectList *list2) {
	struct QREPKeyObjectList *tmp1=NULL;
	struct QREPKeyObjectList *tmp2=NULL;
	struct QREPKeyObjectList *tmp2_predec=NULL;
	struct QREPKeyObjectList *tmp2_next=NULL;
	
	if(list1==NULL)
		return list2;
	if(list2==NULL)	
		return list1;	
	tmp1 = list1;
	while(tmp1!=NULL){
		tmp2_predec=NULL;
		tmp2=list2;
		while(tmp2!=NULL ){
			if(tmp1->entry==tmp2->entry){
				tmp2_next = tmp2->next;
				if(tmp2_predec==NULL)
					list2=tmp2->next;
				else	
					tmp2_predec->next = tmp2->next;
				
				free(tmp2);
				tmp2=tmp2_next;
			}
			else{
				tmp2_predec=tmp2;
				tmp2=tmp2->next;
			}
		}
		tmp1=tmp1->next;
	}
	
	/*now we have list1 and list2 wihtout replicas, we can concatenate*/
		
	tmp1=list1;
	while(tmp1->next!=NULL)
		tmp1=tmp1->next;
	
	tmp1->next=list2;
	
	return list1;
}



int qrep_evalQuery( struct QREP *rep, char *buf, struct QREPKeyObjectList **res ){

	struct qrep_token *tokenChain = qrep_tokenize(rep, buf);
		
	if(qrep_evalTokenChain(tokenChain)<0){
		*res = NULL;
		return -1;
	}
	
	*res = (struct QREPKeyObjectList*)tokenChain->semanticVal;
	if(*res == NULL){
		qrep_freeToken(tokenChain);
		return -1;
	}
		
	qrep_freeToken(tokenChain);
	return 0;
}













