/* C/bricks/vlink/common/ana_vlink.c
 *  
 * implementation of the vlink
 * 
 * KVR entries: interface+vlink (on default) 
 * 
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 *
 */


#include "brick_template.h"
#include "../ana_vlink.h"

#ifdef KERNEL
#include <linux/module.h>
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Christophe Jelger");
//static char *cd = "kern://vlinkd_1"; /* default vlink data url*/
//static char *cc = "kern://vlinkc_1"; /* default vlink control url*/
//static char *nc = "kern://anaMinmex"; /* default minmex control url*/
//static char *debug = "ANA_DEBUG";
#endif

#define MODULE_NAME ana_vlink
char mymodename[]="ana_vlink";

/* Global variables common for kernel and userspace */
struct vlink_s *vlink_head = NULL;
unsigned char vl_index_next = 1; /* start with 1 for vlink_check_mac */
unsigned char vl_cliport;	/* actually used to create unique VMAC addresses */
struct vl_minmex *vl_mx_head = NULL;
unsigned char vl_mx_index_next = 1;


/*********************************************************************************/

void brick_exit(void) {

#ifdef USER
	anaPrint(ANA_NOTICE, "brick_exit called\n");
	user_vlink_exit();
#endif

#ifdef KERNEL
	anaPrint(ANA_NOTICE, "brick_exit called\n");
	kernel_vlink_exit();
#endif	
	
	vlink_clean_all();
}

int brick_start(void) {

	validate_init(mymodename);
#ifdef USER
	return user_vlink_init();
#endif

#ifdef KERNEL
        return kernel_vlink_init();
#endif

}

int vlink_cli_command(char *recvbuf, int len, struct sockaddr_in *replyto)
{
	char sendbuf[2048];
	char *str = NULL;
	char *token, *ptr;
	int i=0;
	struct vlink_s *cur = vlink_head;
	struct ifs_ptr *ifs;
	
	if (len == 0)
	{
		anaPrint(ANA_ERR, "user_vl_cli_command --> len NULL\n");
		return -1;
	}

	if (!recvbuf || !replyto)
	{
		anaPrint(ANA_ERR, "user_vl_cli_command --> recvbuf or replyto NULL!\n");
		return -1;
	}

	/* get command name */
	ptr = (char *)recvbuf;
	token = strsep(&ptr, " ");
	
	if (!token || len == 2)
	{
		strcpy(sendbuf, "--> cannot parse command name!\n");
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return -1;
	}
	
	anaPrint(ANA_DEBUG, "user_vl_cli_command --> received '%s' command\n", recvbuf);
	anaValidate("user_vl_cli_command --> received '%s' command\n", recvbuf);
	
	if ( !strncmp(token, "vlink_create", 12) )
	{
		/* next token must be the vlink ID */
		token = strsep(&ptr, " ");
		if (!token)
		{
			strcpy(sendbuf, "--> cannot parse vlink ID!\n");
			sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return -1;
		}
				
		i = atoi(token);
			
		/* next token, if any, is a string of keywords */
		token = strsep(&ptr, " ");
		
		if (i)
			str = vlink_create(i, token);
		else
			str = NULL; /* cannot get vlink_id */
		
		if (!str)
			strcpy(sendbuf, "--> vlink_create failed!\n");
		else
			sprintf(sendbuf, "--> created %s (with keywords: %s)\n", str, token ? token : "DEFAULT");
				
		anaValidate("sendbuf: %s\n", sendbuf);
		/* send reply */	
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return 1;
	}
	else if ( !strncmp(token, "vlink_del", 9) )
	{
		/* next token must be the vlink name */
		token = strsep(&ptr, " ");
		if (!token)
			strcpy(sendbuf, "--> cannot parse vlink name!\n");
		else if (strlen(token) > VL_NAME_SIZE)
			strcpy(sendbuf, "--> vlink name too long!\n");
		else
		{		
			i = vlink_del(token);
			
			if (i == VL_NOERR)
				sprintf(sendbuf, "--> deleted %s\n", token);
			else if (i == VL_ERR_NULL)
				sprintf(sendbuf, "--> vl_name NULL!\n");
			else if (i == VL_ERR_NOT_FOUND)
				sprintf(sendbuf, "--> %s not found!\n", token);
		}

		anaValidate("sendbuf: %s\n", sendbuf);
		/* send reply */	
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return 1;
	}
	else if ( !strncmp(token, "vlink_up", 8) )
	{
		/* next token must be the vlink name */
		token = strsep(&ptr, " ");
		if (!token)
			strcpy(sendbuf, "--> cannot parse vlink name!\n");
		else if (strlen(token) > VL_NAME_SIZE)
			strcpy(sendbuf, "--> vlink name too long!\n");
		else
		{
			i = vlink_up(token);
			
			if (i == VL_NOERR)
				sprintf(sendbuf, "--> turned on %s\n", token);
			else if (i == VL_ERR_NULL)
				sprintf(sendbuf, "--> vl_name NULL!\n");
			else if (i == VL_ERR_NOT_FOUND)
				sprintf(sendbuf, "--> %s not found!\n", token);
		}
		
		anaValidate("sendbuf: %s\n", sendbuf);
		/* send reply */	
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return 1;
	}
	else if ( !strncmp(token, "vlink_down", 8) )
	{
		/* next token must be the vlink name */
		token = strsep(&ptr, " ");
		if (!token)
			strcpy(sendbuf, "--> cannot parse vlink name!\n");
		else if (strlen(token) > VL_NAME_SIZE)
			strcpy(sendbuf, "--> vlink name too long!\n");
		else
		{
			i = vlink_down(token);
			
			if (i == VL_NOERR)
				sprintf(sendbuf, "--> turned off %s\n", token);
			else if (i == VL_ERR_NULL)
				sprintf(sendbuf, "--> vl_name NULL!\n");
			else if (i == VL_ERR_NOT_FOUND)
				sprintf(sendbuf, "--> %s not found!\n", token);
		}
		
		anaValidate("sendbuf: %s\n", sendbuf);
		/* send reply */	
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return 1;
	}
	else if ( !strncmp(token, "vlink_add_if", 12) ||
		  !strncmp(token, "vlink_rem_if", 12) )
	{
		int add = !strncmp(token, "vlink_add_if", 12);
		
		/* next token must be the vlink name */
		token = strsep(&ptr, " ");
		if (!token)
			strcpy(sendbuf, "--> cannot parse vlink name!\n");
		else if (strlen(token) > VL_NAME_SIZE)
			strcpy(sendbuf, "--> vlink name too long!\n");
		else
		{
			char *vl_name = token;
		
			/* next token must be the interface name */
			token = strsep(&ptr, " ");
			if (!token)
				strcpy(sendbuf, "--> cannot parse if_name!\n");
			else
			{
				if (add)
					i = vlink_add_if(vl_name, token);
				else
					i = vlink_del_if(vl_name, token);
			
				if (i == VL_NOERR)
				{
					if (add)
						sprintf(sendbuf, "--> added interface %s\n", token);
					else
						sprintf(sendbuf, "--> deleted interface %s\n", token);
				}
				else if (i == VL_ERR_NULL)
					sprintf(sendbuf, "--> vl_name or if_name NULL!\n");
				else if (i == VL_ERR_NOMEM)
					sprintf(sendbuf, "--> cannot allocate memory!\n");
				else if (i == VL_ERR_NOT_FOUND)
					sprintf(sendbuf, "--> vl_name or if_name not found!\n");
			}
		}

		anaValidate("sendbuf: %s\n", sendbuf);
		/* send reply */	
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return 1;
	}
	else if ( !strncmp(token, "vlink_add_udp", 13) ||
		  !strncmp(token, "vlink_rem_udp", 13) )
	{
		int add = !strncmp(token, "vlink_add_udp", 13);
		
		/* next token must be the vlink name */
		token = strsep(&ptr, " ");
		if (!token)
			strcpy(sendbuf, "--> cannot parse vlink name!\n");
		else if (strlen(token) > VL_NAME_SIZE)
			strcpy(sendbuf, "--> vlink name too long!\n");
		else
		{
			char *vl_name = token;
			char *ip, *portc;
			struct sockaddr_in peer;
		
			memset(&peer, 0, sizeof(struct sockaddr_in));
			peer.sin_family = AF_INET;
		
			/* next token must be the ipaddr */
			token = strsep(&ptr, ":");
			if (!token)
				strcpy(sendbuf, "--> cannot find ipaddr!\n");
			else
			{
				ip = token;
#ifdef USER
				if (!inet_aton(ip, &peer.sin_addr))
				{
					strcpy(sendbuf, "--> invalid ip address!\n");
					sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
					return 1;
				}
#endif
#ifdef KERNEL
				peer.sin_addr.s_addr = in_aton(ip); /* cannot check return value :-( */
#endif
				token = strsep(&ptr, " ");
				if (!token)
				{
					strcpy(sendbuf, "--> cannot parse port number!\n");
					sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
					return 1;
				}
				portc = token;
				
				peer.sin_port = htons(atoi(portc));
				
				if (add)
					i = vlink_add_udp(vl_name, &peer);
				else
					i = vlink_del_udp(vl_name, &peer);
			
				if (i == VL_NOERR)
				{
					if (add)
						sprintf(sendbuf, "--> added udp peer\n");
					else
						sprintf(sendbuf, "--> removed udp peer\n");
				}
				else if (i == VL_ERR_NULL)
					sprintf(sendbuf, "--> vl_name or udp peer NULL!\n");
				else if (i == VL_ERR_NOMEM)
					sprintf(sendbuf, "--> cannot allocate memory!\n");
				else if (i == VL_ERR_NOT_FOUND)
					sprintf(sendbuf, "--> vl_name or udp peer not found!\n");
			}
		}

		anaValidate("sendbuf: %s\n", sendbuf);
		/* send reply */	
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return 1;
	}
	else if ( !strncmp(token, "vlink_list", 10) )
	{
		int no_udp;
				
		str = sendbuf;
		sprintf(str, "\n--------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
		sprintf(str, " VLINK   | STATE | VLINK_ID |       VL_MAC      | IF LIST ...\n");
		str = sendbuf + strlen(sendbuf);
		sprintf(str, "--------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
		if (cur)
			sprintf(str, "%s", MAGIC_COOKIE); /* add magic cookie and send header */	
		else
			sprintf(str, "--> vlink table is empty\n");
			
		str = sendbuf + strlen(sendbuf);
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		
		if (!cur)
			return 1;
				
		while (cur)
		{
			str = sendbuf;			
			sprintf(str, " %s", cur->vl_name);
			str = sendbuf + strlen(sendbuf);
			
			if (cur->vl_idx < 10)
				sprintf(str, "    ");
			else if (cur->vl_idx < 100)
				sprintf(str, "   ");
			else
				sprintf(str, "  ");
									
			str = sendbuf + strlen(sendbuf);		
			
			sprintf(str, " %s   %8d   " ETH_FMT "   [", cur->vl_flags & IFF_UP ? " UP " : "DOWN", 
				ntohl(cur->vl_id), ETH_HEX(cur->vl_mac) );
			str = sendbuf + strlen(sendbuf);
			
			ifs = cur->vl_iflist;
			
			if (!ifs)
			{
				strcpy(str, "NULL]\n");
				str = sendbuf + strlen(sendbuf);
			}
			else
			{
				while (ifs)
				{
#ifdef KERNEL
					char *ifname = ((struct net_device *)ifs->ptr)->name;
#endif
#ifdef USER
					char *ifname = ((struct ifaddrs *)ifs->ptr)->ifa_name;
#endif
					sprintf(str, "%s ", ifname);
					str = sendbuf + strlen(sendbuf);
					ifs = ifs->next;
				}
				
				strcpy(str-1, "]\n");
				str = sendbuf + strlen(sendbuf);
			}
		
			cur = cur->vl_next;
			
			/* we will send more */
			sprintf(str, "%s", MAGIC_COOKIE);
						
			/* send reply */	
			sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		}
		
		str = sendbuf;
		sprintf(str, "\n----------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
		sprintf(str, " VLINK   | Remote UDP peers ... \n");
		str = sendbuf + strlen(sendbuf);
		sprintf(str, "----------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
		/* we will send more */
		sprintf(str, "%s", MAGIC_COOKIE);
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		
		cur = vlink_head;
		no_udp = 0;
		
		while (cur)
		{
			struct udp_tun *tun = cur->vl_udplist;
#ifdef USER
			char ip_str[INET_ADDRSTRLEN];
#endif
			if (!tun)
			{
				cur = cur->vl_next;
				continue;
			}
		
			no_udp++;		
			str = sendbuf;			
			sprintf(str, " %s", cur->vl_name);
			str = sendbuf + strlen(sendbuf);
			
			if (cur->vl_idx < 10)
				sprintf(str, "    ");
			else if (cur->vl_idx < 100)
				sprintf(str, "   ");
			else
				sprintf(str, "  ");
									
			str = sendbuf + strlen(sendbuf);
#ifdef USER
			inet_ntop(AF_INET, &tun->udpdst.sin_addr, ip_str, INET_ADDRSTRLEN);
			sprintf(str, "%s:%u\n", ip_str, ntohs(tun->udpdst.sin_port));
#endif
#ifdef KERNEL
			sprintf(str, NIPQUAD_FMT ":%u\n", NIPQUAD(tun->udpdst.sin_addr), 
				ntohs(tun->udpdst.sin_port));
#endif
			str = sendbuf + strlen(sendbuf);
			
			tun = cur->vl_udplist->next;
			
			while (tun)
			{
#ifdef USER
				inet_ntop(AF_INET, &tun->udpdst.sin_addr, ip_str, INET_ADDRSTRLEN);
				sprintf(str, "            %s:%u\n", ip_str, ntohs(tun->udpdst.sin_port));
#endif
#ifdef KERNEL
				sprintf(str, "            " NIPQUAD_FMT ":%u\n", 
					NIPQUAD(tun->udpdst.sin_addr), ntohs(tun->udpdst.sin_port));
#endif
				str = sendbuf + strlen(sendbuf);
				tun = tun->next;
			}
			
			cur = cur->vl_next;
			
			sprintf(str, "%s", MAGIC_COOKIE); /* we always send more */
			sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		}
		
		str = sendbuf;
		if (!no_udp)
			sprintf(str, "--> no udp tunnels\n");
		else
			sprintf(str, "\n");
				
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
	}
	else if ( !strncmp(token, "send_dummy", 10) )
	{
		/* debuging only: used to send data from vlconfig */
		
		/* next token must be the vlink name */
		token = strsep(&ptr, " ");
		if (!token)
			strcpy(sendbuf, "--> cannot parse vlink name!\n");
		else if (strlen(token) > VL_NAME_SIZE)
			strcpy(sendbuf, "--> vlink name too long!\n");
		else
			vlink_send("dummy data, for testing only", 28, 0, token);
		
		/* send reply */
		sprintf(sendbuf, "--> trying to send dummy data to %s\n", token);
		anaValidate("sendbuf: %s\n", sendbuf);
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return 1;
	}
	else
	{
		strcpy(sendbuf, "--> Unknown command!\n");
		anaValidate("sendbuf: %s\n", sendbuf);
		/* send reply */	
		sendSocketWrapper(&vl_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
	}

	return 1;
}

void *copyCharChain(void *aux){
	char *aux2 = malloc(strlen((char*)aux)+1);
	strcpy(aux2, (char*)aux); 
	return aux2;
}

char* vlink_create(__u32 vl_id, char *keywords)
{
	struct vlink_s *cur, *prev;
	struct vlink_s *new = NULL;
	anaLabel_t bind;
	struct IDPinfo_s info;

	if (!vl_id)
	{
		anaPrint(ANA_ERR, "vlink_create --> vl_id 0 is not allowed!\n");
		return NULL;
	}
	
	if (vl_index_next == 255)
	{
		anaPrint(ANA_ERR, "vlink_create --> max number of vlinks reached (255)!\n");
		return NULL;
	}
	
	if (!vlink_head)
	{	
		new = malloc(sizeof(struct vlink_s));
	
		if (!new)
		{
			anaPrint(ANA_ERR, "vlink_create --> malloc failed!\n");
			return NULL;
		}
		
		vlink_head = new;
	}
	else
	{
		for (cur = vlink_head; cur; cur = cur->vl_next)
		{
			if (cur->vl_id == htonl(vl_id))
			{
				anaPrint(ANA_ERR, "vlink_create --> vl_id %d already used!\n", vl_id);
				return NULL;
			}
			
			prev = cur;
		}
		
		new = malloc(sizeof(struct vlink_s));
		
		if (!new)
		{
			anaPrint(ANA_ERR, "vlink_create --> malloc failed!\n");
			return NULL;
		}
		
		prev->vl_next = new;
	}

	sprintf(new->vl_name, "vlink%d", vl_index_next);
	new->vl_id = htonl(vl_id);	/* used to know to which vlink a packet belongs to */
	new->vl_idx = vl_index_next;
	new->vl_mac[0] = vl_index_next;
	new->vl_mtu = VLINK_MTU; /* init value */
	
	/* NEW: cli_port is used to make sure that multiple vlink instances create different vmac addresses */
	new->vl_mac[1] = vl_cliport;
	
	if (vl_index_next < 255)
		vl_index_next++;
	
	anaPrint(ANA_DEBUG, "Created vlink %s (vl_id %u)\n", new->vl_name, ntohl(new->vl_id));
	
	/* register vlink_bind callback function
	bind = allocateLabel_multi(vl_mx->minmex->labelSize);
	ret = anaL0multi_registerCallback(vl_mx->minmex, (anaCallback_t)vlink_bind, new->vl_name, 
					  copyCharChain, (freeAuxFunction)free, bind, NULL, IDP_PERM, 1); */
					  
	bind = anaL0_registerCallback((anaCallback_t)&vlink_bind, (void *) new->vl_name, 
					copyCharChain, free, 0, IDP_PERM, NOTHREAD);
					  
	if (!bind) 
	{
		anaPrint(ANA_ERR, "Unable to register Callback IDP with ANA minmex\n");
		vlink_del(new->vl_name);
		return NULL;
	}

	/* set IDPinfo before we publish it in the KVR */
	/* maybe in the future we can use the context and service to provide some useful info */
	/* for now it's all set to 0 by the memset */
		
	memset(&info, 0, sizeof(struct IDPinfo_s));
	info.label = bind;
	info.viewer = 0; /* public */
	info.owner = MY_HANDLE;
	info.ICflag = 0; /* the vlink does not really provide an IC in the ANA sense */
	info.permFlag = 1;
		
	info.MTU = 0; /* IDP is NOT for sending, it's only for binding */	
	info.chanType = 0; /* no real sense with the vlink */
	info.IDPtype = BRICK;
	info.binding.brick.handle = MY_HANDLE;
	info.binding.brick.outputIDP = 0; /* vlink natively forwards over medium or UDP tunnel */

	if (anaL0_setIDPinfo(&info) < 0)
		anaPrint(ANA_DEBUG, "vlink_create --> setIDPinfo failed\n");
	
	/* publish vlink entry in KVR + store bind IDP */
	new->binding = bind;
	if (keywords)
		vlink_publish(new, keywords);
	else
		vlink_publish(new, "interface+vlink");

	return new->vl_name;
}

void vlink_unbind(void *data, int len, anaLabel_t input, void *aux)
{
	struct vlink_s *vlink = NULL;
	char *vl_name = (char *)aux; /* fct arg should be vlink name */
	int exp_len;
	anaLabel_t upcall, bind;
	struct IDPinfo_s info;
	
	if (!vl_name)
	{
		anaPrint(ANA_ERR, "vlink_unbind --> vl_name NULL!\n");
		return;
	}
	
	if (!data)
	{
		anaPrint(ANA_ERR, "vlink_unbind --> data ptr NULL!\n");
		return;
	}

	exp_len = sizeof(anaLabel_t);

	if (len < exp_len)
	{
		anaPrint(ANA_ERR, "vlink_unbind --> data len too small (%d bytes)!\n", len);
		return;
	}

	anaPrint(ANA_DEBUG, "vlink_unbind --> unbind request received for %s\n", vl_name);
	vlink = vlink_lookup(vl_name);
	
	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_unbind --> cannot find %s!\n", vl_name);
		return;
	}
	
	if (!vlink->upcall)
	{
		anaPrint(ANA_ERR, "vlink_unbind --> %s not bound!\n", vl_name);
		return;
	}
	
	upcall = *((anaLabel_t *) data);
	
	if ( upcall != vlink->upcall )
	{
		anaPrint(ANA_ERR, "vlink_unbind --> upcall provided different from vlink->upcall!\n");
		return;
	}
	
	bind = anaL0_registerCallback((anaCallback_t)&vlink_bind, (void *) vlink->vl_name, 
					copyCharChain, free, 0, IDP_PERM, NOTHREAD);
					  
	if (!bind) 
	{
		anaPrint(ANA_ERR, "vlink_unbind --> unable to register vlink_bind callback!\n");
		return;
	}

	/* copy IDPinfo to new IDP (unbind to bind) */
	if (anaL0_getIDPinfo(vlink->binding, vlink->caller, &info) < 0)
		anaPrint(ANA_ERR, "vlink_unbind --> cannot get IDPinfo for unbind idp %x\n", vlink->binding);
	else
	{
		info.label = bind;
		info.viewer = 0; /* public IDP */
		
		if (anaL0_setIDPinfo(&info) < 0)
			anaPrint(ANA_DEBUG, "vlink_unbind --> setIDPinfo failed for bind idp %x\n", bind);
	}

	/* unregister vlink_sendto and vlink_unbind callbacks */
	anaL0_unregisterCallback(vlink->sendto, vlink->caller);
	anaL0_unregisterCallback(vlink->binding, vlink->caller);
	
	/* update binding IDP */
	vlink->binding = bind;
	
	/* publish vlink entry in KVR */
	vlink_publish(vlink, "interface+vlink");
	
	/* reset obsolete IDPs + brick handle (caller) */
	vlink->sendto = 0;
	vlink->upcall = 0;
	vlink->caller = 0;
}

void vlink_bind(void *data, int len, anaLabel_t input, void *aux)
{
	struct vlink_s *vlink = NULL;
	char *vl_name = (char *)aux; /* fct arg should be vlink name */
	int exp_len;
	anaLabel_t replyto, upcall, sendto, unbind;
	anaHandle_t caller;
	char *replybuf; /* for reply message */
	unsigned char *vl_mac; /* for passing vMAC address */
	struct IDPinfo_s info;
	
	if (!vl_name)
	{
		anaPrint(ANA_ERR, "vlink_bind --> vl_name NULL!\n");
		return;
	}
	
	if (!data)
	{
		anaPrint(ANA_ERR, "vlink_bind --> data ptr NULL!\n");
		return;
	}

	exp_len = 2*sizeof(anaLabel_t) + sizeof(anaHandle_t);

	if (len < exp_len)
	{
		anaPrint(ANA_ERR, "vlink_bind --> data len too small (received %d, expected %d bytes)! \n", len, exp_len);
		return;
	}

	anaPrint(ANA_DEBUG, "vlink_bind --> bind request received for %s\n", vl_name);
	vlink = vlink_lookup(vl_name);
	
	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_bind --> cannot find %s!\n", vl_name);
		
		/* TODO --> entry no longer valid: remove from KVR */
		
		return;
	}
	
	if (vlink->upcall)
	{
		anaPrint(ANA_ERR, "vlink_bind --> %s already bound!\n", vl_name);
		return;
	}	
	
	/* before doing anything, check that we can allocate replybuf */
	
	len = 2*sizeof(anaLabel_t) + sizeof(anaHandle_t) + VL_MACLEN;
	replybuf = malloc(len);
	if (!replybuf)
	{
		anaPrint(ANA_ERR, "vlink_bind --> cannot allocate reply buffer!\n");
		return;
	}
	
	/* data sent by the brick binding to the vlink */
	replyto = *((anaLabel_t *) data);
	upcall = *((anaLabel_t *) ((char *)data + sizeof(anaLabel_t) ));
	caller = *((anaHandle_t *) ((char *)data + 2*sizeof(anaLabel_t) ));
	
	/* IDP for sending to the vlink */	
	sendto = anaL0_registerCallback((anaCallback_t)&vlink_send, (void *) vlink->vl_name, 
					copyCharChain, free, caller, IDP_PERM, NOTHREAD);
					 
	if (!sendto) 
	{
		anaPrint(ANA_ERR, "vlink_bind --> unable to register vlink_send callback!\n");
		free(replybuf);
		return;
	}
	
	memset(&info, 0, sizeof(struct IDPinfo_s));
	info.label = sendto;
	info.viewer = caller; /* private IDP */
	info.owner = MY_HANDLE;
	info.ICflag = 1; /* the vlink does not really provide an IC in the ANA sense */
	info.permFlag = 1;
		
	info.MTU = vlink->vl_mtu;	
	info.chanType = 0; /* no real sense with the vlink */
	info.IDPtype = BRICK;
	info.binding.brick.handle = MY_HANDLE;
	info.binding.brick.outputIDP = 0; /* vlink natively forwards over medium or UDP tunnel */

	if (anaL0_setIDPinfo(&info) < 0)
		anaPrint(ANA_DEBUG, "vlink_bind --> setIDPinfo failed for sendto idp %x\n", sendto);
	
	/* IDP for unbinding with the vlink */
	unbind = anaL0_registerCallback((anaCallback_t)&vlink_unbind, (void *) vlink->vl_name, 
					copyCharChain, free, caller, IDP_PERM, NOTHREAD);
					 
	if (!unbind) 
	{
		anaPrint(ANA_ERR, "vlink_bind --> unable to register vlink_unbind callback!\n");
		anaL0_unregisterCallback(sendto, caller);
		free(replybuf);
		return;
	}
	
	/* copy IDPinfo to new IDP (bind to unbind) */
	if (anaL0_getIDPinfo(vlink->binding, 0, &info) < 0)
		anaPrint(ANA_ERR, "vlink_bind --> cannot get IDPinfo for bind idp %x\n", vlink->binding);
	else
	{
		info.label = unbind;
		info.viewer = caller;
		
		if (anaL0_setIDPinfo(&info) < 0)
			anaPrint(ANA_DEBUG, "vlink_bind --> setIDPinfo failed for unbind idp %x\n", unbind);
	}
			
	/* update vlink object */
	
	vlink->sendto = sendto;
	anaL0_unregisterCallback(vlink->binding, 0); /* unregister bind IDP */
	
	vlink_unpublish(vlink); /* remove (bind) entry from KVR */
	vlink->binding = unbind; 
	vlink->upcall = upcall;
	vlink->caller = caller;
	
	/* build reply message and send it */
	
	memcpy(replybuf, &vlink->sendto, sizeof(anaLabel_t));
	memcpy(replybuf+sizeof(anaLabel_t), &vlink->binding, sizeof(anaLabel_t));
	memcpy(replybuf+2*sizeof(anaLabel_t), &myHandle, sizeof(anaLabel_t));
	vl_mac = (unsigned char*) (replybuf + 2*sizeof(anaLabel_t) + sizeof(anaHandle_t));
	memcpy(vl_mac, vlink->vl_mac, VL_MACLEN);
	
	anaPrint(ANA_DEBUG, "vlink_bind --> bind successful for %s\n", vlink->vl_name);
	anaL0_send(replyto, replybuf, len);
	free(replybuf);
}

int vlink_demux(char *data, int len, int iif, struct sockaddr_in *peer) 
{
	__u32 vl_id;
	struct vlink_s *vlink = vlink_head;
	/* Note: data was already checked to be longer than 4 bytes */	

	vl_id = *((__u32 *)data); /* in network order */
	
	if (!vlink_head)
	{
		anaPrint(ANA_ERR, "vlink_demux --> vlink_head is NULL!\n");
		return VL_ERR_NULL;
	}

	while (vlink)
	{
		if (vlink->vl_id == vl_id) {
			vlink_upcall(vlink, data+4, len-4, iif, peer);
		}
		vlink = vlink->vl_next;
	}
	
	return VL_NOERR;
}
	
int vlink_upcall(struct vlink_s *vlink, char *data, int len, int iif, struct sockaddr_in *peer)
{
	struct ifs_ptr *ifs;
	struct udp_tun *tun;
	
	anaValidate("vlink_upcall called with data len %i\n", len);
	if (!vlink->upcall)
	{
		anaPrint(ANA_ERR, "vlink_upcall --> %s (vl_id %u) no upcall!\n", vlink->vl_name, ntohl(vlink->vl_id));
		return -1;
	}
	
	if ( !(vlink->vl_flags & IFF_UP) )
	{
		anaPrint(ANA_ERR, "vlink_upcall --> %s DOWN!\n", vlink->vl_name);
		return 1;
	}

	if (!peer && iif == -1)
	{
		/* node-local delivery between minmexes */
	
		anaL0_send(vlink->upcall, data, len);
		anaPrint(ANA_ERR, "vlink_upcall --> %s (vl_id %u) [from local minmex]\n", vlink->vl_name, ntohl(vlink->vl_id));
		return 1;
	}
	
	if ( !(vlink->vl_iflist) && !(vlink->vl_udplist) )
	{
		/* data is sent by another node but vlink has no UDP peers and no interface set */
		
		anaPrint(ANA_ERR, "vlink_upcall --> %s vl_iflist and vl_udplist are empty!\n", vlink->vl_name);
		return -1;
	}
	
	if (peer) /* data is coming from a UDP tunnel */
	{
#ifdef USER
		char ip_str[INET_ADDRSTRLEN];
#endif
		tun = vlink->vl_udplist;
				
		while (tun)
		{
			if (!memcmp(&tun->udpdst, peer, sizeof(struct sockaddr_in)))
			{
#ifdef USER
				if (anaDebugLevel >= ANA_DEBUG)
					inet_ntop(AF_INET, &peer->sin_addr, ip_str, INET_ADDRSTRLEN);
				anaPrint(ANA_DEBUG, "vlink_upcall --> %s (vl_id %u) [from %s:%u]\n", 
					 vlink->vl_name, ntohl(vlink->vl_id), ip_str, ntohs(peer->sin_port));
#endif
#ifdef KERNEL
				anaPrint(ANA_DEBUG, "vlink_upcall --> %s (vl_id %u) [from " NIPQUAD_FMT ":%u]\n", 
					 vlink->vl_name, ntohl(vlink->vl_id), NIPQUAD(peer->sin_addr), ntohs(peer->sin_port));
#endif
				anaL0_send(vlink->upcall, data, len);
				return 1;
			}
			
			tun = tun->next;				
		}
#ifdef USER
		if (anaDebugLevel >= ANA_ERR)
			inet_ntop(AF_INET, &peer->sin_addr, ip_str, INET_ADDRSTRLEN);
		anaPrint(ANA_ERR, "vlink_upcall --> %s (vl_id %u) recv data from invalid udp peer %s:%u!\n",
			 vlink->vl_name, ntohl(vlink->vl_id), ip_str, ntohs(peer->sin_port) );
#endif
#ifdef KERNEL
		anaPrint(ANA_ERR, "vlink_upcall --> %s (vl_id %u) recv data from invalid udp peer " NIPQUAD_FMT ":%u !\n",
			 vlink->vl_name, ntohl(vlink->vl_id), NIPQUAD(peer->sin_addr), ntohs(peer->sin_port) );
#endif
		return -1;
	}
			
	/* if iif > 0, check that packet is received from an */
	/* interface that belongs to the vlink */
			
	if (iif > 0)
	{
		ifs = vlink->vl_iflist;
			
		while (ifs)
		{
			if (iif == ifs->if_index)
			{
#ifdef KERNEL
				char *ifname = ((struct net_device *)ifs->ptr)->name;
#endif
#ifdef USER
				char *ifname = ((struct ifaddrs *)ifs->ptr)->ifa_name;
#endif
				anaPrint(ANA_DEBUG, "vlink_upcall --> %s (vl_id %u) [from %s]\n",
					 vlink->vl_name, ntohl(vlink->vl_id), ifname);

				anaL0_send(vlink->upcall, data, len);
				return 1;
			}
				
			ifs = ifs->next;
		}
			
		anaPrint(ANA_ERR, "vlink_upcall --> %s (vl_id %u) recv data from invalid if_index %d!\n", 
		 	 vlink->vl_name, ntohl(vlink->vl_id), iif);
	}
	
	return -1;
}

struct vlink_s* vlink_lookup(char *vl_name)
{
	struct vlink_s *cur = vlink_head;
	
	if (!vl_name)
	{
		anaPrint(ANA_ERR, "vlink_lookup --> vl_name is NULL!\n");
		return NULL;
	}
	
	if (!vlink_head) /* vlink list is empty */
	{
		anaPrint(ANA_ERR, "vlink_lookup --> vlink_head NULL!\n");
		return NULL;
	}
	
	while(cur)
	{
		if (!strcmp(vl_name, cur->vl_name))
			return cur;
						
		cur = cur->vl_next;
	}
		
	return NULL; /* vlink not found */
}

int vlink_del(char *vl_name)
{
	struct vlink_s *cur = vlink_head;
	struct vlink_s *prev = NULL;
			
	if (vl_name == NULL)
	{
		anaPrint(ANA_ERR, "vlink_del --> vl_name NULL!\n");
		return VL_ERR_NULL;
	}
	
	if (!vlink_head)
	{
		anaPrint(ANA_ERR, "vlink_del(%s) --> vlink_head NULL!\n", vl_name);
		return VL_ERR_NOT_FOUND; /* list empty */
	}
	
	/* below we cannot use vlink_lookup because we need to find "prev" */

	while(cur)
	{
		if (!strcmp(vl_name, cur->vl_name))
		{			
			if (prev)
				prev->vl_next = cur->vl_next;
			else
				vlink_head = cur->vl_next;
			
			anaPrint(ANA_NOTICE, "vlink_del --> removing %s\n", cur->vl_name);
			vlink_clean(cur);
			return VL_NOERR;
		}
		
		prev = cur;
		cur = cur->vl_next;
	}
	
	anaPrint(ANA_ERR, "vlink_del --> cannot find vlink %s!\n", vl_name);
	return VL_ERR_NOT_FOUND; /* not found */
}

void vlink_clean(struct vlink_s *vlink)
{
	/* !!! this function does not maintain the linked list !!! */	
	/* i.e. the caller (e.g. vlink_del) has to take care of it */
	
	/* free all allocated memory and clean object */
	
	struct ifs_ptr *ifcur, *ifnext;
	struct udp_tun *udpcur, *udpnext;
	
	if (!vlink)
		return;
	
	vlink->vl_flags &= ~IFF_UP; /* vlink is down (stops data flow) */
	
	if (!vlink->upcall)
		vlink_unpublish(vlink); /* remove vlink entry from KVR */
			
	if (vlink->sendto)
		anaL0_unregisterCallback(vlink->sendto, vlink->caller);
	if (vlink->binding)
		anaL0_unregisterCallback(vlink->binding, vlink->caller);
			
	if (vlink->vl_iflist) /* free interface list */
	{
		ifcur = vlink->vl_iflist;
				
		while (ifcur)
		{
			ifnext = ifcur->next;
			free(ifcur);
			ifcur = ifnext;
		}
	}
	
	if (vlink->vl_udplist) /* free list of udp tunnels */
	{
		udpcur = vlink->vl_udplist;
				
		while (udpcur)
		{
			udpnext = udpcur->next;
			free(udpcur);
			udpcur = udpnext;
		}
	}
			
	anaPrint(ANA_DEBUG, "vlink_clean --> freeing %s\n", vlink->vl_name);
	free(vlink);
}

void handleUnpublishReply(char *data, int len, anaLabel_t input, void *aux)
{
	if (anaL1_isError(data)) 
		anaPrint(ANA_ERR, "ana_vlink --> Unpublish error!\n");
	else
		anaPrint(ANA_DEBUG, "ana_vlink --> Unpublish succeeded\n");
		
	anaL0_unregisterCallback(input, 0);
}

void handlePublishReply(char *data, int len, anaLabel_t input, void *aux)
{
	char *name = NULL;
	
	anaL1_decResponse(data, XRP_CLASS_SRC_SRV, 0, (void**)&name);

	if (anaL1_isError(data)) /*if ERROR command */
	{
		if (name)
			anaPrint(ANA_ERR, "ana_vlink --> publish of entry %s failed!\n",name);
		else
			anaPrint(ANA_ERR, "ana_vlink --> publish error!\n");
		
		anaL0_unregisterCallback(input, 0);		
		return;
	}
	
	if (name)
		anaPrint(ANA_DEBUG, "ana_vlink --> publish of entry %s succeeded\n",name);
	else
		anaPrint(ANA_DEBUG, "ana_vlink --> publish succeeded\n");
		
	anaL0_unregisterCallback(input, 0);
}

int vlink_unpublish(struct vlink_s *vlink)
{
	int lenMsg, ret;
	struct context_s ctxt;
	struct service_s srv;
	xrpMsg_t msg;
	
	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_unpublish --> vlink NULL!\n");
		return -1;
	}

	msg = anaL1_allocateMessage();
	
	if (!msg)
	{
		anaPrint(ANA_ERR, "vlink_unpublish --> cannot allocate xrpmsg!\n");
		return -1;
	}
	
	ctxt.value = ".";
	ctxt.valueLen = strlen(".")+1;
	srv.value = vlink->vl_name;
	srv.valueLen = strlen(vlink->vl_name)+1;
	
	lenMsg = anaL1_encCompartmentUnpublish(msg, vlink->binding, &ctxt, &srv);
	
	if (lenMsg < 0)
	{
		anaPrint(ANA_ERR, "vlink_unpublish --> encoding unpublish request failed!\n");
		free(msg);
		return -1;
	}

	ret = anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t) handleUnpublishReply, NULL, NULL, NULL, NOTHREAD); 
	if (ret < 0)
	{
		anaPrint(ANA_ERR, "vlink_unpublish --> sending unpublish request failed!\n");
		free(msg);
		return -1;
	}

	anaPrint(ANA_DEBUG, "vlink_unpublish --> sent %s unpublish\n", vlink->vl_name);
	free(msg);
	return 1;
}

int vlink_publish(struct vlink_s *vlink, char *description)
{
	int lenMsg, ret;
	struct context_s ctxt;
	struct service_s srv;
	xrpMsg_t msg;
	
	char *newDesc;
	int lenNewDesc;
		
	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_publish --> vlink NULL!\n");
		return -1;
	}
	
	if (!description)
		newDesc = vlink->vl_name;
	else
	{
		lenNewDesc = strlen(description) + 2 + strlen(vlink->vl_name); /* extra character for the '+' */
		newDesc = (char *) malloc(lenNewDesc);
		strcat(newDesc, vlink->vl_name);
		strcat(newDesc, "+");
		strcat(newDesc, description);
	}
	
	msg = anaL1_allocateMessage();
	
	if (!msg)
	{
		anaPrint(ANA_ERR, "vlink_publish --> cannot allocate xrpmsg!\n");
		return -1;
	}
	
	ctxt.value = ".";
	ctxt.valueLen = 2;
	srv.value = newDesc;
	srv.valueLen = strlen(newDesc)+1;
	
	lenMsg = anaL1_encCompartmentPublish(msg, vlink->binding, &ctxt, &srv, IDP_PERM);
	if (lenMsg < 0)
	{
		anaPrint(ANA_ERR, "vlink_publish --> encoding publish request failed!\n");
		free(msg);
		return -1;
	}

	ret = anaL1_requestReply(NODE_LABEL, msg, lenMsg, (anaCallback_t)handlePublishReply, NULL, NULL, NULL, NOTHREAD); 
	if (ret < 0)
	{
		anaPrint(ANA_ERR, "vlink_publish --> sending publish request failed!\n");
		free(msg);
		return -1;
	}

	anaPrint(ANA_DEBUG, "vlink_publish(%s) --> key-val publish ok\n", vlink->vl_name);
	free(msg);
	return 1;
}

void vlink_clean_all(void)
{
	struct vlink_s *cur = vlink_head;
	struct vlink_s *next;
	
	anaPrint(ANA_NOTICE, "vlink_clean_all --> cleaning vlinks\n");
	
	while (cur)
	{
		next = cur->vl_next;
		vlink_clean(cur);
		cur = next;
	}
	
	vlink_head = NULL;
}

int vlink_up(char *vl_name)
{
	struct vlink_s *vlink = NULL;

	if (!vl_name)
	{
		anaPrint(ANA_ERR, "vlink_up --> vl_name NULL!\n");
		return VL_ERR_NULL;
	}

	vlink = vlink_lookup(vl_name);

	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_up --> cannot find %s!\n", vl_name);
		return VL_ERR_NOT_FOUND;
	}
		
	vlink->vl_flags |= IFF_UP; /* set IFF_UP flag */
	anaPrint(ANA_NOTICE, "vlink_up(%s)\n", vl_name);
	
	return VL_NOERR;
}

int vlink_down(char *vl_name)
{
	struct vlink_s *vlink = NULL;

	if (!vl_name)
	{
		anaPrint(ANA_ERR, "vlink_down --> vl_name NULL!\n");
		return VL_ERR_NULL;
	}

	vlink = vlink_lookup(vl_name);

	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_down --> cannot find %s!\n", vl_name);
		return VL_ERR_NOT_FOUND;
	}
	
	vlink->vl_flags &= ~IFF_UP; /* unset IFF_UP flag */
	anaPrint(ANA_NOTICE, "vlink_down(%s)\n", vl_name);
	
	return VL_NOERR;
}

int vlink_add_if(char *vl_name, char *if_name)
{
	int if_index = -1;
	struct vlink_s *vlink = NULL;
	struct ifs_ptr *cur, *prev = NULL;
	unsigned char *newmac;
	uint16_t mtu = 0;
#ifdef KERNEL
	struct net_device *dev;
#endif
#ifdef USER
	struct ifaddrs *pf_if = pf_list;
	struct ifreq if_info;
	struct sockaddr_ll *ll;
#endif
	if (!vl_name || !if_name)
	{
		anaPrint(ANA_ERR, "vlink_add_if --> vl_name or if_name NULL!\n");
		return VL_ERR_NULL;
	}

	vlink = vlink_lookup(vl_name);

	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_add_if --> cannot find %s!\n", vl_name);
		return VL_ERR_NOT_FOUND;
	}

#ifdef USER
	while (pf_if)
	{
		if (!strcmp(if_name, pf_if->ifa_name))
		{
			if_index = if_nametoindex(if_name);
			ll = (struct sockaddr_ll *)pf_if->ifa_addr;
			
			if (ll->sll_halen == 6)
				newmac = ll->sll_addr; /* interface MAC address */
			else
				newmac = NULL;
			
			break;
		}
	
		pf_if = pf_if->ifa_next;
	}
#endif

#ifdef KERNEL
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	dev = dev_get_by_name(if_name);
#else
	/* not sure, whether it is ok to put init_net here */
	dev = dev_get_by_name(&init_net, if_name);
#endif
	if (dev)
	{
		if_index = dev->ifindex;
		
		if (dev->addr_len == 6)
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,14)
			newmac = dev->dev_addr;
#else
			newmac = dev->perm_addr;
#endif
		else
			newmac = NULL;
		
		mtu = dev->mtu; /* we need that later */
		dev_put(dev);
	}
#endif
	if (if_index <= 0)
	{
		anaPrint(ANA_NOTICE, "vlink_add_if(%s) --> cannot find %s!\n", vlink->vl_name, if_name);
		return VL_ERR_NOT_FOUND;
	}
	
	/* check if interface is already in vlink interface list */
	
	cur = vlink->vl_iflist;
	
	while (cur)
	{
		if (cur->if_index == if_index)
		{
			anaPrint(ANA_ERR, "vlink_add_if(%s) --> %s already in vl_iflist!\n", 
			 	 vlink->vl_name, if_name);
			return VL_NOERR;
		}
		
		prev = cur;
		cur = cur->next;
	}
	
	cur = malloc(sizeof(struct ifs_ptr));
	
	if (!cur)
	{
		anaPrint(ANA_ERR, "vlink_add_if --> malloc failed!\n");
		return VL_ERR_NOMEM;
	}
	
	cur->if_index = if_index;
	if (prev)
		prev->next = cur;
	else
		vlink->vl_iflist = cur;
#ifdef KERNEL
	cur->ptr = dev;
#endif
#ifdef USER
	cur->ptr = pf_if;
#endif
	anaPrint(ANA_NOTICE, "vlink_add_if(%s) --> added if %s to vl_iflist\n", vlink->vl_name, if_name);
	vlink->vl_ifcnt++;
	
	/* check if MTU must be updated or not */
	/* for KERNEL we have already set mtu */
#ifdef USER
	strncpy(if_info.ifr_name, if_name, IFNAMSIZ);
	if (ioctl(vl_rawsock, SIOCGIFMTU, &if_info) < 0)
		anaPrint(ANA_ERR, "vlink_add_if --> cannot get MTU info for %s\n", if_name);
	else
		mtu = if_info.ifr_mtu;
#endif
	if (mtu)
	{
		mtu -= 4; /* the vlink adds 4 bytes */
	
		if ( mtu < vlink->vl_mtu )
		{
			vlink->vl_mtu = mtu;
			anaPrint(ANA_DEBUG, "vlink_add_if --> setting %s MTU to %d\n",
				 vlink->vl_name, vlink->vl_mtu);
				 
			if (vlink->sendto) /* vlink is bound */
			{
				struct IDPinfo_s info;
				
				if (anaL0_getIDPinfo(vlink->sendto, vlink->caller, &info) < 0)
					anaPrint(ANA_ERR, "vlink_add_if --> cannot get IDPinfo for sendto idp %x\n", vlink->sendto);
				else		
				{
					info.MTU = vlink->vl_mtu;
					
					if (anaL0_setIDPinfo(&info) < 0)
						anaPrint(ANA_ERR, "vlink_add_if --> cannot set IDPinfo for sendto idp %x\n", vlink->sendto);
					else
						anaPrint(ANA_DEBUG, "vlink_add_if --> updated MTU to %d for IDPinfo %x\n",
							 vlink->vl_mtu, vlink->sendto);
				}
			}
		}
	}
	
	/* check if vl_mac must be set or not */
	if (newmac)
		vlink_check_mac(vlink, newmac);
	
	return VL_NOERR;
}

void vlink_check_mac(struct vlink_s *vlink, unsigned char *newmac)
{
	int i;
		
	if (!vlink || !newmac)
		return;
	
	/* ignore newmac if 00:00:00:00:00:00 */
	for (i=0; i<VL_MACLEN; i++)
	{
		if (newmac[i])
			break;
	}
	if (i == VL_MACLEN)
		return;
	
	for (i=2; i<VL_MACLEN; i++)
	{
		/* if one byte (except vl_mac[0] and vl_max[1]) */
		/* of vl_mac is set, exit */
		
		if (vlink->vl_mac[i])
			return;
	}
	
	/* we copy some part of the real MAC address */
	/* --> skip the first byte as it's almost always 00 */
	/* --> skip 2nd byte to fully use last 3 bytes */
	
	memcpy(vlink->vl_mac+2, newmac+2, VL_MACLEN-2);
		
	anaPrint(ANA_NOTICE, "vlink_check_mac(%s) --> vl_mac set to " ETH_FMT "\n", 
		 vlink->vl_name, ETH_HEX(vlink->vl_mac));
}

int vlink_del_if(char *vl_name, char *if_name)
{
	int if_index = -1;
	struct vlink_s *vlink = NULL;
	struct ifs_ptr *cur, *prev;
#ifdef KERNEL
	struct net_device *dev;
#endif
	if (!vl_name || !if_name)
	{
		anaPrint(ANA_ERR, "vlink_del_if --> vl_name or if_name NULL!\n");
		return VL_ERR_NULL;
	}

	vlink = vlink_lookup(vl_name);

	if (!vlink)
	{
		anaPrint(ANA_NOTICE, "vlink_del_if --> cannot find %s!\n", vl_name);
		return VL_ERR_NOT_FOUND;
	}

#ifdef USER
	if_index = if_nametoindex(if_name);
#endif

#ifdef KERNEL
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,24)
	dev = dev_get_by_name(if_name);
#else
	dev = dev_get_by_name(&init_net, if_name);
#endif
	if (dev)
	{
		if_index = dev->ifindex;
		dev_put(dev);
	}
#endif
	if (if_index <= 0)
	{
		anaPrint(ANA_ERR, "vlink_del_if(%s) --> cannot find %s!\n", vlink->vl_name, if_name);
		return VL_ERR_NOT_FOUND;
	}
	
	/* remove interface from vlink interface list */
	cur = vlink->vl_iflist;
	prev = NULL;
	
	if (!cur)
	{
		anaPrint(ANA_ERR, "vlink_del_if(%s) --> vl_iflist is NULL!\n", vlink->vl_name);
		return VL_ERR_NOT_FOUND;
	}
	
	while (cur)
	{
		if (cur->if_index == if_index)
		{
			if (!prev)
				vlink->vl_iflist = cur->next;
			else
				prev->next = cur->next;
			
			free(cur);
			vlink->vl_ifcnt--;
			anaPrint(ANA_DEBUG, "vlink_del_if(%s) --> removed if %s\n", 
			 	 vlink->vl_name, if_name);
			return VL_NOERR;
		}
		
		prev = cur;
		cur = cur->next;
	}

	anaPrint(ANA_ERR, "vlink_del_if(%s) --> if %s not in vl_iflist!\n", 
			 	 vlink->vl_name, if_name);
	
	return VL_ERR_NOT_FOUND;
}

int vlink_add_udp(char *vl_name, struct sockaddr_in *peer)
{
	struct vlink_s *vlink = NULL;
	struct udp_tun *cur, *prev = NULL;
	unsigned char newmac[VL_MACLEN];
	int tmp;

	if (!vl_name || !peer)
	{
		anaPrint(ANA_ERR, "vlink_add_udp --> vl_name or peer NULL!\n");
		return VL_ERR_NULL;
	}

	vlink = vlink_lookup(vl_name);

	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_add_udp --> cannot find %s!\n", vl_name);
		return VL_ERR_NOT_FOUND;
	}
	
	/* check if peer is already in vlink udp list */
	
	cur = vlink->vl_udplist;
	
	while (cur)
	{
		if (!memcmp(peer, &cur->udpdst, sizeof(struct sockaddr_in)))
		{
			anaPrint(ANA_ERR, "vlink_add_udp(%s) --> peer already in vl_udplist!\n", vlink->vl_name);
			return VL_NOERR;
		}
		
		prev = cur;
		cur = cur->next;
	}
	
	cur = malloc(sizeof(struct udp_tun));
	
	if (!cur)
	{
		anaPrint(ANA_ERR, "vlink_add_udp --> malloc failed!\n");
		return VL_ERR_NOMEM;
	}
	
	memcpy(&cur->udpdst, peer, sizeof(struct sockaddr_in));
	if (prev)
		prev->next = cur;
	else
		vlink->vl_udplist = cur;
	
	anaPrint(ANA_NOTICE, "vlink_add_udp(%s) --> added peer to vl_udplist\n", vlink->vl_name);
	vlink->vl_udpcnt++;
	
	/* build newmac and check if vl_mac must be set or not */
	tmp = random();
	memcpy(newmac, &tmp, 4);
	tmp = random();
	memcpy(newmac+4, &tmp, 2);
	vlink_check_mac(vlink, newmac);
	
	return VL_NOERR;
}

int vlink_del_udp(char *vl_name, struct sockaddr_in *peer)
{
	struct vlink_s *vlink = NULL;
	struct udp_tun *cur, *prev = NULL;
	
	if (!vl_name || !peer)
	{
		anaPrint(ANA_ERR, "vlink_del_udp --> vl_name or peer NULL!\n");
		return VL_ERR_NULL;
	}

	vlink = vlink_lookup(vl_name);

	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_del_udp --> cannot find %s!\n", vl_name);
		return VL_ERR_NOT_FOUND;
	}
	
	/* check if peer is in vlink udp list */
	
	cur = vlink->vl_udplist;
	prev = NULL;
	
	if (!cur)
	{
		anaPrint(ANA_ERR, "vlink_del_udp(%s) --> vl_udplist is NULL!\n", vlink->vl_name);
		return VL_ERR_NOT_FOUND;
	}
	
	while (cur)
	{
		if (!memcmp(&cur->udpdst, peer, sizeof(struct sockaddr_in)))
		{
			if (!prev)
				vlink->vl_udplist = cur->next;
			else
				prev->next = cur->next;
			
			free(cur);
			vlink->vl_udpcnt--;
			anaPrint(ANA_DEBUG, "vlink_del_udp(%s) --> removed peer\n", vlink->vl_name);
			return VL_NOERR;
		}
		
		prev = cur;
		cur = cur->next;
	}

	anaPrint(ANA_ERR, "vlink_del_udp(%s) --> peer not in vl_iflist!\n", vlink->vl_name);
	
	return VL_ERR_NOT_FOUND;
}

/* this is the callback function: prototype is fixed by API level 0 */
void vlink_send(void *data, int len, anaLabel_t input, void *aux)
{
	struct vlink_s *vlink = NULL;
	char *vl_name = (char *)aux;
	unsigned int ulen;
	char *ndata;
	
	anaPrint(ANA_DEBUG, "vlink_send --> asked to send %d bytes\n", len);
	
	if (!vl_name)
	{
		anaPrint(ANA_ERR, "vlink_send --> vl_name NULL!\n");
		return;
	}
	
	if (!data)
	{
		anaPrint(ANA_ERR, "vlink_send --> data ptr NULL!\n");
		return;
	}

	if (len <= 0)
	{
		anaPrint(ANA_ERR, "vlink_send --> data len <= 0!\n");
		return;
	}

	vlink = vlink_lookup(vl_name);
	
	if (!vlink)
	{
		anaPrint(ANA_ERR, "vlink_send(%s) --> cannot find vlink!\n", vl_name);
		return;
	}

	if ( !(vlink->vl_flags & IFF_UP) )
	{
		anaPrint(ANA_ERR, "vlink_send(%s) --> vlink DOWN!\n", vl_name);
		return;
	}

	if ( !(vlink->vl_iflist) && !(vlink->vl_udplist) )
	{
		anaPrint(ANA_ERR, "vlink_send(%s) --> vl_iflist and vl_udplist are empty!\n", vl_name);
		return;
	}

	anaPrint(ANA_DEBUG, "vlink_send --> sending packet to %s\n", vl_name);

	/* add vlink_id to data */
	ndata = malloc(len + sizeof(vlink->vl_id));
	if (!ndata)
	{
		anaPrint(ANA_ERR, "vlink_send(%s) --> malloc failed!\n", vlink->vl_name);
		return;
	}

	memcpy(ndata, &(vlink->vl_id), sizeof(vlink->vl_id));
	memcpy(ndata + sizeof(vlink->vl_id), data, len);
	ulen = len + sizeof(vlink->vl_id);
	anaValidate("vlink_send --> sending packet of len %i \n", ulen);
	anaPrint(ANA_DEBUG, "vlink_send --> sending packet of len %i \n", ulen);

#ifdef KERNEL
	kernel_vlink_send(vlink, ndata, ulen);
#endif

#ifdef USER
	user_vlink_send(vlink, ndata, ulen);
#endif
	
	free(ndata);
}
