/* This file defines the basic data types used in ANA */
#ifndef _ANACOMMON_H
#define _ANACOMMON_H

#ifdef USER
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h> 
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <stdint.h>
#include <errno.h>
#include <stdlib.h>

#endif

#ifdef KERNEL
#include <linux/module.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/stat.h> /*memset*/
#include <linux/types.h>
#include <linux/in.h>
#include <linux/time.h>
#include <linux/list.h>
#include <linux/random.h>
#include <linux/mm.h>
#include <linux/rwsem.h>
#include <linux/inet.h>
#include <net/sock.h>
#include <linux/socket.h>
#include <linux/net.h>
#include <linux/in.h>
#include <linux/wait.h>
#include <linux/limits.h>
#include <asm/param.h>
#include <asm/byteorder.h>
#endif

#include "analock.h"
#include "anaError.h"


#define anaHandle_t uint32_t
#define anaLabel_t uint32_t

#define MY_HANDLE myHandle
#define NODE_LABEL myNodeCompLabel

/** @defgroup anaCommon Standard ANA functions
 * ANA has a set of dedicated functions 
 */


/** \ingroup anaL0
 * \brief Definition of the callback function prototype
 */
typedef void (*anaCallback_t)(void *data, int len, anaLabel_t input, void *aux);




/** \ingroup anaL0
 * \brief function prototype for the copyAux funciton
 */
typedef void *(*copyAuxFunction)(void *aux);

/** \ingroup anaL0
 * \brief function prototype for the freeAux function
 */
typedef void (*freeAuxFunction)(void *aux);

unsigned int pow2(unsigned int x);
anaHandle_t generateHandle(void);
anaLabel_t generateLabel(void);
int sendSocketWrapper(void *sock, int domain, void *aux, void *msg, int len);

/* Wrapper functions */
void * __real_malloc(size_t size);
void * __wrap_malloc(size_t size);
unsigned int __wrap_sleep(unsigned int t);
unsigned int __real_sleep(unsigned int t);

#ifdef KERNEL
void *malloc(int size);
void free(void *ptr);
int random(void);
int atoi(char *nptr);
int inet_addr(const char *cp);
int sleep(int seconds);
int gettimeofday(struct timeval *tv, struct timezone *tz);
#endif

struct service_s {
	void *value;
	int valueLen;
};

struct context_s {
	void *value;
	int valueLen;
};


/* < notification stuff >*/

struct anaEventSub_s {
	uint32_t eventMask;
	
	/*according to he event type (IDP or brick regarded), only some of the following fields should be filled */
	
		/*in case regarding an IDP*/
	anaLabel_t label;
	anaHandle_t viewer;
		/*in case regarding a brick*/
	char *name;
};	


struct anaEventNotif_s {
	uint32_t eventID; /* should be only one event (no mask)*/
	
	/*according to he event type (IDP or brick regarded), only some of the following fields should be filled */
		/* IDP oriented event*/
	anaLabel_t label;
	anaHandle_t viewer;
	
		/* in case of a redirect event */
	anaLabel_t destLabel;
	anaHandle_t destViewer;
	
		/* name of the att/detached brick, or brick where Unpublish happened */
	char *name;	
};


/** \ingroup anaL0
 * \brief Definition of the notification event handler function prototype
 */
typedef void (*eventHandler_t)(struct anaEventNotif_s *notif );	


/* < / notification stuff >*/

/* < IDPinfo stuff >*/

enum idp_types {BRICK, REDIRECT, FORK}; /* types indicating the nature of an IDP */

struct nextIDP_s {
	anaLabel_t label;
	anaHandle_t viewer;
	struct nextIDP_s *next; /* in case it is a fork */
};

struct calleeBrick_s {
	anaHandle_t handle;
	anaLabel_t outputIDP;
};

struct IDPinfo_s {

	anaLabel_t label;
	anaHandle_t viewer;
	anaHandle_t owner;
	char permFlag; /* 1 for permanent */	

	char ICflag; /* 1 for IC*/
	struct context_s destContext;
	struct service_s destService;
	uint16_t MTU;	
	char chanType; /* unicast ('u'), anycast ('a'), multicast ('m'), broadcast ('b') */
	
	char IDPtype; /* nature of the IDP, value must be of enum idp_types */ 
	union destination_u {struct calleeBrick_s brick; struct nextIDP_s *IDPs;} binding;	 
};
/* < /IDPinfo stuff >*/

#ifdef KERNEL

#define MAXNAMELEN 30
struct anaBrickFunctions
{
	struct workqueue_struct *wq;
	struct work_struct *work;
	struct dataMessageQueue_s **recvDataQueue;
	analock_t *queueLock;
	int pluginFlag;
	void (*notification)(void);
};
/*
extern struct anaMinmexFunctions *myminmex;
extern struct dataMessageQueue_s *recvDataQueue;
extern analock_t queueLock;
*/
struct anaMinmexFunctions
{
        /*control*/
        struct attachRes_s *	(*attach)	(char *brick_name, void *commAux);
        void    		(*detach)	(anaHandle_t handle, int val);
        int     		(*regist)	(anaHandle_t querier, anaCallback_t fct, 
						void *aux, copyAuxFunction copyAux, 
						freeAuxFunction freeAux, anaHandle_t viewer, 
						int perm, int thread, anaLabel_t *result);
        int     		(*redirect)	(anaHandle_t querier, anaHandle_t originViewer, 
						anaLabel_t originLabel, anaHandle_t destViewer, 
						anaLabel_t destLabel);
	int 			(*forkData)	(anaHandle_t querier, anaHandle_t originViewer, 
						anaLabel_t originLabel, anaHandle_t destViewer, 
						anaLabel_t destLabel);
	int			(*changeLabelAux)(anaHandle_t querier, anaLabel_t label, void *newAux, copyAuxFunction copyAux, 
						freeAuxFunction freeAux);
        int     		(*unregist)	(anaHandle_t querier, anaHandle_t viewer, anaLabel_t label);
        void    		(*heartBeat)	(anaHandle_t handle);
        void    		(*notification)	(char *msg, int len);
	int			(*changeView)	(anaHandle_t querier, anaHandle_t oldViewer, 
						anaLabel_t label, anaHandle_t newViewer);
	int  			(*setIDPinfo)   (anaHandle_t querier, struct IDPinfo_s *info);
	int  			(*getIDPinfo)   (anaHandle_t querier, anaHandle_t viewer, anaLabel_t label, struct IDPinfo_s* toBeFilled);
   	int			(*registerEventHandler)	 (anaHandle_t callee, anaHandle_t owner, struct anaEventSub_s* sub, anaCallback_t notificationWrapper, void *aux, int auxLen, int threadFlag );	
	int 			(*unregisterEventHandler)	(anaHandle_t owner, struct anaEventSub_s* sub );
	int			(*notifyEvent)	(anaHandle_t notifier, struct anaEventNotif_s *notif );
	
	 /*data*/
        int   			(*callback)	(anaHandle_t querier, anaLabel_t label, anaHandle_t viewer, 
						void *data, int len);
        void    (*brickDataChannel)(struct anaBrickFunctions * brickDataFunctions, 
                                     anaHandle_t handle);
};
int unregisterAnaMinmex(char * name);
int registerAnaMinmex(struct anaMinmexFunctions * anaMinmexFct, int mode, char * name);
struct anaMinmexFunctions * getMinmexFunctions(char * name);



#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
struct wq_wrapper{
        struct sock * sk;
};

#else
struct wq_wrapper{
        struct work_struct worker;
	struct sock * sk;
};
#endif /* VERSION */
#endif /* KERNEL */

struct dataMessageQueue_s {
	anaCallback_t callBackFct; 		/* function handling the message or notification */
	anaLabel_t inputLabel;
	void *data;
	int dataLen;	
	void *aux;				/* auxilary argument of the callback function  */
	int thread;
	struct dataMessageQueue_s *next;
};

enum anaEvents {IDP_DEL, IDP_REDIR, IDP_UNPUB, IDP_INFOUP, BRICK_ATT, BRICK_DET};

/** @defgroup anaEvents ANA Events
  * standard ANA event identifiers to subscribe/notify to/about
  */

/** \ingroup anaEvents
 */
#define IDP_DELETED_EVENT 1 << IDP_DEL  /**< event regarding the deletion of an IDP */
/** \ingroup anaEvents
 */
#define IDP_REDIRECTED_EVENT 1 << IDP_REDIR  /**< event regarding the redirection of an IDP */
/** \ingroup anaEvents
 */
#define IDP_UNPUBLISHED_EVENT 1 << IDP_UNPUB /**< event regarding the unpublication of an IDP */
/** \ingroup anaEvents
 */
#define IDP_INFOUPDATED_EVENT 1 << IDP_INFOUP /**< event regarding the update of information about an IDP */
/** \ingroup anaEvents
 */
#define BRICK_ATTACHED_EVENT 1 << BRICK_ATT  /**< event regarding the attachement of a Brick */
/** \ingroup anaEvents
 */
#define BRICK_DETACHED_EVENT 1 << BRICK_DET  /**< event regarding the detachment of a Brick */


#ifdef USER
struct commSpecs_s {  /* this structure contains the specifics to communicate with a .so plugin */
	analock_t *queueLock;			/* lock on the message queue */
	struct dataMessageQueue_s **msgQueue;	/* queue where to deposit the message (at tail) */
	analock_cond_t *condLock; 		/* condition lock to wake up the awaiting thread */
	void *pluginHandle;
	int pluginFlag;   	/* if set to 1 means the brick is a .so plugin, oherwise, it is attached through the gates plugin*/
}; 

struct pluginit_s {
	char *commandLine;
	void *pluginHandle;
};
#endif /* USER */


/* generic netlink enums*/
enum {
        ATTR_UNSPEC,
        ATTR_ID,
        ATTR_ID2,
        __ATTR_MAX,
};
#define ATTR_MAX (__ATTR_MAX - 1)

enum {
        COMMAND_UNSPEC,
        COMMAND_CONTROL,
        COMMAND_DATA,
        __COMMAND_MAX,
};
#define COMMAND_MAX (__COMMAND_MAX - 1)


#endif /* _ANACOMMON_H */

