#ifndef ANALIB0_H_
#define ANALIB0_H_

#include "anaCommon.h"
#include "anaThread.h"
#include "anatimer.h"


#define IDP_PERM 1
#define NOTHREAD 0
#define THREAD 1 
#define VIEWER_PUBLIC 0

#define HEARTBEAT_INTERVAL  5

#define MY_HANDLE myHandle
#define NODE_LABEL myNodeCompLabel

extern anaHandle_t myHandle;
extern anaLabel_t myNodeCompLabel;

extern int attached;

#ifdef PLUGIN_MODE

#include "minmex_decl.h"

#ifdef USER
struct dataMessageQueue_s *messageQueue;
analock_cond_t waitLock; 
analock_t queueLock;

void *mypluginHandle;
int heartBeatTimer;
#endif


#ifdef KERNEL
extern struct anaMinmexFunctions *myminmex;
extern struct dataMessageQueue_s *recvDataQueue;
extern analock_t queueLock;
extern analock_rw_t attachLock;
#endif

#endif

struct condLock0_s{
	analock_cond_t cond;
	analock_t condLock;
	int flag;
};


#ifdef GATES_MODE


extern analock_t bigLock;

struct anaMinmexSpecs_s {

	anaHandle_t handle;			/* The handle chosen by the minmex to identify the brick	*/
	anaLabel_t minmexCompLabel;	/* label identifying the IDP to reach the minmex compartment via the data link */
	
	/*anaLabel_t notificationLabel; */
		
	int myDataMode;
	int myDataDescriptor;   	/* must be value specicied in anaTypes.h*/
        struct socket * myDataSocket;
        struct anaBrickFunctions * myDataFunctions;
       	
	void *myDataAux;			/* complementary argument depending on the mode*/
	 					 		/* e.g for UDP this would be a sockaddr_in */
	int myControlMode;					 
	int myControlDescriptor;
	struct socket * myControlSocket;
	void *myControlAux;
	
	int minmexDataMode;
	int minmexDataDescriptor;
        struct socket * minmexDataSocket;   	
	void *minmexDataAux;			
	 					 		
	int minmexControlMode;					 
	int minmexControlDescriptor;
	struct socket *minmexControlSocket;
	void *minmexControlAux;
	int age;
	
	int sizeAttachMsg;
	void *buffer;			/* buffer for data reception, will be allocated according to the MTU*/	
	struct shadowIDT_s *shadowT;	/* shadow dispatch table associated with minmex */
	unsigned char *attachMessage; 	/* This String is builtaccording to the information aboce by the initMinmexSpecs() function, it is sent to the minmex by
					  		 the attach() function */
}; 

#ifdef USER /* a gate brick can only be in user space */
struct anaMinmexSpecs_s *myMinmex;
anaLabel_t myHBLabel;

//#ifdef USER
struct threadAux {
	anaCallback_t fct;	/* the function to be executed */
	void * aux;		/* the aux argument */
	copyAuxFunction cpy;	/* the function to copy the aux (needed for structs with pointer etc) */
	freeAuxFunction free;	/* the function to free the aux */
};


struct callbackArgs {
	struct anaMinmexSpecs_s * minmex;
	anaCallback_t function;
	char * data;
	int len;
	anaLabel_t input;
	void * aux;
};

#endif


struct anaMinmexSpecs_s * anaL0_initMinmexSpecs(char *myDataURI, char *myControlURI, 
                                   char *minmexControlURI, int desiredDataMode, int timeoutVal);

int anaL0_attach(struct anaMinmexSpecs_s * minmexSpec, int tableBitSize);
void anaL0_heartBeat(int sig);
void freeMinmexSpecs(struct anaMinmexSpecs_s * spec);
#ifdef USER
int listenLoop(); 
#endif

#endif

int anaL0_send( const anaLabel_t label , void *data , int lenData);  

#ifdef GATES_MODE
int anaL0_send_ack(const anaLabel_t label , void *data , int lenData);  
#endif

int anaL0_redirect( const anaLabel_t originLabel, const anaHandle_t originViewer, const  anaLabel_t destLabel,  const anaHandle_t destViewer);
int anaL0_forkData( const  anaLabel_t originLabel, const  anaHandle_t originViewer,  const anaLabel_t destLabel, const  anaHandle_t destViewer);
anaLabel_t anaL0_registerCallback( anaCallback_t fct, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux, 
						const anaHandle_t viewer, int perm, int thread);

int anaL0_changeLabelAux( const anaLabel_t label, void *aux, copyAuxFunction copyAux, freeAuxFunction freeAux);
int anaL0_changeLabelView( const anaLabel_t label,  const anaHandle_t oldViewer,  const anaHandle_t newViewer);
int anaL0_unregisterCallback( const anaLabel_t label, const  anaHandle_t viewer);



int anaL0_getIDPinfo(anaLabel_t label,  anaHandle_t viewer, struct IDPinfo_s *result);
int anaL0_clearIDPinfo(struct IDPinfo_s *result);
int anaL0_setIDPinfo( struct IDPinfo_s *info);

int anaL0_eventSubscribe(struct anaEventSub_s* sub, eventHandler_t handler, int threadFlag);
int anaL0_eventUnsubscribe(struct anaEventSub_s* sub);
int anaL0_eventNotify( struct anaEventNotif_s *notif);


void anaL0_stopANA(int ret);
void dataLoop(struct condLock0_s *lock);
 

/** \ingroup anaCommon
 * \brief This function needs to be provided from every brick. 
 * It will be called after the brick is correctly initialized and attached to the minmex
 */
int brick_start(void);

/** \ingroup anaCommon
 * \brief This function needs to be provided from every brick. 
 * It will be called when a brick will be detached from the minmex and should therefore do all sort of clean up
 */
void brick_exit(void);


void heartBeat(void);

#ifdef KERNEL
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
void recvData(struct work_struct *work);
#else
void recvData(void *work);
#endif
#endif


#endif
