#ifndef _ANA_AL2_H
#define _ANA_AL2_H

#include "anaLib1.h"

#define DEFAULT_MQ_BITSIZE 10


extern struct MQTable_s MQTable;

struct condLock_s{
	analock_cond_t cond;
	analock_t condLock;
	analock_t valueLock;
	void *value;
	int valueLen;
};
/**
 * \ingroup anaL2
 * \brief message received in AL2 callback functions or returned by anaL2_receive function
 */
struct anaL2_message{
	int dataLen;	/**< size in bytes of the received data */
	void *data;	/**< pointer to the received data */
	struct context_s *senderContext; /**< pointer to the context of the sender */	
	struct service_s *senderService; /**< pointer to the service of the sender */
	anaLabel_t idp; /**< the idp on which the message is received */
	anaLabel_t responseLabel;	/**< the idp provided by the compartment through which the message was received, to reply to the sender */
	void *bufferPointer;	/**< pointer to the XRP_CMD_DATA message into which the previous fields point */
};

typedef void (*AL2Callback_t) (struct anaL2_message *msg);

struct anaL2_lkpResponse{
	
	void *description;
	int descLen;
	
	anaLabel_t label;
	struct anaL2_lkpResponse *next;
};

struct anaL2_serviceWrap{
	AL2Callback_t funct;
	struct anaL2_message *msg;
};

void anaL2_freeMsg(struct anaL2_message *msg);

void anaL2_freeLkpResponse(struct anaL2_lkpResponse *resp);

int anaL2_forwardData(anaLabel_t destIDP, void *data , int lenData, struct context_s *senderContext, struct service_s *senderService, anaLabel_t responseLabel);

anaLabel_t anaL2_publish(anaLabel_t compIDP, struct context_s *context, struct service_s *service, AL2Callback_t function, int separateThread, struct timespec *timeout);

int anaL2_unpublish(anaLabel_t compIDP, anaLabel_t label, struct context_s *ctxt, struct service_s *service, struct timespec *timeout);

anaLabel_t anaL2_resolve(anaLabel_t compIDP, struct context_s *context, struct service_s *service, char chanType, struct service_s *querierDescription, struct timespec *timeout);

int anaL2_lookup(anaLabel_t compIDP, struct context_s *context, struct service_s *service, struct anaL2_lkpResponse **result, struct service_s *querierDescription, struct timespec *timeout);

struct anaL2_message * anaL2_receive(anaLabel_t input, struct timespec *timeout);

char * anaL2_requestReply(anaLabel_t dest, xrpMsg_t msg, int reqlen, int * respLen, struct timespec *timeout);

int anaL2_initDefault(void);

#endif /* _ANA_AL2_H */
