#ifndef _ANATIMER_H
#define _ANATIMER_H

#include "anaCommon.h"

#ifdef USER
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/wait.h>
#include <string.h>
#include <pthread.h>
#endif /* USER */

#ifdef KERNEL
#include <linux/delay.h>
#include <linux/time.h>
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/jiffies.h> /* does not exist in 2.4 */
#else
#define msecs_to_jiffies(msecs) (HZ * (msecs)/1000)
#endif
#endif /* KERNEL */

/* Common stuff */

#define ANATIMER_ABSOLUTE 1
#define ANATIMER_PERIODIC 1

typedef void (*fct_t)(void *data);

struct anatimer_s {
	unsigned int id;
	struct anatimer_s *next;
	fct_t action;
	void *data;
	int running;
	struct timeval period;
	struct timeval timeout;
	unsigned char fix_clock;
};

void anatimer_init(void);
void anatimer_exit(void);
int anatimer_add(long msec, unsigned char period, unsigned char fix_clock, 
		 fct_t action, void *fct_data, int datalen);
int anatimer_del(unsigned int id);
int anatimer_restart(unsigned int id, long msec);
int anatimer_change(unsigned int id, long msec);

#endif /* _ANATIMER_H */
