#ifndef _QUICKREP_H
#define _QUICKREP_H

#include "anaCommon.h"
#include "hashFunction.h"

struct QREPKeyword{
	char *keyWord;
	struct QREPKeyword *next;
};

struct QREPEntry{
	void *name;
	int nameLen;
	void* value;
	struct QREPKeyword *keyWordsList;
	struct QREPEntry *next; /* for hash conflict cases*/	
};

struct QREPKeyObjectList{
	struct QREPEntry *entry;
	struct QREPKeyObjectList *next;
};

struct QREPKey{
	char *keyWord;
	struct QREPKeyObjectList *entriesList;
	struct QREPKey *next;   /* for hash conflict cases*/ 
};

/*the repository*/
struct QREP{
	int sizeEntryTab;
	int sizeKeyTab;
	int hashMaskEntry;
	int hashMaskKey;
	struct QREPEntry **entryTab;
	struct QREPKey **keyTab; 
};

typedef void (*freeValFunction_t) (void *value);


typedef void (*valueFunction_t) (void *value, void *aux);



ub4 QREPhash( ub1 *k, ub4 length, ub4 initval, ub4 maskLength);

enum{QREP_OPERATOR, QREP_OPERAND};
#define QREP_BUFSIZE 500

struct qrep_token{
	int type; /*indicates whether token is operand or operator*/
	char *lexicalVal;
	void *semanticVal;
	struct qrep_token * prev; /* double chained list*/
	struct qrep_token * next;	
};
int initQRepository(struct QREP *rep, int hashMaskEntryTab, int hashMaskKeyTab /*, int deadEntryLimit */);
int qrep_name2hash(struct QREP *rep, char *name, int nameLen);
int qrep_key2hash(struct QREP *rep, char *key);
void * qrep_cloneValue(void *value, int valueSize);
struct QREPEntry *allocateQREPEntry(char *name, int nameLen ,void *value, int valueSize);
int QREPaddKeywordToEntry(struct QREPEntry *entry, char *keyWord);
int qrep_updateKey(struct QREP *rep, char *keyword, struct QREPEntry *entry);
int qrep_freeKey(struct QREP *rep, struct QREPKey *key, int h);
int qrep_removeFromKeysList(struct QREP *rep, struct QREPEntry *toRemove, char *keyword);
struct QREPEntry *getQREPEntryByName(struct QREP *rep, char *name, int nameLen);
int addQREPEntry(struct QREP *rep, struct QREPEntry *entry);
struct QREPKeyObjectList * QREPfindByKey(struct QREP *rep, char *key);
int QREPfindByQuery(struct QREP *rep, char *query, struct QREPKeyObjectList **res);
void QREPfreeQueryRes(struct QREPKeyObjectList *list);
/* this function does not free the repository i.e is does not deallocate it's table but rather frees all the
entries */
void freeQRepository(struct QREP *rep, freeValFunction_t fct);
int qrep_evalQuery( struct QREP *rep, char *buf, struct QREPKeyObjectList **res );
int  qrep_evalTokenChain(struct qrep_token *tkChain);
struct QREPKeyObjectList *  qrep_intersectObjLists(struct QREPKeyObjectList *list1 , struct QREPKeyObjectList *list2);
struct QREPKeyObjectList *  qrep_minusObjLists(struct QREPKeyObjectList *list1 , struct QREPKeyObjectList *list2);
struct QREPKeyObjectList *  qrep_uniteObjLists(struct QREPKeyObjectList *list1 , struct QREPKeyObjectList *list2); 
struct QREPKeyObjectList *  qrep_getCopyObjList( struct QREP *rep , char *keyword);
int freeQREPEntry(struct QREP *rep, struct QREPEntry *entry, freeValFunction_t freeValueFunction);
#endif
