#ifndef IP2ANA_H
#define IP2ANA_H

/* ipt_ip2ana.h
 * 
 * IP to ANA migration/adaptation layer
 * Shared object extension for the iptables command
 *
 * Author: Christophe Jelger <Christophe.Jelger@unibas.ch>
 * 	   	
 * Copyright (c) 2008, University of Basel
 *  
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include <linux/in6.h>

struct udp_ports {
	
	/* UDP port numbers to communicate with ip2ana brick */

	u_int16_t ipt_port; /* in network order */
	u_int16_t brick_port; /* in network order */
	u_int16_t i2a_port; /* in network order */
	u_int16_t a2i_port; /* in network order */
};


/* for timing out pending DNS requests */

#define DNSREQS_EXPIRE_SEC 5 /* timeout in seconds */
#define TIMER_CLICK_SEC 1 /* period for checking pending queries */

#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,6,0)
#include <linux/jiffies.h> /* does not exist in 2.4 */
#else
#define msecs_to_jiffies(msecs) (HZ * (msecs)/1000)
#endif

/* return instructions when checking data in dns_request_for_ana() */
#define IP2ANA_DNSREQ 1
#define IP2ANA_DATA 0
#define IP2ANA_LETGO -1

/* MAX number of pending DNS requests */
#define MAX_DNS_REQS 128

extern int dns_request_for_ana(struct sk_buff *skb);
extern void send_dns_reply(struct sk_buff *skb, char *result);
extern int store_dnsreq(struct sk_buff *, uint16_t trans_id);
extern void timer_click(unsigned long dummy);
extern void clean_arp_entries(void);
extern void clean_dnsreqs(void);

extern int netbox_neigh_map(void *addr, int family, char *eth);
extern int l3deliver(void *data, int len);

/* to check whether PTR queries are for ANA or not */
#define ANA_IP_SUBNET "10.42.0.0"
#define ANA_IP_MASK "255.255.0.0"
extern struct in_addr ana_ip4_subnet;
extern struct in_addr ana_ip4_mask;

#define ANA_IPV6_SUBNET { { { 0x00,0x10,0x00,0x42,0,0,0,0,0,0,0,0,0,0,0,0 } } }
#define ANA_IPV6_PFXLEN 32 /* bits */
extern struct in6_addr ana_ip6_subnet;
extern int ana_ip6_pfxlen;

#endif /* IP2ANA_H */
