/* C/minmex/common/parser.c
 * 
 * parser functions for the KVR
 *
 * Author: Ghazi Bouabene <ghazi.bouabene@unibas.ch>
 *	
 * Copyright (c) 2008, Ghazi Bouabene, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


/* temporary */

#include "../include/nodeComp.h"
#define BUFSIZE 500

void freeToken(struct token *tk){
	if(tk!=NULL){
		free(tk->lexicalVal);
		free(tk);
	}
}

int reduceParentheses(struct token *tkChain){ 
	struct token *openPar=NULL;
	struct token *closePar=NULL;
	struct token *tmpTok=tkChain;
	struct token *tmpTok2 = NULL;
	struct token *next = NULL;
	struct token *next2 = NULL;
	
	while(tmpTok!=NULL){
		next=tmpTok->next;	
		if(tmpTok->type==OPERATOR && tmpTok->lexicalVal[0]=='('){
			int found =0;
			int intermediateOpar = 0;			
			openPar=tmpTok;
			/* we look for the closing parentheses */
			
			tmpTok2=tmpTok->next;
			while(!found && tmpTok2!= NULL){
				
				next2=tmpTok2->next;
				if(tmpTok2->type==OPERATOR && tmpTok2->lexicalVal[0]=='(')
					intermediateOpar++;	
				
				if(tmpTok2->type==OPERATOR && tmpTok2->lexicalVal[0]==')') {
					if(intermediateOpar>0)
						intermediateOpar--;
					else {	/* we found the closing parenthesis*/
						closePar = tmpTok2;
						found = 1;
						/*we need to cut the expression included in parenthesis form the chain*/
						
						if(closePar->prev == openPar)
							return -1; /* we do not allow () */
						else
							closePar->prev->next = NULL;
						
											
						if(evalTokenChain(openPar->next)<0)
							return -1;
						/* the expression between parentheses has been reduced
						inside the first token after openPar, we need to move it in OPAR token */
						openPar->type = OPERAND;
						openPar->semanticVal = openPar->next->semanticVal;
						
						freeToken(openPar->next);
						
						openPar->next=closePar->next;
						next = closePar->next;
						if(closePar->next!=NULL)
							closePar->next->prev=openPar;
						
						freeToken(closePar);
					}
				}
				
				tmpTok2=next2;
			}
			if(!found){
				anaPrint(ANA_ERR, "Missing closing parenthesis\n");
				return -1;
			}
				
			
		}
		
			
		tmpTok=next;
	}
	
	return 0;
}


/* this function is always called on tokenChains that do notcontain parentheses*/
int reduceAND(struct token *tkChain){ 
	struct token *next=NULL;
	while(tkChain!=NULL){
		next=tkChain->next;
		if(tkChain->type==OPERATOR && tkChain->lexicalVal[0]=='*'){
					
			/* first we check if previous and next tokens are Operands
			   if not we return an error	
			*/	
			if(tkChain->prev==NULL || tkChain->prev->type!=OPERAND ||tkChain->next==NULL || tkChain->next->type!=OPERAND ){
				anaPrint(ANA_ERR, "invalid operands for AND operator \n");
				return -1;
			}
			
			tkChain->prev->semanticVal = intersectObjLists((struct KVRKeyObjectList *)tkChain->prev->semanticVal, (struct KVRKeyObjectList *)tkChain->next->semanticVal);
			tkChain->prev->next = tkChain->next->next;
			next = tkChain->next->next;
			if(next!=NULL)
				next->prev = tkChain->prev;
			
			freeToken(tkChain->next);
			freeToken(tkChain);
		}	
		tkChain = next;
	}
	return 0;
}

/* this function is always called on tokenChains that do notcontain parentheses nor AND tokens*/
int reduceORMINUS(struct token *tkChain){ 
	struct token *next=NULL;
	while(tkChain!=NULL){
		next=tkChain->next;
		
		if(tkChain->type==OPERATOR && (tkChain->lexicalVal[0]=='+' || tkChain->lexicalVal[0]=='-')){
			struct KVRKeyObjectList * res = NULL;	
			/* first we check if previous and next tokens are Operands
			  if not we return an error	
			*/	
			if(tkChain->prev==NULL || tkChain->prev->type!=OPERAND ||tkChain->next==NULL || tkChain->next->type!=OPERAND ){
				anaPrint(ANA_ERR, "invalid operands %c operator\n", tkChain->lexicalVal[0] );
				return -1;
			}
			
			if( tkChain->lexicalVal[0] == '-')
				res = minusObjLists((struct KVRKeyObjectList *)tkChain->prev->semanticVal, (struct KVRKeyObjectList *)tkChain->next->semanticVal);
			if( tkChain->lexicalVal[0] == '+')
				res = uniteObjLists((struct KVRKeyObjectList *)tkChain->prev->semanticVal, (struct KVRKeyObjectList *)tkChain->next->semanticVal);
			
			tkChain->prev->semanticVal = res;
			tkChain->prev->next = tkChain->next->next;
			next=tkChain->next->next;
			if(next!=NULL)
				next->prev = tkChain->prev;	
			
			freeToken(tkChain->next);
			freeToken(tkChain);
		}
		tkChain = next;
	}
	return 0;
}

int evalTokenChain(struct token *tkChain){ 

	if(tkChain==NULL)
		return -1;
	/* first we reduce the expressions between parentheses */
	if (reduceParentheses(tkChain)<0){
		anaPrint(ANA_ERR, "parentheses problem\n");
		return -1;
	}	
	
	/* then we reduce the AND(*) rule*/
	if (reduceAND(tkChain)<0){
		anaPrint(ANA_ERR, "AND problem\n");
		return -1;
	}

	/* and finally we reduce the OR, MINUS rules */
	if (reduceORMINUS(tkChain)<0){
		anaPrint(ANA_ERR, "OR,MINUS problem\n");
		return -1;
	}
	
	return 0;
}


struct token * tokenize(char *expr){
	int offset=0;
	int bufferOffset=0;
	struct token *tokList=NULL;
	struct token *predecessor=NULL;
	int exprLen=0;
	char buf[BUFSIZE];
	char c=1;

	if(expr==NULL)
		return NULL;
	
	exprLen=strlen(expr)+1;	
	memset(buf,0,BUFSIZE);
	while(offset<exprLen){
		c = expr[offset];
		
		if(c == 0){
			struct token *tmpTok = NULL;
			if(bufferOffset>0){ /*we create an operand token*/
				tmpTok = (struct token *) malloc(sizeof(struct token));
				tmpTok->type = OPERAND;
				tmpTok->lexicalVal = (char *) malloc(bufferOffset+1);
				tmpTok->lexicalVal[bufferOffset] = 0;
				strncpy(tmpTok->lexicalVal, buf, bufferOffset);
				
				/* and the semantic one */
				tmpTok->semanticVal = (struct KVRKeyObjectList *) getCopyObjList(tmpTok->lexicalVal);
				
				if(predecessor==NULL)
					tokList = tmpTok;
				else {
					predecessor->next = tmpTok;
					tmpTok->prev = predecessor;
				}
				predecessor = tmpTok;
				/*we clear the buffer and reset the offset to 0*/
				memset(buf,0,BUFSIZE);
				bufferOffset=0;
			}
			return tokList;
		}
				
		if(c =='+' || c =='-' || c =='*' || c =='(' || c ==')' ) {
			
				struct token *tmpTok = NULL;
				struct token *tmpTok2 = NULL;
				if(bufferOffset>0){ /*we create an operand token*/
					tmpTok = (struct token *) malloc(sizeof(struct token));
					tmpTok->type = OPERAND;
					/* we set the lexical value*/
					tmpTok->lexicalVal = (char *) malloc(bufferOffset+1);
					tmpTok->lexicalVal[bufferOffset] = 0;
					strncpy(tmpTok->lexicalVal, buf, bufferOffset);
					
					/* and the semantic one */
					tmpTok->semanticVal = (struct KVRKeyObjectList *) getCopyObjList(tmpTok->lexicalVal);
											
					if(predecessor==NULL)
						tokList = tmpTok;
					else {
						predecessor->next = tmpTok;
						tmpTok->prev = predecessor;
					}
					predecessor = tmpTok;
					/*we clear the buffer and reset the offset to 0*/
					memset(buf,0,BUFSIZE);
					bufferOffset=0;
				}
				
				tmpTok2 = (struct token *) malloc(sizeof(struct token));
				tmpTok2->type = OPERATOR;
				tmpTok2->lexicalVal = (char *) malloc(1);
				tmpTok2->lexicalVal[0] = c;
				tmpTok2->semanticVal=NULL;
				
				if(predecessor==NULL)
					tokList = tmpTok2;
				else {
					predecessor->next = tmpTok2;
					tmpTok2->prev = predecessor;
				}
				
				predecessor = tmpTok2;	
		}			
		else
			if(bufferOffset<BUFSIZE){
				buf[bufferOffset]=c;
				bufferOffset++;
			}
			
			
		offset++;
	}
	/* if we end up here the expression is too long, we retunr NULL*/
	return NULL;
}	

void displayTokenChain(struct token *tokenChain){
	struct KVRKeyObjectList *tmp;
	while(tokenChain!=NULL){
		if(tokenChain->type==OPERAND)
			anaPrint(ANA_DEBUG, "%s [ ", tokenChain->lexicalVal);
		else
			anaPrint(ANA_DEBUG, "%c [ ", tokenChain->lexicalVal[0]);	
		
		tmp = (struct KVRKeyObjectList *)tokenChain->semanticVal;
		while(tmp!=NULL){
			anaPrint(ANA_DEBUG, "%s, ", tmp->entry->name);
			tmp=tmp->next;
		}
		
		tokenChain = tokenChain->next;
		anaPrint(ANA_DEBUG, "] --> ");
	
	}
	anaPrint(ANA_DEBUG, "\n");
}


struct KVRKeyObjectList * getCopyObjList(char *keyword){
	
	struct KVRKeyObjectList *original= findByKey(&kValRep, keyword); 
	struct KVRKeyObjectList *copy=NULL;
	struct KVRKeyObjectList *predec=NULL;
	struct KVRKeyObjectList *tmp=NULL;
	/* mow we make a copy */
	predec = NULL;
	
	while(original!=NULL){
		tmp = (struct KVRKeyObjectList *) malloc(sizeof(struct KVRKeyObjectList));
		tmp->entry = original->entry;
		if(predec==NULL)
			copy = tmp;
		else{
			predec->next=tmp;
		}
		predec = tmp;		
		original = original->next;
	}
	
	
	return copy;
}


/* this function gives back a list containing elements common to list and list2
 * ATTention : this function frees list1 and list2 at the end (since it is a reduction(YACC)), 
 * therefore list1 and list2 must be copies of the findByKey(struct KVR *rep, char *keyword) query
 */
struct KVRKeyObjectList * intersectObjLists(struct KVRKeyObjectList *list1 , struct KVRKeyObjectList *list2) {
	
	struct KVRKeyObjectList *tmp1;
	struct KVRKeyObjectList *tmp2;
	struct KVRKeyObjectList *tmp2_predec;
	struct KVRKeyObjectList *tmp1_next;
	struct KVRKeyObjectList *tmp2_next;
	struct KVRKeyObjectList *res = NULL;
	int added;
	
	if(list1==NULL || list2 == NULL)
		return NULL;
	tmp1 = list1;
	while(tmp1!=NULL){
		tmp2=list2;
		tmp2_predec=NULL;
		added=0;
		tmp1_next = tmp1->next;
		while(tmp2!=NULL && !added){
			if(tmp1->entry==tmp2->entry){
				tmp2_next=tmp2->next;
				/* we add tmp1 to res*/
				added = 1;
				tmp1->next = res;
				res = tmp1;
				/*we free tmp2 */
				if(tmp2_predec == NULL)
					list2 = tmp2->next;
				else
					tmp2_predec->next = tmp2->next;
				free(tmp2);
				
			}
			else {
				tmp2_predec = tmp2;
				tmp2=tmp2->next;
			}		
		}
		/*if not added we free tmp1*/
		if(!added)
			free(tmp1);
		
		tmp1=tmp1_next;	
	}
	
	/*now we free the elements of list2 that didn't match with tmp1 */
	while(list2!=NULL){
		tmp1=list2->next;
		free(list2);
		list2=tmp1;
	}
	return res;	
}

/* list1 minus list2 i.e  the order of parameters is important here*/
struct KVRKeyObjectList * minusObjLists(struct KVRKeyObjectList *list1 , struct KVRKeyObjectList *list2) {
	struct KVRKeyObjectList *tmp1;
	struct KVRKeyObjectList *tmp2;
	struct KVRKeyObjectList *tmp1_predec;
	struct KVRKeyObjectList *tmp1_next;
	struct KVRKeyObjectList *tmp2_predec;
	int removed;
	
	if(list2 == NULL)
		return list1;
	
	tmp1=list1;
	tmp1_predec=NULL;
	while(tmp1!=NULL){
		tmp2=list2;
		tmp2_predec=NULL;
		tmp1_next=tmp1->next;
		removed = 0;
		while(tmp2!=NULL && !removed){
			if(tmp1->entry==tmp2->entry){
				/* we remove tmp1 */
				if(tmp1_predec==NULL)
					list1=tmp1->next;
				else
					tmp1_predec->next=tmp1->next;
				
				/*we also remove tmp2*/
				if(tmp2_predec==NULL)
					list2=tmp2->next;
				else
					tmp2_predec->next = tmp2->next;
				free(tmp2);	
				
				removed=1;							
			}
			else{
				tmp2_predec=tmp2;	
				tmp2=tmp2->next;
			}
		}
		if(!removed)
			tmp1_predec=tmp1;
		tmp1=tmp1_next;
	
	}
	return list1;
}


/* this function unites the 2 lists without duplicates, 
 I didn't use minus in this function since it frees list2 and it shoud stay that way*/

struct KVRKeyObjectList * uniteObjLists(struct KVRKeyObjectList *list1 , struct KVRKeyObjectList *list2) {
	struct KVRKeyObjectList *tmp1=NULL;
	struct KVRKeyObjectList *tmp2=NULL;
	struct KVRKeyObjectList *tmp2_predec=NULL;
	struct KVRKeyObjectList *tmp2_next=NULL;
	
	if(list1==NULL)
		return list2;
	if(list2==NULL)	
		return list1;	
	tmp1 = list1;
	while(tmp1!=NULL){
		tmp2_predec=NULL;
		tmp2=list2;
		while(tmp2!=NULL ){
			if(tmp1->entry==tmp2->entry){
				tmp2_next = tmp2->next;
				if(tmp2_predec==NULL)
					list2=tmp2->next;
				else	
					tmp2_predec->next = tmp2->next;
				
				free(tmp2);
				tmp2=tmp2_next;
			}
			else{
				tmp2_predec=tmp2;
				tmp2=tmp2->next;
			}
		}
		tmp1=tmp1->next;
	}
	
	/*now we have list1 and list2 wihtout replicas, we can concatenate*/
		
	tmp1=list1;
	while(tmp1->next!=NULL)
		tmp1=tmp1->next;
	
	tmp1->next=list2;
	
	return list1;
}



int evalQuery(char *buf , struct KVRKeyObjectList **res ){
	struct token *tokenChain = NULL;
	if(buf==NULL || res==NULL){
		return -1;
	}
		
	tokenChain = tokenize(buf);
		
	if(evalTokenChain(tokenChain)){
		*res = NULL;
		return -1;
	}
	
	*res = (struct KVRKeyObjectList *)tokenChain->semanticVal;
	freeToken(tokenChain);
	return 0;
}













