/* C/minmex/common/statusFunctions.c
 * 
 * functions which allow a user to browse the KVR, IDT, and BT
 *
 * Author: Christophe Jelger <christophe.jelger@unibas.ch>
 *	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */

#include "../include/anaNode.h"
#include "../include/statusInterface.h"

int sendWithCookie(char *sendbuf, int len, void *sock, struct sockaddr_in * replyto){
	
	int mcLen;
	
	if(sendbuf==NULL || len<0 || sock ==NULL)
		return -1;
	
	mcLen = strlen(MAGIC_COOKIE);
	strncpy(sendbuf+len, MAGIC_COOKIE, mcLen);

	return sendSocketWrapper(sock, AF_INET, replyto, sendbuf, len+mcLen);
}


int status_cli_command(char *recvbuf, int len, struct sockaddr_in *replyto)
{
	char sendbuf[BUFSIZE];
	char tmp[BUFSIZE];
	char *str;
	char *token, *ptr;
	char *ownerName=NULL;
	char *viewerName=NULL;
	char *calleeName=NULL;
	int i;
	
	if (len == 0)
	{
		anaPrint(ANA_ERR, "status_cli_command --> len NULL\n");
		return -1;
	}

	if (!recvbuf || !replyto)
	{
		anaPrint(ANA_ERR, "status_cli_command --> recvbuf or replyto NULL!\n");
		return -1;
	}

	/* get command name */
	ptr = (char *)recvbuf;
	token = strsep(&ptr, " ");
	memset(sendbuf, 0, sizeof(sendbuf));
       	if (!token || len == 2)
	{
		strcpy(sendbuf, "--> cannot parse command name!\n");
		sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		return -1;
	}
	
	anaPrint(ANA_DEBUG, "status_cli_command --> received: %s (%d bytes)\n", recvbuf, len);
		
	if ( !strncmp(token, "show_IDT", 7) )
	{	
                int j = 0;
                struct IDTentry_s *tmpIDTE = NULL;
                str = sendbuf;

                sprintf(str, "IDT hash table size: %i\n",IDT.size); str = sendbuf + strlen(sendbuf);
                sprintf(str, "Non permanent IDT entries expire after not beeing used for %i heart beats\n",IDT.deadEntryLimit);
		str = sendbuf + strlen(sendbuf);
                sprintf(str, "Heart beat interval : %d seconds\n",HEART_BEAT_INTERVAL); str = sendbuf + strlen(sendbuf);
		sprintf(str, "IDT Entries Listing:\n\n");str = sendbuf + strlen(sendbuf);
                sprintf(str, "----------------------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);

		sprintf(str, "%8s %15s %15s %15s %8s %8s\n" , "IDP", "Owner", "View", "Callee", "Age", "Index" );
		str = sendbuf + strlen(sendbuf);

		sprintf(str, "----------------------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
		for (i = 0; i < (int) IDT.size; i++){

			tmpIDTE =IDT.tab[i];

			while (tmpIDTE!= NULL){ 
				memset(tmp, 0, BUFSIZE);
                                j++;
				ownerName = handle2name(&BT, tmpIDTE->owner);
				viewerName = handle2name(&BT, tmpIDTE->viewer);
				if (viewerName==NULL)
					viewerName = "public";

				/*to know the callee we need to test the callBack functions */	
				if (tmpIDTE->callBackFct == (minmexCallback_t)&sendData  && tmpIDTE->aux!=NULL){
					struct IDTaux_s *tmp = (struct IDTaux_s *)tmpIDTE->aux ;
					calleeName=handle2name(&BT, tmp->callee);
				
				}	
				if (tmpIDTE->callBackFct == (minmexCallback_t)&redirectData && tmpIDTE->aux!=NULL){
					calleeName = (char *) malloc(15);
					sprintf(calleeName, "%8x", ((struct redirectAux_s *)(tmpIDTE->aux))->label );
				}
				

				if (tmpIDTE->callBackFct == (minmexCallback_t)&forkData && tmpIDTE->aux!=NULL){
					int *nbForks = (int *)tmpIDTE->aux;
					int offsetCallee=0, i;
					
					struct redirectAux_s *redirA = NULL;	
									
					
					calleeName = (char *) malloc( ((*nbForks) * 10)+1 );
					
					redirA = (struct redirectAux_s *)((tmpIDTE->aux) + sizeof(int));
					for(i=0;i<*nbForks;i++){
						anaPrint(ANA_DEBUG, "*********** pointer to calleeName(%p) redirA(%p).\n", calleeName, redirA);
						sprintf(calleeName+offsetCallee, "%8x, ",  redirA->label);
						
						redirA = (struct redirectAux_s *)((tmpIDTE->aux) + sizeof(int) + sizeof(struct redirectAux_s));
						offsetCallee += 10;
					}
				
									
				}	
				
				if (tmpIDTE->callBackFct == (minmexCallback_t)& dispatchNodeCompMsg)			
					calleeName="Node Comp"; 
				if (tmpIDTE->age!=-1)
					sprintf(tmp, "  %8x %13s %15s %15s %8d %8d\n",tmpIDTE->label, ownerName, viewerName, calleeName, tmpIDTE->age, i); 
                                else
					sprintf(tmp, "  %8x %13s %15s %15s %8s %8d\n",tmpIDTE->label, ownerName, viewerName, calleeName, "Perm", i);
				
				if((strlen(sendbuf)+strlen(tmp)+1)<(BUFSIZE - strlen(MAGIC_COOKIE))){
					strcat(sendbuf, tmp);	
				}
				else { /*we send the sendbuf with "magic cookie" */
					sendWithCookie(sendbuf, strlen(sendbuf)+1, &status_clisock, replyto);
					/*we copy tmp's content into sendbuf*/
					strcpy(sendbuf, tmp);	
				}
								
				if (tmpIDTE->callBackFct == (minmexCallback_t)&redirectData || tmpIDTE->callBackFct == (minmexCallback_t)&forkData)
					free(calleeName);

				tmpIDTE=tmpIDTE->next;
			} 
			
                }

		memset(tmp, 0, BUFSIZE);
		sprintf(tmp, "----------------------------------------------------------------------------------------------\n");
		str = tmp + strlen(tmp);
		sprintf(str, "Total number of IDT Entries: %i\n",j); 
		
		if ((strlen(sendbuf)+strlen(tmp)+1)<(BUFSIZE - strlen(MAGIC_COOKIE))){
			strcat(sendbuf, tmp);	

			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		}
		else { /*we send the sendbuf with "magic cookie" */
			sendWithCookie(sendbuf, strlen(sendbuf)+1, &status_clisock, replyto);

			sendSocketWrapper(&status_clisock, AF_INET, replyto, tmp, strlen(tmp));			
		}
		
		
              
        }
	else if ( !strncmp(token, "show_KVR", 8) )
	{
		struct KVREntry *tmpKVRE = NULL;
                int j = 0;
                char *tmpName=NULL;
                str = sendbuf;
                
                sprintf(str, "Entry Hash Table size: %i\n",kValRep.sizeEntryTab); str = sendbuf + strlen(sendbuf);
                sprintf(str, "Key Hash Table size: %i\n",kValRep.sizeKeyTab); str = sendbuf + strlen(sendbuf);
                
                sprintf(str, "Non permanent entries expire after not beeing used for %i heart beats\n",kValRep.deadEntryLimit);
		str = sendbuf + strlen(sendbuf);
                sprintf(str, "Heart beat interval : %d seconds\n",HEART_BEAT_INTERVAL); str = sendbuf + strlen(sendbuf);
		 sprintf(str, "KVR Entries Listing:\n\n");str = sendbuf + strlen(sendbuf);
                sprintf(str, "--------------------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
         
	       	sprintf(str, "%15s %15s %10s %8s %15s\n", "Name", "Owner", "IDP  ", "Age", "Keywords" );   
	   	str = sendbuf + strlen(sendbuf);
		sprintf(str, "--------------------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
             
	         
		for (i = 0; i < kValRep.sizeEntryTab; i++){
			tmpKVRE = kValRep.entryTab[i];
			
                        while (tmpKVRE != NULL){
                                struct KVRKeyword * key;
				j++;
				tmpName = handle2name(&BT, tmpKVRE->owner);
				
		         	if(tmpName==NULL)
					tmpName = "unknown";
					
				if (tmpKVRE->age != -1)	
					sprintf(tmp, "%15s %15s   %8x %8d     ", tmpKVRE->name, tmpName, tmpKVRE->IDP, tmpKVRE->age); 
                              	else
			      		sprintf(tmp, "%15s %15s   %8x %8s     ", tmpKVRE->name, tmpName, tmpKVRE->IDP, "Perm");
			       
			        str = tmp + strlen(tmp);
                               				       
			        key = tmpKVRE->keyWordsList;
                                while (key) {
                                        sprintf(str, "%s",key->keyWord);
					str = tmp + strlen(tmp);
                                        key = key->next;
					if (key)
					{
						sprintf(str, ","); str++;
					}
                                }
                                sprintf(str, "\n");
                        	
				
				if ((strlen(sendbuf)+strlen(tmp)+1)<(BUFSIZE - strlen(MAGIC_COOKIE)))
					strcat(sendbuf, tmp);	
				else {
					sendWithCookie(sendbuf, strlen(sendbuf)+1, &status_clisock, replyto);
					strcpy(sendbuf, tmp);	
				}
				
				tmpKVRE = tmpKVRE->next;
			}
                }
		
		memset(tmp, 0, BUFSIZE);
		sprintf(tmp, "--------------------------------------------------------------------------------------------\n");
		str = tmp + strlen(tmp);
		sprintf(str, "Total number of KVR Entries: %i\n",j); 
		
		if ((strlen(sendbuf)+strlen(tmp)+1)<(BUFSIZE - strlen(MAGIC_COOKIE))){
			strcat(sendbuf, tmp);

			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		}
		else { 
			sendWithCookie(sendbuf, strlen(sendbuf)+1, &status_clisock, replyto);

			sendSocketWrapper(&status_clisock, AF_INET, replyto, tmp, strlen(tmp));			
		}
    		
	}
	else if ( !strncmp(token, "show_bricks", 11) )
	{
          	struct BTentry_s *tmpBTE;
		str = sendbuf;
                
                sprintf(str, "Bricks are automatically detached after being inactive for %i heart beats\n", DEAD_CLIENT_THRESHOLD); 
               	str = sendbuf + strlen(sendbuf);
	        sprintf(str, "Hash table for storing bricks is of size %d\n", BT.size); 
               	str = sendbuf + strlen(sendbuf);
                sprintf(str, "-----------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
		sprintf(str, "%15s %8s %8s\n", "Name", "Age", "Mode" );
		str = sendbuf + strlen(sendbuf);
		sprintf(str, "-----------------------------------------------------------------------------------\n");
		str = sendbuf + strlen(sendbuf);
               	       
		for(i = 0; i < (int) BT.size; i++){
			tmpBTE = BT.tab[i];
                        while( tmpBTE!= NULL){
				
				if(tmpBTE->pluginFlag)			
	                           	sprintf(tmp, "%15s %8d %8s\n",tmpBTE->brickName, tmpBTE->age, "plugin"); 
                               	else
					sprintf(tmp, "%15s %8d %8s\n",tmpBTE->brickName, tmpBTE->age, "gates"); 
                               	if((strlen(sendbuf)+strlen(tmp)+1)<(BUFSIZE - strlen(MAGIC_COOKIE))){
					strcat(sendbuf, tmp);	
				}
				else{ /*we send the sendbuf with "magic cookie" */
					sendWithCookie(sendbuf, strlen(sendbuf)+1, &status_clisock, replyto);
					/*we copy tmp's content into sendbuf*/
					strcpy(sendbuf, tmp);	
				}
				tmpBTE = tmpBTE->next;
			}
                }
		
		memset(tmp, 0, BUFSIZE);
		sprintf(tmp, "-----------------------------------------------------------------------------------\n");
		str = tmp + strlen(tmp);
		sprintf(str, "Total number of Bricks: %i\n", BT.nbBricks); 
		
		if((strlen(sendbuf)+strlen(tmp)+1)<(BUFSIZE - strlen(MAGIC_COOKIE))){
			strcat(sendbuf, tmp);

			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
		}
		else { /*we send the sendbuf with "magic cookie" */
			
			sendWithCookie(sendbuf, strlen(sendbuf)+1, &status_clisock, replyto);
			sendSocketWrapper(&status_clisock, AF_INET, replyto, tmp, strlen(tmp));			
		}
    		
	}
#ifdef USER
	else if ( !strncmp(token, "load_brick", 10) )
	{	
		void *handle = NULL;
		struct pluginit_s *pluginit = NULL;
		int (*so_brick_init)(void*);
		char *(*so_brickname)(void);
		/* first we extract the path of the brick .so */
		char *path = strsep(&ptr, " "); /* now ptr points to the auxiliary arguments to be passed to the brick */
		char *error = NULL;
		unsigned long int ret;
		if (path==NULL){
			strcpy(sendbuf, "--> Invalid plugin path!\n");		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
		
		/* we load the brick .so */
		handle = dlopen(path, RTLD_LAZY);
		if (!handle){
			sprintf(sendbuf, "--> Unable to open plugin err = %s!\n", dlerror());		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
		
			
		/* we resolve the symbol of the init function */
		so_brick_init = dlsym(handle, "so_brick_init");
		error = dlerror();
		if (error != NULL)
		{
			anaPrint(ANA_ERR, "Error resolving plugin init symbol: %s\n", error);
			dlclose(handle);
			sprintf(sendbuf, "--> Unable to resolve plugin's initialisation symbol, dlerror()=%s !\n",error);		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
		
		/* we resolve the symbol of the brickname function */
		so_brickname = dlsym(handle, "so_brickname");
		error = dlerror();
		if (error != NULL)
		{
			anaPrint(ANA_ERR, "Error resolving plugin brickname symbol: %s\n", error);
			dlclose(handle);
			sprintf(sendbuf, "--> Unable to resolve plugin's brickname symbol, dlerror()=%s !\n",error);		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
				
		/* we initialize the brick in a new thread */
		pluginit = (struct pluginit_s *) malloc(sizeof(struct pluginit_s));
		if(ptr!=NULL){
			pluginit->commandLine = (char *) malloc(strlen(ptr)+1);
			strcpy(pluginit->commandLine, ptr);
		}
		pluginit->pluginHandle = handle;
		
		ret = (unsigned long int ) _anaThread(0, (void *)so_brick_init, pluginit, so_brickname());
	
		if(ret < 0){
			
			free(pluginit->commandLine);
			free(pluginit);
			strcpy(sendbuf, "--> Unable to launch plugins thread!\n");		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			dlclose(handle);
				
			return 1;
		}
		
		anaPrint(ANA_NOTICE, "Started init thread for new brick \"%s\"\n", (char *)so_brickname());
		strcpy(sendbuf, "--> brick's thread launched!\n");		
		sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
	}
	else if ( !strncmp(token, "unload_brick", 12) )
	{
		int ret;
		char *brickname = strsep(&ptr, " "); /* should point to brick name */
		anaHandle_t h;
	
		if (!brickname)
		{
			strcpy(sendbuf, "--> Invalid brick name!\n");		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
		
		/* try to find brick handle */
	
		h = brickname2handle(&BT, brickname);
		if (!h)
		{
			strcpy(sendbuf, "--> Cannot find brick with that name!\n");		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
		ret = kickBrick(h);
		if (!ret)
		{
			strcpy(sendbuf, "--> successfully unloaded brick\n");		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
		else
		{
			strcpy(sendbuf, "--> problem when unloading brick!\n");		
			sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
			return 1;
		}
	}
#endif	
	else {
		strcpy(sendbuf, "--> Unknown command!\n");
		sendSocketWrapper(&status_clisock, AF_INET, replyto, sendbuf, strlen(sendbuf));
	}	
	
	return 1;
}
