#ifndef _ANANODE_H
#define _ANANODE_H

#include "anaCommon.h"

#ifdef USER
#include <sys/select.h>
#include <signal.h>
#include <pthread.h>
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <errno.h>
#include <poll.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <asm/types.h>
#include <sys/un.h>
#endif /* USER */

#ifdef KERNEL
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/byteorder/generic.h>
#endif /* KERNEL */

#include "anatimer.h"
#include "dispatchTable.h"
#include "clientTable.h"
#include "notifTable.h"
#include "nodeComp.h"
#include "statusInterface.h"
#include "minmex_decl.h"

#define HEART_BEAT_INTERVAL  10     /* define the interval in seconds for expecting  heartBeat from bricks  */
#define DEAD_CLIENT_THRESHOLD 10    /* define the number of heartBeat checks with no life sign from brick , 
				     * before declaring the brick as Dead */
#define DEAD_IDTENTRY_THRESHOLD 15  /* defines the number of heartBeat checks with no use of the entry 
				     * => over this limit we delete the entry */

#define MAX_NB_PERMANENT 128 
#define MAX_NB_VOLATILE 1024

#define DEFAULT_IDT_HASHMASK 10
#define DEFAULT_BT_HASHMASK 6
#define DEFAULT_NT_HASHMASK 6

#define DEFAULT_MTU 200


struct redirectAux_s{
	anaHandle_t handle;
	anaLabel_t label;
};


struct IDTaux_s {
	anaHandle_t callee; 			/* indicates the owner of the IDP */
	anaCallback_t callBackFct; 		/* function handling the IDP */
	void *aux;				/* auxilary argument of the callback function  */
	copyAuxFunction copyAux;
	freeAuxFunction freeAux;
	int thread;
};



/* The information dispatch table */
extern IDT_t IDT;
/* table storing the bricks */
extern BrickTable_t BT;

/* notification table */
extern notifTable_t NT;

/*array holding the subscription to all events of a certain type */
extern struct notifAux_s *allEvents[32];

extern int anaDebugLevel;

int sendData(void *msg, int len, anaLabel_t input, struct IDTaux_s *idtAux  );
int sendNotification(anaHandle_t handle, void *msg, int len);
int kickBrick(anaHandle_t handle);
int createsLoop(anaHandle_t originViewer, anaLabel_t originLabel, anaHandle_t destViewer, anaLabel_t destLabel);
int kickAllBricks(void);
int existsIDP(anaLabel_t label);

int forkData(void *msg, int len, anaLabel_t input, void *dest);
int redirectData(void *msg, int len, anaLabel_t input, struct redirectAux_s *dest);

int checkIDTAges(void);
int  clearBrickEvents(notifTable_t *Table, anaHandle_t owner);
int  minmex_notifyEvent(anaHandle_t notifier, struct anaEventNotif_s *notif );

#ifdef KERNEL
extern struct genl_family ANAFamGN;
extern struct workqueue_struct *wq;
void sleepfunction(unsigned long ptr);
#endif

#endif /* _ANANODE_H */

