/* C/minmex/kernel/KstatusInterface.c
 * 
 * kernel specific functions for command line interface
 * 
 * Author: Christophe Jelger, <Christophe.Jelger@unibas.ch>
 *	
 * Copyright (c) 2008, Christophe Jelger, University of Basel
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/stat.h>
#include <linux/string.h>
#include <linux/delay.h>
#include <linux/random.h>
#include <linux/net.h>
#include <linux/socket.h>
#include <linux/in.h>
#include <linux/mm.h>
#include <linux/types.h>
#include <linux/udp.h>


#include "../include/anaNode.h"
#include "../include/statusInterface.h"

#include <linux/workqueue.h>

struct socket *status_clisock = NULL;  /* socket for recv CLI commands */
struct wq_wrapper wq_status_data;

void kernel_status_exit(void)
{
      	if (status_clisock)
	{
		anaPrint(ANA_DEBUG, "Closing status_clisock socket\n");
                sock_release(status_clisock);
		status_clisock = NULL;
        }
}

void kernel_status_cli_data(struct work_struct *control){
	struct  wq_wrapper * foo = container_of(control, struct  wq_wrapper, worker);
	int len = 0;
	/* as long as there are messages in the receive queue of this socket*/
	while((len = skb_queue_len(&foo->sk->sk_receive_queue)) > 0){ 
		struct sk_buff *skb = NULL;
		struct sockaddr_in replyto;
		struct udphdr *udph;
		int udplen = sizeof(struct udphdr);

        	anaPrint(ANA_DEBUG, "packets in udp control queue: %i\n",len);
       	        skb = skb_dequeue(&foo->sk->sk_receive_queue);
	
		udph = (struct udphdr*) skb->data;
	
		memset(&replyto, 0, sizeof(struct sockaddr_in));
		replyto.sin_family = AF_INET;
		replyto.sin_addr.s_addr = in_aton("127.0.0.1");
		replyto.sin_port = udph->source;
	
		/* status_cli_command is shared (userspace + kernel code) */
		/* so we don't call it directly in the work queue */
		status_cli_command(skb->data + udplen, skb->len - udplen, &replyto);
        	kfree_skb(skb);
	}

}


int initStatusInterface(){
        struct sockaddr_in ctrl;    /* for the CLI */
        int t = 0;
	int i;
	
       	if ((t = sock_create(PF_INET, SOCK_DGRAM, IPPROTO_UDP, &status_clisock)) < 0)
	{
		anaPrint(ANA_ERR, "cannot create UDP CLI socket");
                return t;
	}

	memset(&ctrl, 0, sizeof(struct sockaddr_in));
	ctrl.sin_family = AF_INET;
	ctrl.sin_addr.s_addr = in_aton("127.0.0.1");
	ctrl.sin_port = htons(UDP_ANASTATUS);

	/* since one can run multiple minmexes on a device, each minmex */
	/* must use a different UDP port for the status socket */

	for (i=0; i<MAX_UDP_BIND_TRIES; i++)
	{
		ctrl.sin_port = htons(UDP_ANASTATUS+i);
		t = status_clisock->ops->bind(status_clisock, (struct sockaddr *) &ctrl, sizeof(ctrl));
	
		if (t>=0)
			break;
	}

	if (t>=0)
		anaPrint(ANA_NOTICE, "status_clisock bound to UDP port %u\n", ntohs(ctrl.sin_port));
	else
	{
                anaPrint(ANA_ERR, "cannot bind UDP status_clisock socket!");
                sock_release(status_clisock);
                status_clisock = NULL;
                return t;
	}
        
        /* callback to handle incoming UDP messages for CLI */
	status_clisock->sk->sk_data_ready = kernel_status_cli;
		
	/*the minmex's work queue wq will be used*/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,19)
	INIT_WORK(&wq_status_data.worker, kernel_status_cli_data);
#else
        INIT_WORK(&wq_status_data.worker, kernel_status_cli_data, &wq_status_data);      
#endif	

	return 0;

}



void kernel_status_cli(struct sock *sk, int len)
{
	wq_status_data.sk = sk;
	queue_work(wq, &wq_status_data.worker);
}



