/* C/shared/KanaControl.c
 * 
 * Module that ties together minmex and bricks in kernel space
 *
 * Author: Ariane Keller <ariane.keller@tik.ee.ethz.ch> 
 *
 * Copyright (c) 2008, Ariane Keller, Swiss federal institute of technology (ETH Zurich)
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 *
 */


#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/list.h>
#include <linux/timex.h>
#include <asm/msr.h>
#include "anaCommon.h"

static char * debug = "ANA_DEBUG";
module_param(debug, charp, 0);
MODULE_PARM_DESC(debug, "debug Level");

int anaDebugLevel;
MODULE_LICENSE("Dual BSD/GPL");
MODULE_AUTHOR("Ariane Keller");
struct anaMinmex
{
        struct list_head list;
        char name[MAXNAMELEN];
        int mode;
        struct anaMinmexFunctions *minmexFct;
};

struct anaMinmex minmexList;

int unregisterAnaMinmex(char * name){
        struct list_head *pos, *p2;
        list_for_each_safe(pos, p2, &minmexList.list){
                struct anaMinmex *tmp = list_entry(pos, struct anaMinmex, list);
                int len = strlen(name) < MAXNAMELEN ? strlen(name):MAXNAMELEN;
		if(!(memcmp(tmp->name, name, len))){
                        list_del(pos);
                        kfree(tmp);
			return 0;
                }
        }
        return -1;
}

EXPORT_SYMBOL_GPL (unregisterAnaMinmex);

struct anaMinmexFunctions * getMinmexFunctions(char * name){
        struct list_head *pos;
        list_for_each(pos, &minmexList.list){
                struct anaMinmex *tmp = list_entry(pos, struct anaMinmex, list);
                int len = strlen(name) < MAXNAMELEN ? strlen(name):MAXNAMELEN;
                if (!(memcmp(tmp->name, name, len))){
                        return tmp->minmexFct;
                }
        }
        return NULL;
}

EXPORT_SYMBOL_GPL (getMinmexFunctions);

int registerAnaMinmex(struct anaMinmexFunctions * anaMinmexFct, int mode, char * name){
        struct list_head *pos;
        int len;
        struct anaMinmex *tmp;
        list_for_each(pos, &minmexList.list){
                struct anaMinmex *tmp = list_entry(pos, struct anaMinmex, list);
                len = strlen(name) < MAXNAMELEN ? strlen(name):MAXNAMELEN;
                len = len > strlen(tmp->name) ? len:strlen(tmp->name);
                if(!(memcmp(tmp->name, name, len))){
                        return -1; /*already an entry with this name in the list*/
                }
        }
        tmp = kmalloc(sizeof(struct anaMinmex), GFP_KERNEL);
        tmp->mode = mode;
        tmp->minmexFct = anaMinmexFct;
        len = strlen(name) < MAXNAMELEN ? strlen(name):MAXNAMELEN; 
        memcpy(tmp->name, name, len);
        list_add_tail(&(tmp->list),&minmexList.list);
	return 0;
}

EXPORT_SYMBOL_GPL (registerAnaMinmex);


int test(int i) {
	return i+1;
}

static int __init anaControl_init( void )
{
        INIT_LIST_HEAD(&minmexList.list);
       	anaDebugLevel = ANA_DEBUG ;
	if(!memcmp(debug, "ANA_NONE", 8))
		anaDebugLevel = ANA_NONE;
	if(!memcmp(debug, "ANA_EMERG", 9))
		anaDebugLevel = ANA_EMERG;
	if(!memcmp(debug, "ANA_ERR", 7))
		anaDebugLevel = ANA_ERR;
	if(!memcmp(debug, "ANA_NOTICE", 10))
		anaDebugLevel = ANA_NOTICE;
	if(!memcmp(debug, "ANA_DEBUG", 9))
		anaDebugLevel = ANA_DEBUG;
        return 0;
}

static void __exit anaControl_exit( void )
{
        anaPrint(ANA_DEBUG, "EXIT ANA CONTROL\n");    
}

module_init(anaControl_init);
module_exit(anaControl_exit);
