#! /bin/sh

#############################################################################
# create_referencefile = 1 -> a new referencefile is generated
# create_referencefile = 0 -> test output is compared against referencefile
############################################################################# 
create_referencefile=0
user=$1		#perform user space validation (derived from validation.sh script)
kernel=$2	#perform kernel space validation

userok=0;	#whether tests have failed or not
kernelok=0;
ok=0
#$1 newly created file $2 reference file
user_validate () {
## verify output ##
## we need to cut the date in front of each message:
#awk '{ print (substr($0,28)) }' $1 > test_ana/validOut/tmp
if [ $create_referencefile == 0 ]; then
	cmp -s $1 $2
	if [ $? == 0 ]; then
	#	echo "Test output agrees with reference output"
		ok=1
	else
		echo "Test output differs from reference output"
		diff $1 $2
		ok=0
	fi
else
	cp $1 $2
	ok=1
fi #create_referencefile
## remove current result file ##
rm $1
}


#########################
# user space validation #
#########################
if [ $user == 0 ]; then
#	echo "no user space validation"
	userok=1
else
#####   Test the MINMEX  #####
echo "validate user mode Minmex"
## start Bricks ##
./bin/minmex -D ANA_VALIDATE &> /dev/null &

## wait until test has finished ##
sleep 2

## kill all bricks ##
killall -s SIGINT minmex

## verify output ##
ok=0
user_validate test_ana/validOut/anaMinmex test_ana/referenceData/referencedata_anaMinmex.txt
userok=$ok

#####  Test ANA misc functionality (timer, locks, threads) #####
echo "validate user mode ANA misc functionality (timer, locks, threads)"
## start Bricks ##
./bin/minmex -D ANA_DEBUG &> /dev/null &
sleep 1
./bin/mxconfig load brick $PWD/so/validate_misc.so D=ANA_VALIDATE > /dev/null
## wait until test has finished ##
sleep 6

## kill all bricks ##
./bin/mxconfig unload brick validate_misc > /dev/null
killall -s SIGINT minmex

## verify output ##
ok=0
user_validate test_ana/validOut/validate_misc test_ana/referenceData/referencedata_validate_misc.txt
if [ $userok == 1 ] && [ $ok == 1 ]; then
	userok=1
else
	userok=0
fi


#####  Test the AL0 functionality  #####
echo "validate user mode AL0 functionality"
## start Bricks ##
./bin/minmex -D ANA_DEBUG &> /dev/null &
sleep 1
./bin/mxconfig load brick $PWD/so/validate_al0.so D=ANA_VALIDATE > /dev/null

## wait until test has finished ##
sleep 10

## kill all bricks ##
./bin/mxconfig unload brick validate_al0 > /dev/null
killall -s SIGINT minmex

## verify output ##
ok=0
user_validate test_ana/validOut/validate_al0 test_ana/referenceData/referencedata_validate_al0.txt
if [ $userok == 1 ] && [ $ok == 1 ]; then
	userok=1
else
	userok=0
fi

fi  #user


###########################
# kernel space validation #
###########################
if [ $kernel == 0 ]; then
#	echo "no kernel space validation"
	kernelok=1
else
#####   Test the MINMEX  #####
echo "validate kernel mode Minmex"
## start Bricks ##
insmod modules/anaControl.ko
insmod modules/anaMinmex.ko D=ANA_VALIDATE

## wait until test has finished ##
sleep 2

## verify output ##
cp /proc/anaMinmex/anaMinmex test_ana/validOut/anaMinmex
ok=0
user_validate test_ana/validOut/anaMinmex test_ana/referenceData/referencedata_anaMinmex.txt
kernelok=$ok

##remove the bricks
rmmod anaMinmex
rmmod anaControl
if [ $? != 0 ]; then
kernelok=0
fi


#####  Test ANA misc functionality (timer, locks, threads) #####
echo "validate kernel mode ANA misc functionality (timer, locks, threads)"
## start Bricks ##
insmod modules/anaControl.ko
insmod modules/anaMinmex.ko
insmod modules/anaValidateMisc.ko D=ANA_VALIDATE

## wait until test has finished ##
sleep 6


## verify output ##
cp /proc/validate_misc/validate_misc test_ana/validOut/anaValidateMisc
ok=0
user_validate test_ana/validOut/anaValidateMisc test_ana/referenceData/referencedata_validate_misc.txt
if [ $kernelok == 1 ] && [ $ok == 1 ]; then
	kernelok=1
else
	kernelok=0
fi


##remove the bricks
rmmod anaValidateMisc
rmmod anaMinmex
rmmod anaControl
if [ $? != 0 ]; then
kernelok=0
fi


#####  Test the AL0 functionality  #####
echo "validate kernel mode AL0 functionality"
## start Bricks ##
insmod modules/anaControl.ko
insmod modules/anaMinmex.ko D=ANA_DEBUG
insmod modules/anaValidateAL0.ko D=ANA_VALIDATE

## wait until test has finished ##
sleep 10

## verify output ##
cp /proc/validate_al0/validate_al0 test_ana/validOut/anaValidateAL0
ok=0
user_validate test_ana/validOut/anaValidateAL0 test_ana/referenceData/referencedata_validate_al0.txt
if [ $kernelok == 1 ] && [ $ok == 1 ]; then
	kernelok=1
else
	kernelok=0
fi

##remove the bricks
rmmod anaValidateAL0
rmmod anaMinmex
rmmod anaControl
if [ $? != 0 ]; then
kernelok=0
fi

fi

###########################
#     check validation    #
###########################
if [ $userok == 0 -o $kernelok == 0 ]; then
	echo "there was an errror during basic validation"
	exit 1
else
	echo "basic validation successful"
	exit 0
fi


