#! /bin/sh

#############################################################################
# create_referencefile = 1 -> a new referencefile is generated
# create_referencefile = 0 -> test output is compared against referencefile
############################################################################# 
create_referencefile=0
user=$1		#perform user space validation (derived from validation.sh script)
kernel=$2	#perform kernel space validation

userok=0;	#whether tests have failed or not
kernelok=0;

ok=0;

#$1 newly created file $2 reference file
user_validate () {
	## verify output ##
	## we need to cut the date in front of each message:
	#awk '{ print (substr($0,28)) }' $1 > test_ana/validOut/tmp
	if [ $create_referencefile == 0 ]; then
		cmp -s $1 $2
		if [ $? == 0 ]; then
		#	echo "Test output agrees with reference output"
			ok=1
		else
			echo "Test output differs from reference output"
			diff $1 $2
			ok=0
		fi
	else
		cp $1 $2
		ok=1
	fi #create_referencefile
	## remove current result file ##
	rm $1
	return
}


#########################
# user space validation #
#########################
if [ $user == 0 ]; then
#	echo "no user space validation"
	userok=1
else

##### Test IP Brick, purely node local test is not really helpful... #####
echo "validate IP Brick in user space"
./bin/minmex &> /dev/null &

sleep 2
./bin/mxconfig load brick $PWD/so/vlink.so > /dev/null
sleep 2
./bin/vlconfig create 1 > /dev/null
./bin/vlconfig add_if vlink1 lo > /dev/null
./bin/vlconfig up vlink1 > /dev/null

./bin/mxconfig load brick $PWD/so/eth-vl.so > /dev/null
sleep 2
./bin/mxconfig load brick $PWD/so/ip_enc.so D=ANA_VALIDATE > /dev/null
sleep 2
./bin/mxconfig load brick $PWD/so/ip_sum.so > /dev/null
sleep 2
./bin/mxconfig load brick $PWD/so/ip_fwd.so > /dev/null
sleep 2
./bin/mxconfig load brick $PWD/so/gatesPlug.so -c unix:///tmp/anaControl_gatesPlug_abcd &> /dev/null
sleep 1
./bin/ip_cfg -n unix:///tmp/anaControl_gatesPlug_abcd -c unix:///tmp/anaControl_gatesPlug_cdab -d unix:///tmp/anaData_gatesPlug_cdab -D ANA_VALIDATE -i 10.0.0.2 -m 255.255.255.0 -e eth01 &> /dev/null &
sleep 1
./bin/ip_cfg -n unix:///tmp/anaControl_gatesPlug_abcd -c unix:///tmp/anaControl_gatesPlug_cdab2 -d unix:///tmp/anaData_gatesPlug_cdab2 -D ANA_VALIDATE -g 10.0.0.1 &> /dev/null &


## wait until test has finished ##
sleep 5

## kill all bricks ##
killall -s SIGINT minmex

## verify output ##
ok=0
user_validate test_ana/validOut/ip_local test_ana/referenceData/referencedata_ip_local.txt
userok=$ok

ok=0
user_validate test_ana/validOut/ip_config test_ana/referenceData/referencedata_ip_config.txt
if [ $userok == 1 ] && [ $ok == 1 ]; then
	userok=1
else
	userok=0
fi


fi #user


###########################
# kernel space validation #
###########################
if [ $kernel == 0 ]; then
#	echo "no kernel space validation"
	kernelok=1
else
	##### ip #####
	echo "validate IP Brick in kernel space"

insmod modules/anaControl.ko
insmod modules/anaMinmex.ko
insmod modules/anaVlink.ko 
sleep 1
./bin/vlconfig create 1 > /dev/null
./bin/vlconfig add_if vlink1 lo > /dev/null
./bin/vlconfig up vlink1 > /dev/null
sleep 1
insmod modules/eth-vl1.ko
insmod modules/ip_main.ko D=ANA_VALIDATE
insmod modules/ip_forward.ko 
insmod modules/ip_checksum.ko 
sleep 1
insmod modules/anaGates.ko a=c_genetlink://ANA,d_genetlink://ANA
sleep 1
./bin/ip_cfg -n genetlink://ANA -d genetlink://ANA -c genetlink://ANA -D ANA_VALIDATE -i 10.0.0.2 -m 255.255.255.0 -e eth01 &> /dev/null &
sleep 1
./bin/ip_cfg -n genetlink://ANA -d genetlink://ANA -c genetlink://ANA -D ANA_VALIDATE -g 10.0.0.1 &> /dev/null &
sleep 1
insmod modules/ip_routing.ko
sleep 5

## verify output ##
cp /proc/ip_local/ip_local test_ana/validOut/ip_local
ok=0
user_validate test_ana/validOut/ip_local test_ana/referenceData/referencedata_ip_local.txt
kernelok=$ok
ok=0
user_validate test_ana/validOut/ip_config test_ana/referenceData/referencedata_ip_config.txt
if [ $kernelok == 1 ] && [ $ok == 1 ]; then
	kernelok=1
else
	kernelok=0
fi

rmmod ip_routing
rmmod ip_forward
rmmod ip_checksum
rmmod ip_main
rmmod eth-vl1
rmmod anaVlink
rmmod anaGates
rmmod anaMinmex
rmmod anaControl


fi

###########################
#     check validation    #
###########################
if [ $userok == 0 -o $kernelok == 0 ]; then
	echo "there was an errror during validation of IP"
	exit 1
else
	echo "validation of IP successful"
	exit 0
fi

