#! /bin/sh

#############################################################################
# create_referencefile = 1 -> a new referencefile is generated
# create_referencefile = 0 -> test output is compared against referencefile
############################################################################# 
create_referencefile=0
user=$1		#perform user space validation (derived from validation.sh script)
kernel=$2	#perform kernel space validation

userok=0;	#whether tests have failed or not
kernelok=0;

ok=0;	
#$1 newly created file $2 reference file
user_validate () {
	## verify output ##
	## we need to cut the date in front of each message:
	#awk '{ print (substr($0,28)) }' $1 > test_ana/validOut/tmp
	if [ $create_referencefile == 0 ]; then
		cmp -s $1 $2
		if [ $? == 0 ]; then
		#	echo "Test output agrees with reference output"
			ok=1
		else
			echo "Test output differs from reference output"
			diff $1 $2
			ok=0
		fi
	else
		cp $1 $2
		ok=1
	fi #create_referencefile
	## remove current result file ##
	rm $1
	return
}


#########################
# user space validation #
#########################
if [ $user == 0 ]; then
#	echo "no user space validation"
	userok=1
else
#####   Test the brickAL0_template  #####
echo "validate user mode API AL0 template brick"
## start Bricks ##
./bin/minmex &> /dev/null &
sleep 1
./bin/mxconfig load brick $PWD/so/brickAL0_template.so D=ANA_VALIDATE > /dev/null
## wait until test has finished ##
sleep 8 

## kill all bricks ##
killall -s SIGINT minmex

## verify output ##
ok=0
user_validate test_ana/validOut/brick0 test_ana/referenceData/referencedata_brick0.txt
userok=$ok


#####   Test the brickAL2_template_*  #####
echo "validate user mode API AL2 template bricks"
## start Bricks ##
./bin/minmex &> /dev/null &
sleep 1
./bin/mxconfig load brick $PWD/so/brickAL2_template.so D=ANA_VALIDATE > /dev/null
sleep 1
./bin/mxconfig load brick $PWD/so/brickAL2_template2.so D=ANA_VALIDATE > /dev/null

## wait until test has finished ##
sleep 8 

## kill all bricks ##
killall -s SIGINT minmex

## verify output ##
ok=0
user_validate test_ana/validOut/brick2_client test_ana/referenceData/referencedata_brick2_client.txt
userok=$ok
ok=0
user_validate test_ana/validOut/brick2_server test_ana/referenceData/referencedata_brick2_server.txt

if [ $userok == 1 ] && [ $ok == 1 ]; then
	userok=1
else
	userok=0
fi

fi

###########################
# kernel space validation #
###########################
if [ $kernel == 0 ]; then
#	echo "no kernel space validation"
	kernelok=1
else
#####   Test the API AL0 template brick  #####
echo "validate kernel mode API AL0 template brick"
## start Bricks ##
insmod modules/anaControl.ko
insmod modules/anaMinmex.ko
insmod modules/anaBrickAL0.ko D=ANA_VALIDATE

## wait until test has finished ##
sleep 8

## verify output ##
ok=0
cp /proc/brick0/brick0 test_ana/validOut/brick0
user_validate test_ana/validOut/brick0 test_ana/referenceData/referencedata_brick0.txt
kernelok=$ok

##remove the bricks
rmmod anaBrickAL0
rmmod anaMinmex
rmmod anaControl
if [ $? != 0 ]; then
kernelok=0
fi

#####   Test the API AL2 template bricks  #####
echo "validate kernel mode API AL2 template brick"
## start Bricks ##
insmod modules/anaControl.ko
insmod modules/anaMinmex.ko
insmod modules/anaBrickAL2_server.ko D=ANA_VALIDATE
sleep 1
insmod modules/anaBrickAL2_client.ko D=ANA_VALIDATE

## wait until test has finished ##
sleep 8

## verify output ##
ok=0
cp /proc/brick2_server/brick2_server test_ana/validOut/brick2_server
user_validate test_ana/validOut/brick2_server test_ana/referenceData/referencedata_brick2_server.txt
kernelok=$ok
ok=0
cp /proc/brick2_client/brick2_client test_ana/validOut/brick2_client
user_validate test_ana/validOut/brick2_client test_ana/referenceData/referencedata_brick2_client.txt

if [ $kernelok == 1 ] && [ $ok == 1 ]; then
	kernelok=1
else
	kernelok=0
fi


##remove the bricks
rmmod anaBrickAL2_client
rmmod anaBrickAL2_server
rmmod anaMinmex
rmmod anaControl
if [ $? != 0 ]; then
kernelok=0
fi

fi


###########################
#     check validation    #
###########################
if [ $userok == 0 -o $kernelok == 0 ]; then
	echo "there was an errror during validation of the API template brick"
	exit 1
else
	echo "validation of the API template brick successful"
	exit 0
fi

